/*
 *  Semaphore.cpp
 *  Copyright (C) 2001-2003 by Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

#include "Threading/Threading.hpp"

using namespace Threading;

//! Post a signal that one ressource has been released.
void Semaphore::post(void)
{
   // acquire exclusive access to count
   lock();
   // increment count
   ++mCount;
   // if waiting threads exist, signal one of them to wake up
   if(mCount <= 0) signal();
   // release access to count
   unlock();
}

//! Try to acquire one ressource, but don't block if ressources are exhausted.
bool Semaphore::tryWait(void)
{
   // acquire exclusive access to count
   lock();
   // if ressource is exhausted, return false...
   if(mCount <= 0)
   {
      // after releasing access to count
      unlock();
      return false;
   }
   // decrement count
   --mCount;
   // if ressource is exhausted wait for ressource
   while(mCount < 0) Condition::wait();
   // release access to count
   unlock();
   // return true; ressource is acquired
   return true;
}

//! Wait to acquire one ressource.
void Semaphore::wait(void)
{
   // acquire exclusive access to count
   lock();
   // decrement count
   --mCount;
   // if ressource is exhausted wait for ressource
   while(mCount < 0) Condition::wait();
   // release access to count
   unlock();
}
