/*
 *  Thread.cpp
 *  Copyright (C) 2001-2003 by Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

#include "Threading/Threading.hpp"

#ifdef WIN32
#include <windows.h>
typedef HANDLE pthread_t;
#else // Unix...
#include <pthread.h>
#include <unistd.h>
#include <sys/errno.h>
#endif

using namespace Threading;

/*! \brief Create thread.

Method Thread::run must be called in order to start executing Thread::main.
*/
Thread::Thread(void) : mThread(0), mCancel(false), mRunning(false) 
{
}

/*! \brief Delete thread.

\attention Deleting a running thread can lead to unexpected results. The calling thread should always wait for this thread to terminate by calling Thread::wait prior to thread deletion. It can also call Thread::cancel in order to request early termination for this thread. But thread cancellation cannot be imposed upon this thread, it will be deferred to the next cancellation point. In order for this thread to allow rapid thread cancellation, it must thus call method Thread::makeCancellationPoint as often as possible when it is safe to cancel the thread. A running thread that omits cancellation points and never terminates will block the calling thread indefinitely.

A good strategy is to always define a destructor in the derived class, and to make this destructor wait for thread termination. For instance:
\code
class MyThread : public Threading::Thread {...};

MyThread::~MyThread(void) {wait();}
\endcode

If the destructor in the derived class (e.g. MyThread) does not wait for thread termination (or is not defined), the potential problem is that the compiler will delete all of its data members before calling the destructor of base class %Thread. Thus, the thread could still be running but with deleted data members, which indeed can lead to unexpected results! 
*/
Thread::~Thread(void)
{
   lock();
   if(mThread)
   {
      // try to cancel thread execution
      mCancel = true;
      // wait for thread termination
      if(mRunning) Condition::wait();
      // detach thread
      pthread_t* lThread = (pthread_t*) mThread;
#ifdef WIN32
      ::WaitForSingleObject(*lThread, INFINITE);
#else // Unix...
      ::pthread_join(*lThread, 0);
#endif
      // delete native structure
      delete lThread;
      mThread = 0;
   }
   unlock();
}

/*! \brief Cancel execution of thread.

The cancellation process requested by this function runs asynchronously with the calling thread. The later can call function Thread::wait in order to wait for thread termination. The cancelled thread will not terminate before it reaches its next cancellation point, or before it terminates by itself. Before deleting a cancelled thread, the caller should always wait for thread termination. 

The caller should always lock the thread (using the lock method) prior to asking for cancellation. If it wishes to wait immediately for thread termination, it should keep the thread locked and call the wait method with argument true (see Thread::wait). 
*/
void Thread::cancel(void)
{
   mCancel = true;
}

/*! \brief Check wheter or not the thread is currently running.

The caller should always lock the thread (using the lock method) prior to checking the running flag.
*/
bool Thread::isRunning(void)
{
   return mRunning;   
}

/*! \brief Create cancellation point.

This function should be called by Thread::main as often as possible in order to enable thread cancellation. A Thread::cancel request can only be honored at a future cancellation point.
*/
void Thread::makeCancellationPoint(void)
{
   lock();
   // exit thread if cancellation was requested
   if(mCancel)
   {
      // signal all that thread has terminated
      mRunning = false;
      broadcast();
      unlock();
      // now exit thread
#ifdef WIN32
      ::ExitThread(0);
#else // Unix...
      ::pthread_exit(0);
#endif
   }
   unlock();
}

/*! \brief Create and startup thread.

This function allocates memory for the native thread structure, and then creates the thread which starts up asynchronously. The calling thread will block until this thread as actually started to run. 

Any error will raise a Threading:Exception.
*/
void Thread::run(void)
{
   lock();
   // allocate native structure
   pthread_t* lThread = new pthread_t;
   if(lThread == 0) throw Exception(eOtherError, "Thread::run() can't allocate thread!");
   // create thread
#ifdef WIN32
   if((*lThread = CreateThread(0, 0, (LPTHREAD_START_ROUTINE)startup, this, 0, 0)) == 0)
#else // Unix
   if(::pthread_create(lThread, 0, startup, this) != 0)  
#endif
      throw Exception(eOtherError, "Thread::run() can't create thread!");
   mThread = lThread;
   // wait for thread start up
   if(!mRunning) Condition::wait();
   unlock();   
}

/*! \brief Sleep calling thread for \c inTime seconds. 

A negative value will throw a Threading::Exception.
*/
void Thread::sleep(float inTime)
{
   if(inTime < 0) throw Exception(eOtherError, "Thread::sleep() invalid time");
#ifdef WIN32
   // argument must be in milliseconds
   ::Sleep((unsigned int)(1000*inTime));
#else // Unix...
   // argument must be in microseconds
   if(::usleep((unsigned int)(1000000*inTime)) != 0)
      throw Exception(eOtherError, "Thread::sleep() can't sleep");
#endif
}

/*! \brief Startup thread.

This function is first called when the thread starts up and immediately calls function Thread::main. After main returns (or is cancelled), it cleans up by signalling any waiting thread, and then terminates.
*/
void* Thread::startup(void* inThread)
{
   Thread* lThread = (Thread*) inThread;
   // signal any waiting thread that this thread is starting to run
   lThread->lock();
   lThread->mRunning = true;
   lThread->signal();
   lThread->unlock();
   // execute main
   lThread->main();
   // signal all that thread as terminated
   lThread->lock();
   lThread->mRunning = false;
   lThread->broadcast();
   lThread->unlock();
   return 0;
}

/*! \brief %Wait for thread termination.

This method will wait for thread termination. If argument \c inLock is true (default), it will first lock the thread mutex before checking wheter or not the thread is still running. If the thread has already terminated, it will promptly return after unlocking the mutex. Otherwise, it will wait for the thread to terminate by itself. 

If argument \c inLock is false, this method assumes that the caller has already locked the mutex (using Thread::lock) and is responsible for unlocking it.
*/
void Thread::wait(bool inLock)
{
   if(inLock) lock();
   if(mRunning) Condition::wait();
   if(inLock) unlock();
}
