/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/CrossoverConstrainedOp.cpp
 *  \brief  Source code of class GP::CrossoverConstrainedOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:20 $
 */

#include "beagle/GP.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a GP constrained tree crossover operator.
 *  \param inMatingPbName Individual mating probability parameter name used in register.
 *  \param inDistribPbName Distribution probability parameter name used in register.
 *  \param inName Name of the operator.
 */
GP::CrossoverConstrainedOp::CrossoverConstrainedOp(std::string inMatingPbName,
                                                   std::string inDistribPbName,
                                                   std::string inName) :
  Beagle::GP::CrossoverOp(inMatingPbName, inDistribPbName, inName)
{ }


/*!
 *  \brief Mate two GP individuals for a constrained tree crossover.
 *  \param ioIndiv1   First individual to mate.
 *  \param ioContext1 Evolutionary context of the first individual.
 *  \param ioIndiv2   Second individual to mate.
 *  \param ioContext2 Evolutionary context of the second individual.
 *  \return True if the individuals are effectively mated, false if not.
 */
bool GP::CrossoverConstrainedOp::mate(Beagle::Individual& ioIndiv1, Beagle::Context& ioContext1,
                                      Beagle::Individual& ioIndiv2, Beagle::Context& ioContext2)
{
  // Initial parameters checks
  Beagle_AssertM(ioIndiv1.size() > 0);
  Beagle_AssertM(ioIndiv1.size() == ioIndiv2.size());
  Beagle_ValidateParameterM(mNumberAttempts->getWrappedValue()>0,"gp.try",">0");

  // Cast method arguments.
  GP::Individual& lIndiv1   = castObjectT<GP::Individual&>(ioIndiv1);
  GP::Individual& lIndiv2   = castObjectT<GP::Individual&>(ioIndiv2);
  GP::Context&    lContext1 = castObjectT<GP::Context&>(ioContext1);
  GP::Context&    lContext2 = castObjectT<GP::Context&>(ioContext2);

  // Get parameters in local values, with the total number of nodes of an individual.
  bool             lMatingDone     = false;
  float            lDistrProba     = mDistributionProba->getWrappedValue();
  unsigned int     lMaxTreeDepth   = mMaxTreeDepth->getWrappedValue();
  GP::Tree::Handle lOldTreeHandle1 = lContext1.getGenotypeHandle();
  unsigned int     lOldTreeIndex1  = lContext1.getGenotypeIndex();
  GP::Tree::Handle lOldTreeHandle2 = lContext2.getGenotypeHandle();
  unsigned int     lOldTreeIndex2  = lContext2.getGenotypeIndex();
  unsigned int     lSizeIndiv1     = 0;
  for(unsigned int i=0; i<lIndiv1.size(); i++) lSizeIndiv1 += lIndiv1[i]->size();

  Beagle_LogDebugM(  
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GP::CrossoverConstrainedOp",
    std::string("First individual to mate (before constrained GP crossover): ")+
    lIndiv1.serialize()
  );
  Beagle_LogDebugM(
    ioContext1.getSystem().getLogger(),
    "crossover", "Beagle::GP::CrossoverConstrainedOp",
    std::string("Second individual to mate (before constrained GP crossover): ")+
    lIndiv2.serialize()
  );
  
  // Crossover loop. Try the given number of attempts to mate two individuals.
  for(unsigned int lAttempt=0; lAttempt < mNumberAttempts->getWrappedValue(); lAttempt++) {

    // Choose a node in all the individual node.
    unsigned int lChoosenNode1 =
      lContext1.getSystem().getRandomizer().rollInteger(0, lSizeIndiv1-1);

    // Get the tree in which the choosen node is. Change the global node index to the tree's index.
    unsigned int lChoosenTree = 0;
    for(; lChoosenTree<lIndiv1.size(); lChoosenTree++) {
      if(lChoosenNode1 < lIndiv1[lChoosenTree]->size()) break;
      Beagle_AssertM(lChoosenNode1 >= lIndiv1[lChoosenTree]->size());
      lChoosenNode1 -= lIndiv1[lChoosenTree]->size();
    }
    Beagle_AssertM(lChoosenTree < lIndiv1.size());

    // Choose a type of node (branch or leaf) following the distribution probability and change the
    // node for another node of the same tree if the types mismatch.
    GP::Tree& lTree1 = *lIndiv1[lChoosenTree];
    if(lTree1.size() > 1) {
      bool lTypeNode1 =
        (lContext1.getSystem().getRandomizer().rollUniform(0.0, 1.0) < lDistrProba);
      while((lTree1[lChoosenNode1].mPrimitive->getNumberArguments() != 0) != lTypeNode1) {
        lChoosenNode1 = lContext1.getSystem().getRandomizer().rollInteger(0, lTree1.size()-1);
      }
    }

    // Choose a node in the second individual in the corresponding tree.
    GP::Tree& lTree2 = *lIndiv2[lChoosenTree];
    unsigned int lChoosenNode2 =
      lContext2.getSystem().getRandomizer().rollInteger(0, lTree2.size()-1);

    // Choose a type of node (branch or leaf) following the distribution probability and change the
    // node for another node of the same tree if the types mismatch.
    if(lTree2.size() > 1) {
      bool lTypeNode2 =
        (lContext2.getSystem().getRandomizer().rollUniform(0.0, 1.0) < lDistrProba);
      while((lTree2[lChoosenNode2].mPrimitive->getNumberArguments() != 0) != lTypeNode2) {
        lChoosenNode2 = lContext2.getSystem().getRandomizer().rollInteger(0, lTree2.size()-1);
      }
    }

    // Set the first context to the node of the first tree.
    // Check if depth is ok. Do a new crossover attempt if not.
    lTree1.setContextToNode(lChoosenNode1, lContext1);
    unsigned int lNewDepthTree1 =
      lContext1.getCallStackSize() + lTree2.getTreeDepth(lChoosenNode2) - 1;
    if(lNewDepthTree1 > lMaxTreeDepth) continue;

    // Set the first context to the node of the second tree.
    // Check if depth is ok. Do a new crossover attempt if not.
    lTree2.setContextToNode(lChoosenNode2, lContext2);
    unsigned int lNewDepthTree2 =
      lContext2.getCallStackSize() + lTree1.getTreeDepth(lChoosenNode1) - 1;
    if(lNewDepthTree2 > lMaxTreeDepth) continue;

    // Mate the trees.
    Beagle_LogVerboseM(
      ioContext1.getSystem().getLogger(),
      "crossover", "Beagle::GP::CrossoverConstrainedOp",
      std::string("Trying to mate the ")+uint2ordinal(lChoosenTree+1)+
      std::string(" trees")
    );
    Beagle_LogVerboseM(
      ioContext1.getSystem().getLogger(),
      "crossover", "Beagle::GP::CrossoverConstrainedOp",
      std::string("Trying to exchange the ")+uint2ordinal(lChoosenNode1+1)+
      std::string(" node of the first tree with the ")+uint2ordinal(lChoosenNode2+1)+
      std::string(" node of the second tree")
    );
       
    mateTrees(lTree1, lChoosenNode1, lContext1, lTree2, lChoosenNode2, lContext2);

    // If one tree is not valid, undo the crossover and do a new crossover attempt. 
    lContext1.setGenotypeHandle(lIndiv1[lChoosenTree]);
    lContext1.setGenotypeIndex(lChoosenTree);
    lContext2.setGenotypeHandle(lIndiv2[lChoosenTree]);
    lContext2.setGenotypeIndex(lChoosenTree);
    if(lTree1.validateSubTree(lChoosenNode1,lContext1) &&
       lTree2.validateSubTree(lChoosenNode2,lContext2)) {
      lMatingDone = true;
      Beagle_LogVerboseM(
        ioContext1.getSystem().getLogger(),
        "crossover", "Beagle::GP::CrossoverConstrainedOp",
        "Constrained tree GP crossover valid"
      );
      break; // The crossover is valid.
    }
    else {   // Undo crossover.
      Beagle_LogVerboseM(
        ioContext1.getSystem().getLogger(),
        "crossover", "Beagle::GP::CrossoverConstrainedOp",
        "Constrained tree GP crossover invalid"
      );
      mateTrees(lTree1, lChoosenNode1, lContext1, lTree2, lChoosenNode2, lContext2);
      continue;
    }
  }

  // Replace the contexts.
  lContext1.setGenotypeHandle(lOldTreeHandle1);
  lContext1.setGenotypeIndex(lOldTreeIndex1);
  lContext2.setGenotypeHandle(lOldTreeHandle2);
  lContext2.setGenotypeIndex(lOldTreeIndex2);

  if(lMatingDone) {
    Beagle_LogDebugM(
      ioContext1.getSystem().getLogger(),
      "crossover", "Beagle::GP::CrossoverConstrainedOp",
      std::string("First individual mated (after constrained tree GP crossover): ")+
      lIndiv1.serialize()
    );
    Beagle_LogDebugM(
      ioContext1.getSystem().getLogger(),
      "crossover", "Beagle::GP::CrossoverConstrainedOp",
      std::string("Second individual mated (after constrained tree GP crossover): ")+
      lIndiv2.serialize()
    );
  }
  else {
    Beagle_LogVerboseM(
      ioContext1.getSystem().getLogger(),
      "crossover", "Beagle::GP::CrossoverConstrainedOp",
      "No constrained tree GP crossover done"
    );
  }
  
  return lMatingDone;
}
