/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/Individual.cpp
 *  \brief  Source code of class GP::Individual.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.8 $
 *  $Date: 2004/08/23 05:06:27 $
 */

#include "beagle/GP.hpp"

using namespace Beagle;


/*!
 *  \brief Contruct a GP individual.
 *  \par Note:
 *    Use GP::Tree allocator to allocate the genotypes.
 */
GP::Individual::Individual() :
  Beagle::Individual(new GP::Tree::Alloc)
{ }


/*!
 *  \brief Contruct a GP individual from a GP tree allocator.
 *  \param inGenotypeAlloc Allocator of GP trees.
 */
GP::Individual::Individual(GP::Tree::Alloc::Handle inGenotypeAlloc) :
  Beagle::Individual(inGenotypeAlloc)
{ }


/*!
 *  \brief Contruct a GP individual from a genotype allocator and a fitness allocator.
 *  \param inGenotypeAlloc Allocator of GP trees.
 *  \param inFitnessAlloc Fitness allocator.
 *  \param inN Number of genotypes of the GP individual.
 */
GP::Individual::Individual(GP::Tree::Alloc::Handle inGenotypeAlloc,
                           Fitness::Alloc::Handle inFitnessAlloc,
                           size_type inN) :
  Beagle::Individual(inGenotypeAlloc, inFitnessAlloc, inN)
{ }


/*!
 *  \brief Interpret the GP individual.
 *  \param outResult Datum containing the result of the interpretation.
 *  \param ioContext GP evolutionary context.
 *  \throw Beagle::ObjectException When individual or tree are empty.
 *  \throw Beagle::AssertException When context is not correctly set.
 */
void GP::Individual::run(GP::Datum& outResult, GP::Context& ioContext)
{
  Beagle_AssertM(&ioContext.getIndividual() == this);
  if(empty()) throw Beagle_ObjectExceptionM("Could not interpret, individual has no trees!");
  if((*this)[0]->empty()) throw Beagle_ObjectExceptionM("Could not interpret, tree is empty!");
  Tree::Handle lOldTreeHandle = ioContext.getGenotypeHandle();
  unsigned int lOldTreeIndex  = ioContext.getGenotypeIndex();
  ioContext.setGenotypeIndex(0);
  ioContext.setGenotypeHandle((*this)[0]);
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "individual", "Beagle::GP::Individual",
    std::string("Running the ")+uint2ordinal(ioContext.getIndividualIndex()+1)+
    std::string(" individual")
  );
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "individual", "Beagle::GP::Individual",
    std::string("The individual is: ")+serialize()
  );
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "individual", "Beagle::GP::Individual",
    std::string("Executing the first tree root node \"")+
    (*(*this)[0])[0].mPrimitive->getName()+"\""
  );
  ioContext.pushCallStack(0);  
  (*(*this)[0])[0].mPrimitive->execute(outResult, ioContext);
  ioContext.popCallStack();
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "individual", "Beagle::GP::Individual",
    std::string("Result of execution: ")+outResult.serialize()
  );
  ioContext.setGenotypeIndex(lOldTreeIndex);
  ioContext.setGenotypeHandle(lOldTreeHandle);
}

