/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/InitFullOp.cpp
 *  \brief  Source code of class GP::InitFullOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.6 $
 *  $Date: 2004/07/04 21:45:31 $
 */

#include "beagle/GP.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a GP "full" intialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
GP::InitFullOp::InitFullOp(std::string inReproProbaName, std::string inName) :
  Beagle::InitializationOp(inReproProbaName, inName),
  mMaxTreeDepth(NULL),
  mMinTreeDepth(NULL)
{ }


/*!
 *  \brief Initialize the GP "full" intialization operator.
 *  \param ioSystem System of the evolution.
 */
void GP::InitFullOp::initialize(Beagle::System& ioSystem)
{
  Beagle::InitializationOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("gp.init.maxdepth")) {
    mMaxTreeDepth = castHandleT<UInt>(ioSystem.getRegister()["gp.init.maxdepth"]);
  } else {
    mMaxTreeDepth = new UInt(5);
    Register::Description lDescription(
      "Maximum initial tree depth",
      "UInt",
      "5",
      "Maximum depth for newly initialized trees."
    );
    ioSystem.getRegister().addEntry("gp.init.maxdepth", mMaxTreeDepth, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("gp.init.mindepth")) {
    mMinTreeDepth = castHandleT<UInt>(ioSystem.getRegister()["gp.init.mindepth"]);
  } else {
    mMinTreeDepth = new UInt(2);
    Register::Description lDescription(
      "Minimum initial tree depth",
      "UInt",
      "2",
      "Minimum depth for newly initialized trees."
    );
    ioSystem.getRegister().addEntry("gp.init.mindepth", mMinTreeDepth, lDescription);
  }
}


/*!
 *  \brief Initialize the trees of an individual using the "half-and-half" approach.
 *  \param outIndividual Individual to initialize.
 *  \param ioContext Evolution context.
 *  \throw Beagle::RunTimeException If the min/max depths are incorrectly set.
 */
void GP::InitFullOp::initIndividual(Beagle::Individual& outIndividual, Beagle::Context& ioContext)
{
#ifndef BEAGLE_NDEBUG
  if(*mMinTreeDepth > *mMaxTreeDepth) {
    std::string lMessage = "GP::InitFullOp::initIndividual: Minimum tree depth is superior to the ";
    lMessage += "maximum tree depth. Could not initialize the individuals!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
#endif // BEAGLE_NDEBUG
  Beagle_ValidateParameterM(mMinTreeDepth->getWrappedValue()>0,"gp.init.mindepth",">0");
  
  GP::Individual& lIndividual = castObjectT<GP::Individual&>(outIndividual);
  GP::Context& lContext = castObjectT<GP::Context&>(ioContext);

  lIndividual.resize(lContext.getSystem().getPrimitiveSuperSet().size());
  GP::Tree::Handle lOldTreeHandle = lContext.getGenotypeHandle();
  unsigned int lOldTreeIndex = lContext.getGenotypeIndex();
  for(unsigned int i=0; i<lIndividual.size(); i++) {
    lContext.setGenotypeHandle(lIndividual[i]);
    lContext.setGenotypeIndex(i);
    unsigned int lTreeDepth =
      lContext.getSystem().getRandomizer().rollInteger(mMinTreeDepth->getWrappedValue(),
                                                       mMaxTreeDepth->getWrappedValue());
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "initialization", "Beagle::GP::InitFullOp",
      std::string("Initializing \'full\' method the ")+
      uint2ordinal(i+1)+std::string(" tree with depth ")+uint2str(lTreeDepth)
    );
    
    initTreeFull(*lIndividual[i], lTreeDepth, lContext);
  }

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "initialization", "Beagle::GP::InitFullOp",
    std::string("Initialized individual: ")+lIndividual.serialize()
  );
      
  lContext.setGenotypeIndex(lOldTreeIndex);
  lContext.setGenotypeHandle(lOldTreeHandle);
}


/*!
 *  \brief Initialize a GP sub-tree of a specified depth using the "full" approach.
 *  \param ioTree Tree containing the sub-tree to initialize.
 *  \param inSubTreeDepth Depth of the sub-tree to initialize.
 *  \param ioContext Evolutionary context.
 *  \return Generated sub-tree size.
 */
unsigned int GP::InitFullOp::initSubTreeFull(GP::Tree& ioTree,
                                             unsigned int inSubTreeDepth,
                                             GP::Context& ioContext) const
{
  Beagle_AssertM(inSubTreeDepth>0);
  GP::PrimitiveSet& lPrimitSet =
    *ioContext.getSystem().getPrimitiveSuperSet()[ioContext.getGenotypeIndex()];
  GP::Primitive::Handle lPrimit = NULL;
  if(inSubTreeDepth == 1) {
    lPrimit = lPrimitSet.selectPrimitiveWithoutArgument(ioContext.getSystem());
    if(!lPrimit) {
      std::string lMessage = "There is no leaf (primitive without argument) in the ";
      lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
      lMessage += " primitive set!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
  else {
    lPrimit = lPrimitSet.selectPrimitiveWithArguments(ioContext.getSystem());
    if(!lPrimit) {
      std::string lMessage = "There is no branch (primitive with arguments) in the ";
      lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
      lMessage += " primitive set!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
  unsigned int lNodeIndex = ioTree.size();
  ioTree.push_back(GP::Node(lPrimit->giveReference(ioContext),0));
  unsigned int lSubTreeSize = 1;
  for(unsigned int i=0; i<ioTree[lNodeIndex].mPrimitive->getNumberArguments(); i++) {
    lSubTreeSize += initSubTreeFull(ioTree, inSubTreeDepth-1, ioContext);
  }
  ioTree[lNodeIndex].mSubTreeSize = lSubTreeSize;
  return lSubTreeSize;
}


/*!
 *  \brief Initialize a GP tree of a specified depth using the "full" approach.
 *  \param outTree Tree to initialize.
 *  \param inTreeDepth Depth of the tree to initialize.
 *  \param ioContext Evolutionary context.
 */
void GP::InitFullOp::initTreeFull(GP::Tree& outTree,
                                  unsigned int inTreeDepth,
                                  GP::Context& ioContext) const
{
  Beagle_AssertM(inTreeDepth>0);
  outTree.resize(0);
  ioContext.emptyCallStack();
  initSubTreeFull(outTree, inTreeDepth, ioContext);
}
