/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/InitGrowConstrainedOp.cpp
 *  \brief  Source code of class GP::InitGrowConstrainedOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.6 $
 *  $Date: 2004/07/04 21:45:31 $
 */

#include "beagle/GP.hpp"

#include <sstream>

using namespace Beagle;


/*!
 *  \brief Construct a constrained GP tree "grow" intialization operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
GP::InitGrowConstrainedOp::InitGrowConstrainedOp(std::string inReproProbaName,
                                                 std::string inName) :
  InitGrowOp(inReproProbaName, inName),
  mNumberAttempts(NULL)
{ }


/*!
 *  \brief Initialize a GP constrained sub-tree of a specified depth using the "grow" approach.
 *  \param ioTree Tree containing the sub-tree to initialize.
 *  \param inMinDepth Minimal depth of the sub-tree to initialize.
 *  \param inMaxDepth Maximal depth of the sub-tree to initialize.
 *  \param ioContext Evolutionary context.
 *  \return Generated sub-tree size.
 */
unsigned int GP::InitGrowConstrainedOp::initConstrainedSubTreeGrow(GP::Tree& ioTree,
                                                                   unsigned int inMinDepth,
                                                                   unsigned int inMaxDepth,
                                                                   GP::Context& ioContext) const
{
  Beagle_AssertM(inMaxDepth >= inMinDepth);
  Beagle_AssertM(inMinDepth>0);
  GP::PrimitiveSet& lPrimitSet =
    *ioContext.getSystem().getPrimitiveSuperSet()[ioContext.getGenotypeIndex()];
  GP::Primitive::Handle lPrimit = NULL;
  const unsigned int lNodeIndex = ioTree.size();
  for(unsigned int lAttempt=0; lAttempt < mNumberAttempts->getWrappedValue(); lAttempt++) {
    if(inMinDepth > 1) {
      lPrimit = lPrimitSet.selectPrimitiveWithArguments(ioContext.getSystem());
      if(!lPrimit) {
        std::string lMessage = "There is no branch (primitive with argument) in the ";
        lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
        lMessage += " primitive set!";
        throw Beagle_RunTimeExceptionM(lMessage);
      }
    }
    else if(inMaxDepth == 1) {
      lPrimit = lPrimitSet.selectPrimitiveWithoutArgument(ioContext.getSystem());
      if(!lPrimit) {
        std::string lMessage = "There is no leaf (primitive without arguments) in the ";
        lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
        lMessage += " primitive set!";
        throw Beagle_RunTimeExceptionM(lMessage);
      }
    }
    else {
      lPrimit = lPrimitSet.selectPrimitive(ioContext.getSystem());
      if(!lPrimit) {
        std::string lMessage = "There is no primitive in the ";
        lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
        lMessage += " primitive set!";
        throw Beagle_RunTimeExceptionM(lMessage);
      }
    }
    lPrimit = lPrimit->giveReference(ioContext);
    ioTree.push_back(GP::Node(lPrimit, 0));
    ioContext.pushCallStack(lNodeIndex);
    if(lPrimit->validate(ioContext)) {
      unsigned int lSubTreeSize = 1;
      unsigned int lMinDepth = (inMinDepth > 1) ? (inMinDepth-1) : 1;
      bool lGoodInit = true;
      for(unsigned int i=0; i<lPrimit->getNumberArguments(); i++) {
        unsigned int lArgSubTreeSize =
          initConstrainedSubTreeGrow(ioTree, lMinDepth, inMaxDepth-1, ioContext);
        if(lArgSubTreeSize == 0) {
          for(unsigned int j=1; j<lSubTreeSize; j++) ioTree.pop_back();
          lGoodInit = false;
          break;
        }
        lSubTreeSize += lArgSubTreeSize;
      }
      if(lGoodInit) {
        ioContext.popCallStack();
        ioTree[lNodeIndex].mSubTreeSize = lSubTreeSize;
        return lSubTreeSize;
      }
    }
    ioContext.popCallStack();
    ioTree.pop_back();
  }
  return 0;  // Could not initialize this node correctly, backtracking instead.
}


/*!
 *  \brief Initialize the constrained GP tree "grow" intialization operator.
 *  \param ioSystem System of the evolution.
 */
void GP::InitGrowConstrainedOp::initialize(Beagle::System& ioSystem)
{
  InitGrowOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("gp.try")) {
    mNumberAttempts = castHandleT<UInt>(ioSystem.getRegister()["gp.try"]);
  } else {
    mNumberAttempts = new UInt(2);
    std::string lLongDescrip = "Maximum number of attempts to modify a GP tree in a genetic ";
    lLongDescrip += "operation. As there is topological constraints on GP trees (i.e. tree ";
    lLongDescrip += "depth limit), it is often necessary to try a genetic operation several times.";
    Register::Description lDescription(
      "Max number of attempts",
      "UInt",
      "2",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("gp.try", mNumberAttempts, lDescription);
  }
}


/*!
 *  \brief Initialize a constrained GP tree of a specified depth using the "grow" approach.
 *  \param outTree Tree to initialize.
 *  \param inMinDepth Minimal depth of the tree to initialize.
 *  \param inMaxDepth Maximal depth of the tree to initialize.
 *  \param ioContext Evolutionary context.
 */
void GP::InitGrowConstrainedOp::initTreeGrow(GP::Tree& outTree,
                                             unsigned int inMinDepth,
                                             unsigned int inMaxDepth,
                                             GP::Context& ioContext) const
{
  Beagle_AssertM(inMaxDepth >= inMinDepth);
  Beagle_AssertM(inMinDepth>0);
  outTree.resize(0);
  ioContext.emptyCallStack();
  while(initConstrainedSubTreeGrow(outTree, inMinDepth, inMaxDepth, ioContext) == 0);
}

