/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/MutationShrinkOp.cpp
 *  \brief  Source code of class GP::MutationShrinkOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:20 $
 */

#include "beagle/GP.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a GP shrink mutation operator.
 *  \param inMutationPbName Mutation shrink probability.
 *  \param inName Name of the operator.
 */
GP::MutationShrinkOp::MutationShrinkOp(std::string inMutationPbName,
                                       std::string inName) :
  Beagle::MutationOp(inMutationPbName, inName)
{ }


/*!
 *  \brief Initialize the GP shrink mutation operator.
 *  \param ioSystem System of the evolution.
 */
void GP::MutationShrinkOp::initialize(Beagle::System& ioSystem)
{
  Beagle::MutationOp::initialize(ioSystem);
  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    ioSystem.getRegister().deleteEntry(mMutationPbName);
  }

  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    mMutationProba = castHandleT<Float>(ioSystem.getRegister()[mMutationPbName]);
  } else {
    mMutationProba = new Float(float(0.05));
    std::string lLongDescrip = "Shrink mutation probability for an individual. ";
    lLongDescrip += "Shrink mutation consists in replacing a branch (a node with one ";
    lLongDescrip += "or more arguments) with one of his child node. This erases ";
    lLongDescrip += "the chosen node and the other child nodes.";
    Register::Description lProbaDescription(
      "Individual shrink mutation prob.",
      "Float",
      "0.05",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry(mMutationPbName, mMutationProba, lProbaDescription);
  }
}


/*!
 *  \brief Shrink mutate a GP individual.
 *  \param ioIndividual GP individual to shrink mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool GP::MutationShrinkOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
  GP::Individual& lIndividual  = castObjectT<GP::Individual&>(ioIndividual);
  GP::Context& lContext        = castObjectT<GP::Context&>(ioContext);
  
  unsigned int lNbNodes = 0;
  for(unsigned int i=0; i<lIndividual.size(); i++) lNbNodes += lIndividual[i]->size();
  if(lNbNodes == 0) return false;
  unsigned int lChoosenNode = lContext.getSystem().getRandomizer().rollInteger(0, lNbNodes-1);
  unsigned int lChoosenTree = 0;
  for(; (lChoosenTree+1)<lIndividual.size(); lChoosenTree++) {
    if(lChoosenNode < lIndividual[lChoosenTree]->size()) break;
    else lChoosenNode -= lIndividual[lChoosenTree]->size();
  }
  GP::Tree::Handle lActualTree = lIndividual[lChoosenTree];
  if(lActualTree->size() < 2) return false;

  GP::Tree::Handle lOldTreeHandle = lContext.getGenotypeHandle();
  unsigned int lOldTreeIndex = lContext.getGenotypeIndex();

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationShrinkOp",
    std::string("Individual before GP shrink mutation: ")+
    ioIndividual.serialize()
  );
           
  lIndividual[lChoosenTree] = castHandleT<GP::Tree>(lIndividual.getTypeAlloc()->allocate());
  while((*lActualTree)[lChoosenNode].mPrimitive->getNumberArguments() == 0) {
    lChoosenNode = lContext.getSystem().getRandomizer().rollInteger(0, lActualTree->size()-1);
  }
  lIndividual[lChoosenTree]->clear();
  lIndividual[lChoosenTree]->insert(lIndividual[lChoosenTree]->end(),
                                    lActualTree->begin(),
                                    lActualTree->begin()+lChoosenNode);
  unsigned int lChoosenArg = lContext.getSystem().getRandomizer().rollInteger(0,
                               (*lActualTree)[lChoosenNode].mPrimitive->getNumberArguments()-1);
  unsigned int lChoosenArgIndex = lChoosenNode + 1;
  for(unsigned int k=0; k<lChoosenArg; k++) {
    lChoosenArgIndex += (*lActualTree)[lChoosenArgIndex].mSubTreeSize;
  }
  unsigned int lChoosenArgSubTreeSize = (*lActualTree)[lChoosenArgIndex].mSubTreeSize;
  lIndividual[lChoosenTree]->insert(lIndividual[lChoosenTree]->end(),
                                    lActualTree->begin()+lChoosenArgIndex,
                                    lActualTree->begin()+lChoosenArgIndex+lChoosenArgSubTreeSize);
  unsigned int lChoosenNodeSubTreeSize = (*lActualTree)[lChoosenNode].mSubTreeSize;
  lIndividual[lChoosenTree]->insert(lIndividual[lChoosenTree]->end(),
                                    lActualTree->begin()+lChoosenNode+lChoosenNodeSubTreeSize,
                                    lActualTree->end());
  lActualTree->setContextToNode(lChoosenNode, lContext);
  unsigned int lDiffSize =
    (*lActualTree)[lChoosenNode].mSubTreeSize - (*lActualTree)[lChoosenArgIndex].mSubTreeSize;
  for(unsigned int l=0; l<(lContext.getCallStackSize()-1); l++) {
    (*lIndividual[lChoosenTree])[lContext.getCallStackElement(l)].mSubTreeSize -= lDiffSize;
  }

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationShrinkOp",
    std::string("Replacing the ")+uint2ordinal(lChoosenNode+1)+
    std::string(" node of the ")+uint2ordinal(lChoosenTree+1)+
    std::string(" tree with its ")+uint2ordinal(lChoosenArg+1)+
    std::string(" argument, that is the ")+
    uint2ordinal(lChoosenArgIndex+1)+std::string(" node")
  );
  
  lContext.setGenotypeHandle(lOldTreeHandle);
  lContext.setGenotypeIndex(lOldTreeIndex);

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationShrinkOp",
    std::string("Individual after GP shrink mutation: ")+
    ioIndividual.serialize()
  );

  return true;
}

