/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/MutationStandardOp.cpp
 *  \brief  Source code of class GP::MutationStandardOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:20 $
 */

#include "beagle/GP.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a GP standard mutation operator.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inMaxRegenDepthName Maximum regeneration depth in mutation parameter name.
 *  \param inName Name of the operator.
 */
GP::MutationStandardOp::MutationStandardOp(std::string inMutationPbName,
                                           std::string inMaxRegenDepthName,
                                           std::string inName) :
  Beagle::MutationOp(inMutationPbName, inName),
  mMaxRegenDepthName(inMaxRegenDepthName)
{ }


/*!
 *  \brief Initialize the GP standard mutation operator.
 *  \param ioSystem System of the evolution.
 */
void GP::MutationStandardOp::initialize(Beagle::System& ioSystem)
{
  Beagle::MutationOp::initialize(ioSystem);
  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    ioSystem.getRegister().deleteEntry(mMutationPbName);
  }
  
  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    mMutationProba = castHandleT<Float>(ioSystem.getRegister()[mMutationPbName]);
  } else {
    mMutationProba = new Float(float(0.05));
    std::string lLongDescrip = "Standard mutation probability for an individual. ";
    lLongDescrip += "A standard mutation replaces a sub-tree with a randomly generated one.";
    Register::Description lProbaDescription(
      "Individual std mutation prob.",
      "Float",
      "0.05",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry(mMutationPbName, mMutationProba, lProbaDescription);
  }

  if(ioSystem.getRegister().isRegistered("gp.tree.maxdepth")) {
    mMaxTreeDepth = castHandleT<UInt>(ioSystem.getRegister()["gp.tree.maxdepth"]);
  } else {
    mMaxTreeDepth = new UInt(17);
    Register::Description lDescription(
      "Maximum tree depth",
      "UInt",
      "17",
      "Maximum allowed depth for the trees."
    );
    ioSystem.getRegister().addEntry("gp.tree.maxdepth", mMaxTreeDepth, lDescription);
  }

  if(ioSystem.getRegister().isRegistered(mMaxRegenDepthName)) {
    mMaxRegenerationDepth = castHandleT<UInt>(ioSystem.getRegister()[mMaxRegenDepthName]);
  } else {
    mMaxRegenerationDepth = new UInt(5);
    std::string lLongDescrip = "Maximum depth for standard mutation. ";
    lLongDescrip += "A standard mutation replaces a sub-tree with a randomly generated one.";
    Register::Description lDescription(
      "Maximum regeneration depth",
      "UInt",
      "5",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry(mMaxRegenDepthName, mMaxRegenerationDepth, lDescription);
  }
}


/*!
 *  \brief Initialize a GP sub-tree of a specified depth using the "grow" approach.
 *  \param ioTree Tree containing the sub-tree to initialize.
 *  \param inMinDepth Minimal depth of the sub-tree to initialize.
 *  \param inMaxDepth Maximal depth of the sub-tree to initialize.
 *  \param ioContext Evolutionary context.
 *  \return Generated sub-tree size.
 */
unsigned int GP::MutationStandardOp::initSubTreeGrow(GP::Tree& ioTree,
                                                     unsigned int inMinDepth,
                                                     unsigned int inMaxDepth,
                                                     GP::Context& ioContext) const
{
  Beagle_AssertM(inMaxDepth >= inMinDepth);
  Beagle_AssertM(inMinDepth>0);
  GP::PrimitiveSet& lPrimitSet =
    *ioContext.getSystem().getPrimitiveSuperSet()[ioContext.getGenotypeIndex()];
  GP::Primitive::Handle lPrimit = NULL;
  if(inMinDepth > 1) {
    lPrimit = lPrimitSet.selectPrimitiveWithArguments(ioContext.getSystem());
    if(!lPrimit) {
      std::string lMessage = "There is no branch (primitive with arguments) in the ";
      lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
      lMessage += " primitive set!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
  else if(inMaxDepth == 1) {
    lPrimit = lPrimitSet.selectPrimitiveWithoutArgument(ioContext.getSystem());
    if(!lPrimit) {
      std::string lMessage = "There is no leaf (primitive without argument) in the ";
      lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
      lMessage += " primitive set!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
  else {
    lPrimit = lPrimitSet.selectPrimitive(ioContext.getSystem());
    if(!lPrimit) {
      std::string lMessage = "There is no primitive in the ";
      lMessage += uint2ordinal(ioContext.getGenotypeIndex()+1);
      lMessage += " primitive set!";
      throw Beagle_RunTimeExceptionM(lMessage);
    }
  }
  unsigned int lNodeIndex = ioTree.size();
  ioTree.push_back(GP::Node(lPrimit->giveReference(ioContext),1));
  unsigned int lSubTreeSize = 1;
  unsigned int lMinDepth = (inMinDepth > 1) ? (inMinDepth-1) : 1;
  for(unsigned int i=0; i<ioTree[lNodeIndex].mPrimitive->getNumberArguments(); i++) {
    lSubTreeSize += initSubTreeGrow(ioTree, lMinDepth, inMaxDepth-1, ioContext);
  }
  ioTree[lNodeIndex].mSubTreeSize = lSubTreeSize;
  return lSubTreeSize;
}


/*!
 *  \brief Standard mutate a GP individual.
 *  \param ioIndividual GP individual to standard mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool GP::MutationStandardOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
  GP::Individual& lIndividual        = castObjectT<GP::Individual&>(ioIndividual);
  GP::Context& lContext              = castObjectT<GP::Context&>(ioContext);
  unsigned int lMaxTreeDepth         = mMaxTreeDepth->getWrappedValue();
  unsigned int lMaxRegenerationDepth = mMaxRegenerationDepth->getWrappedValue();

  unsigned int lNbNodes = 0;  
  for(unsigned int i=0; i<lIndividual.size(); i++) lNbNodes += lIndividual[i]->size();
  if(lNbNodes == 0) return false;
  unsigned int lChoosenNode = lContext.getSystem().getRandomizer().rollInteger(0, lNbNodes-1);           
  unsigned int lChoosenTree = 0;
  for(; (lChoosenTree+1)<lIndividual.size(); lChoosenTree++) {
    if(lChoosenNode < lIndividual[lChoosenTree]->size()) break;
    else lChoosenNode -= lIndividual[lChoosenTree]->size();
  }

  unsigned int lOldGenotypeIndex = lContext.getGenotypeIndex();
  GP::Tree::Handle lOldGenotypeHandle = lContext.getGenotypeHandle();

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationStandardOp",
    std::string("Individual before GP standard mutation: ")+
    ioIndividual.serialize()
  );
  
      
  GP::Tree::Handle lActualTree = lIndividual[lChoosenTree];
  GP::Tree::Handle lNewTree    = castHandleT<GP::Tree>(lIndividual.getTypeAlloc()->allocate());
  unsigned int lChoosenNodeSubTreeSize = (*lActualTree)[lChoosenNode].mSubTreeSize;
  lNewTree->insert(lNewTree->end(), lActualTree->begin(), lActualTree->begin()+lChoosenNode);
  lContext.setGenotypeIndex(lChoosenTree);
  lContext.setGenotypeHandle(lActualTree);
  lContext.emptyCallStack();
  lActualTree->setContextToNode(lChoosenNode, lContext);
  lContext.popCallStack();
  unsigned int lTreeDepth =
    lContext.getSystem().getRandomizer().rollInteger(1, lMaxRegenerationDepth);
  lTreeDepth = minOf<unsigned int>(lMaxTreeDepth - lContext.getCallStackSize(), lTreeDepth);
  lIndividual[lChoosenTree] = lNewTree;
  lContext.setGenotypeHandle(lNewTree);

  initSubTreeGrow(*lNewTree, 1, lTreeDepth, lContext);

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationStandardOp",
    std::string("GP standard mutate the ")+uint2ordinal(lChoosenNode+1)+
    std::string(" node of the ")+uint2ordinal(lChoosenTree+1)+
    std::string(" tree with depth ")+uint2str(lTreeDepth)
  ); 

  lNewTree->insert(lNewTree->end(),
                   lActualTree->begin()+lChoosenNode+lChoosenNodeSubTreeSize,
                   lActualTree->end());
  unsigned int lDiffSize =
    (*lActualTree)[lChoosenNode].mSubTreeSize - (*lNewTree)[lChoosenNode].mSubTreeSize;
  for(unsigned int l=0; l<lContext.getCallStackSize(); l++) {
    (*lNewTree)[lContext.getCallStackElement(l)].mSubTreeSize -= lDiffSize;
  }
  
  lContext.setGenotypeIndex(lOldGenotypeIndex);
  lContext.setGenotypeHandle(lOldGenotypeHandle);

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationStandardOp",
    std::string("Individual after GP standard mutation: ")+
    ioIndividual.serialize()
  );

  return true;
}


/*!
 *  \brief Read a mutation operator from XML tree.
 *  \param inNode XML node to use to read crossover operator.
 *  \param inOpMap Operator map to use to read crossover operator.
 */
void GP::MutationStandardOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string mMutationPbReadName = inNode->getTagAttribute("mutationpb");
  if(mMutationPbReadName.empty() == false) mMutationPbName = mMutationPbReadName;
  std::string mMaxRegenDepthReadName = inNode->getTagAttribute("maxdepth");
  if(mMaxRegenDepthReadName.empty() == false) mMaxRegenDepthName = mMaxRegenDepthReadName;
}


/*!
 *  \brief Write mutation operator into XML streamer.
 *  \param ioStreamer XML streamer to write mutation operator into.
 */
void GP::MutationStandardOp::writeContent(XMLStreamer& ioStreamer) const
{
  Beagle::MutationOp::writeContent(ioStreamer);
  ioStreamer.insertAttribute("maxdepth", mMaxRegenDepthName);
}

