/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/MutationSwapOp.cpp
 *  \brief  Source code of class GP::MutationSwapOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:20 $
 */

#include "beagle/GP.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a GP swap mutation operator.
 *  \param inMutationPbName Mutation probability parameter name used in register.
 *  \param inDistribPbName Swap mutation distribution probability parameter name.
 *  \param inName Name of the operator.
 */
GP::MutationSwapOp::MutationSwapOp(std::string inMutationPbName,
                                   std::string inDistribPbName,
                                   std::string inName) :
  Beagle::MutationOp(inMutationPbName, inName),
  mDistribPbName(inDistribPbName)
{ }


/*!
 *  \brief Initialize the GP swap mutation operator.
 *  \param ioSystem System of the evolution.
 */
void GP::MutationSwapOp::initialize(Beagle::System& ioSystem)
{
  Beagle::MutationOp::initialize(ioSystem);
  
  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    ioSystem.getRegister().deleteEntry(mMutationPbName);
  }

  if(ioSystem.getRegister().isRegistered(mMutationPbName)) {
    mMutationProba =
      castHandleT<Float>(ioSystem.getRegister().getEntry(mMutationPbName));
  } else {
    mMutationProba = new Float(float(0.05));
    std::string lLongDescrip = "Swap mutation probability for an individual. ";
    lLongDescrip += "Swap mutation consists in exchanging the primitive associated to a ";
    lLongDescrip += "node by one having the same number of arguments.";
    Register::Description lProbaDescription(
      "Individual swap mutation prob.",
      "Float",
      "0.05",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry(mMutationPbName, mMutationProba, lProbaDescription);
  }

  if(ioSystem.getRegister().isRegistered(mDistribPbName)) {
    mDistributionProba = castHandleT<Float>(
      ioSystem.getRegister().getEntry(mDistribPbName));
  } else {
    mDistributionProba = new Float(float(0.5));
    std::string lLongDescrip =
      "Probability that a swap mutation point is a branch (node with sub-trees). ";
    lLongDescrip += "Value of 1.0 means that all swap mutation points are branches, ";
    lLongDescrip += "and value of 0.0 means that all swap mutation points are leaves. ";
    lLongDescrip += "Swap mutation consists in exchanging the primitive associated to a ";
    lLongDescrip += "node by one having the same number of arguments.";
    Register::Description lDistributionDescription(
      "Swap mutation distrib. prob.",
      "Float",
      "0.5",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry(mDistribPbName, mDistributionProba, lDistributionDescription);
  }
}


/*!
 *  \brief Swap mutate a GP individual.
 *  \param ioIndividual GP individual to swap mutate.
 *  \param ioContext Context of the evolution.
 *  \return True if the individual is effectively mutated, false if not.
 */
bool GP::MutationSwapOp::mutate(Beagle::Individual& ioIndividual, Beagle::Context& ioContext)
{
  GP::Individual& lIndividual  = castObjectT<GP::Individual&>(ioIndividual);
  GP::Context& lContext        = castObjectT<GP::Context&>(ioContext);
  double lDistrProba           = mDistributionProba->getWrappedValue();
  
  unsigned int lNbNodes = 0;
  for(unsigned int i=0; i<lIndividual.size(); i++) lNbNodes += lIndividual[i]->size();
  if(lNbNodes == 0) return false;
  unsigned int lChoosenNode = lContext.getSystem().getRandomizer().rollInteger(0, lNbNodes-1);
  unsigned int lChoosenTree = 0;
  for(; (lChoosenTree+1)<lIndividual.size(); lChoosenTree++) {
    if(lChoosenNode < lIndividual[lChoosenTree]->size()) break;
    else lChoosenNode -= lIndividual[lChoosenTree]->size();
  }
  GP::Tree& lTree = *lIndividual[lChoosenTree];
  if(lTree.size() == 0) return false;
     
  GP::Tree::Handle lOldTreeHandle = lContext.getGenotypeHandle();
  unsigned int lOldTreeIndex = lContext.getGenotypeIndex();

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationSwapOp",
    std::string("Individual before GP swap mutation: ")+ioIndividual.serialize()
  );
  
  if(lTree.size() > 1) {
    bool lTypeNode = (lContext.getSystem().getRandomizer().rollUniform(0., 1.) < lDistrProba);
    while((lTree[lChoosenNode].mPrimitive->getNumberArguments() != 0) != lTypeNode) {
      lChoosenNode = lContext.getSystem().getRandomizer().rollInteger(0, lTree.size()-1);
    }
  }
  Primitive::Handle lOriginalPrimitive = lTree[lChoosenNode].mPrimitive;

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationSwapOp",
    std::string("Swap mutating the ")+uint2ordinal(lChoosenNode+1)+
    std::string(" node (primitive: \"")+lOriginalPrimitive->getName()+
    std::string("\" nb args: ")+uint2str(lOriginalPrimitive->getNumberArguments())+
    std::string(") of the ")+uint2ordinal(lChoosenTree+1)+std::string(" tree")
  ); 
  
  PrimitiveSet& lPrimitiveSet = *lContext.getSystem().getPrimitiveSuperSet()[lChoosenTree];
  unsigned int lNbArgsPrimit = lTree[lChoosenNode].mPrimitive->getNumberArguments();

  Primitive::Handle lChoosenPrimitive =
    lPrimitiveSet.selectPrimitiveWithKArguments(lNbArgsPrimit, lContext.getSystem());
  lTree[lChoosenNode].mPrimitive = lChoosenPrimitive->giveReference(lContext);
 
  lContext.setGenotypeHandle(lOldTreeHandle);
  lContext.setGenotypeIndex(lOldTreeIndex);

  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "mutation", "Beagle::GP::MutationSwapOp",
    std::string("Individual after GP swap mutation: ")+
    ioIndividual.serialize()
  ); 
  
  return true;
}


/*!
 *  \brief Read a mutation operator from XML tree.
 *  \param inNode XML node to use to read crossover operator.
 *  \param inOpMap Operator map to use to read crossover operator.
 */
void GP::MutationSwapOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string mMutationPbReadName = inNode->getTagAttribute("mutationpb");
  if(mMutationPbReadName.empty() == false) mMutationPbName = mMutationPbReadName;
  std::string mDistribPbReadName = inNode->getTagAttribute("distrpb");
  if(mDistribPbReadName.empty() == false) mDistribPbName = mDistribPbReadName;
}


/*!
 *  \brief Write mutation operator into XML streamer.
 *  \param ioStreamer XML streamer to write mutation operator into.
 */
void GP::MutationSwapOp::writeContent(XMLStreamer& ioStreamer) const
{
  Beagle::MutationOp::writeContent(ioStreamer);
  ioStreamer.insertAttribute("distrpb", mDistribPbName);
}

