/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/Primitive.cpp
 *  \brief  Implementation of the type GP::Primitive.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/08/24 19:23:48 $
 */

#include "beagle/GP.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a primitive.
 *  \param inNumberArguments Number of arguments of the primitive.
 *  \param inName Name of the primitive.
 */
GP::Primitive::Primitive(unsigned int inNumberArguments, const char* inName) :
  mName(inName),
  mNumberArguments(inNumberArguments)
{ }


/*!
 *  \brief Construct a primitive.
 *  \param inNumberArguments Number of arguments of the primitive.
 *  \param inName Name of the primitive.
 */
GP::Primitive::Primitive(unsigned int inNumberArguments, std::string inName) :
  mName(inName),
  mNumberArguments(inNumberArguments)
{ }


/*!
 *  \brief  Return the nth argument requested return type.
 *  \param  inN Index of the argument to get the type.
 *  \return Char string with the argument's node type.
 *  \throw  Beagle::ObjectException If the method is not overdefined is a subclass.
 */
std::string GP::Primitive::getArgType(unsigned int inN) const
{
  throw Beagle_UndefinedMethodInternalExceptionM("getArgType","GP::Primitive");
  return "";
}


/*!
 *  \brief  Return the children's node index.
 *  \param  inN Index of the children as arguments (index 0 is first arguments).
 *  \param  ioContext Actual evolutionary context.
 *  \return Index of the given nth children.
 *  \throw  Beagle::AssertException If the children index is out-of-bound.
 */
unsigned int GP::Primitive::getChildrenNodeIndex(unsigned int inN, GP::Context& ioContext) const
{
  Beagle_BoundCheckAssertM(inN, 0, mNumberArguments-1);
  GP::Tree& lActualTree = ioContext.getGenotype();
  unsigned int lNodeIndex = ioContext.getCallStackTop() + 1;
  for(unsigned int i=0; i<inN; i++) lNodeIndex += lActualTree[lNodeIndex].mSubTreeSize;
  return lNodeIndex;
}


/*!
 *  \brief  Return the current primitive node return type.
 *  \return Char the primitive's returning type (a char).
 *  \throw  Beagle::ObjectException If the method is not overdefined is a subclass.
 */
std::string GP::Primitive::getReturnType() const
{
  throw Beagle_UndefinedMethodInternalExceptionM("getReturnType","GP::Primitive");
  return "";
}


/*!
 *  \brief Get the primitive value.
 *  \param outValue Object in which the actual primitive value is written.
 *  \throw Beagle::ObjectException If the method is not overdefined is a subclass.
 */
void GP::Primitive::getValue(Object& outValue)
{
  throw Beagle_UndefinedMethodInternalExceptionM("getValue","GP::Primitive");
}


/*!
 *  \brief Give a reference to the current primitive.
 *  \param ioContext Reference to the evolution context.
 *  \return Handle to the refered primitive.
 */
GP::Primitive::Handle GP::Primitive::giveReference(GP::Context& ioContext)
{
  return GP::Primitive::Handle(this);
}


/*!
 *  \brief Return whether primitive's getValue and setValue are effective.
 *  \return True if getValue/setValue method are effective, false if not.
 */
bool GP::Primitive::haveValue() const
{
  return false;
}


/*!
 *  \brief Initialize the primitive.
 *  \param ioSystem Evolutionary system used.
 */
void GP::Primitive::initialize(GP::System& ioSystem)
{ }


/*!
 *  \brief Compare equality of two primitives. Name, number of args. and attribute must be the same.
 *  \param inRightObj Primitive to compare to the actual one.
 *  \return True if primitive are identical, false if not.
 */
bool GP::Primitive::isEqual(const Object& inRightObj) const
{
  const GP::Primitive& lRightPrimitive = castObjectT<const GP::Primitive&>(inRightObj);
  if((mName != lRightPrimitive.mName) ||
     (mNumberArguments != lRightPrimitive.mNumberArguments)) return false;
  return true;
}


/*!
 *  \brief Read the primitive.
 *  \param inNode XML node used to read the primitive from.
 *  \throw IOException If primitive format is not respected.
 */
void GP::Primitive::read(XMLNode::Handle& inNode)
{
  if(inNode->getType() != XMLNode::eTag)
    throw Beagle_IOExceptionNodeM(*inNode, "tag expected!");

  if(inNode->getTagName() != mName) {
    std::ostringstream lOSS;
    lOSS << "tag <" << mName << "> expected, but ";
    lOSS << "got tag <" << inNode->getTagName() << "> instead!";
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str()); 
  }
}


/*!
 *  \brief Set the value of a primitive.
 *  \param inValue Value of the primitive.
 *  \throw Beagle::ObjectException If the method is not overdefined is a subclass.
 */
void GP::Primitive::setValue(const Object& inValue)
{
  throw Beagle_UndefinedMethodInternalExceptionM("setValue","GP::Primitive");
}


/*!
 *  \brief Validate the primitive connections in the tree when using contrained or strongly-typed GP.
 *  \param ioContext Evolutionary context.
 *  \return True if the tree topology is correct, false if not.
 *  \throw Beagle::AssertException If the context is in a bad state.
 */
bool GP::Primitive::validate(GP::Context& ioContext) const
{
  Beagle_AssertM(ioContext.getCallStackSize() != 0);
  if(ioContext.getCallStackTop() == 0) {
    PrimitiveSet& lPrimitSet =
      *ioContext.getSystem().getPrimitiveSuperSet()[ioContext.getGenotypeIndex()];
    return (getReturnType() == lPrimitSet.getRootType());
  }
  GP::Tree& lActualTree        = ioContext.getGenotype();
  unsigned int lParentIndex    = ioContext.getCallStackElement(ioContext.getCallStackSize()-2);
  unsigned int lChildIndex     = lParentIndex + 1;
  unsigned int lArgsIndexChild = 0;
  while(lChildIndex != ioContext.getCallStackTop()) {
    lChildIndex += lActualTree[lChildIndex].mSubTreeSize;
    Beagle_AssertM(lChildIndex < lActualTree.size());
    lArgsIndexChild++;
    Beagle_AssertM(lArgsIndexChild < lActualTree[lParentIndex].mPrimitive->getNumberArguments());
  }
  std::string lTyping = lActualTree[lParentIndex].mPrimitive->getArgType(lArgsIndexChild);
  return (getReturnType() == lTyping);
}


/*!
 *  \brief Write the primitive.
 *  \param ioStreamer XML stream to write the primitive into.
 */
void GP::Primitive::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag(mName);
  writeContent(ioStreamer);
  ioStreamer.closeTag(false);
}


/*!
 *  \brief Write the primitive content.
 *  \param ioStreamer XML stream to write the primitive into.
 */
void GP::Primitive::writeContent(XMLStreamer& ioStreamer) const
{ }
