/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/System.cpp
 *  \brief  Implementation of the type GP::System.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:20 $
 */

#include "beagle/GP.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a new GP system.
 */
GP::System::System() :
  Beagle::System(new GP::Context::Alloc),
  mPrimitiveSuperSet(new GP::PrimitiveSuperSet)
{ }


/*!
 *  \brief Construct a new GP system.
 *  \param inSuperSet Handle to the primitive super set to use.
 */
GP::System::System(GP::PrimitiveSuperSet::Handle inSuperSet) :
  Beagle::System(new GP::Context::Alloc),
  mPrimitiveSuperSet(inSuperSet)
{ }


/*!
 *  \brief Construct a new GP system.
 *  \param inSuperSet Handle to the primitive super set to use.
 *  \param inContextAllocator Context allocator to use.
 */
GP::System::System(GP::PrimitiveSuperSet::Handle inSuperSet,
                   Beagle::Context::Alloc::Handle inContextAllocator) :
  Beagle::System(inContextAllocator),
  mPrimitiveSuperSet(inSuperSet)
{ }


/*!
 *  \brief Construct a new GP system.
 *  \param inSuperSet Handle to the primitive super set to use.
 *  \param inContextAllocator Context allocator to use.
 *  \param inRandomizer Handle to the randomizer used.
 *  \param inRegister Handle to the register used.
 *  \param inLogger Handle to the logger used.
 */
GP::System::System(GP::PrimitiveSuperSet::Handle inSuperSet,
                   Beagle::Context::Alloc::Handle inContextAllocator,
                   Randomizer::Handle inRandomizer,
                   Register::Handle inRegister,
                   Logger::Handle     inLogger) :
  Beagle::System(inContextAllocator, inRandomizer, inRegister, inLogger),
  mPrimitiveSuperSet(inSuperSet)
{ }


/*!
 *  \brief Construct a new GP system.
 *  \param inSet Handle to the primitive set to use.
 */
GP::System::System(GP::PrimitiveSet::Handle inSet) :
  Beagle::System(new GP::Context::Alloc),
  mPrimitiveSuperSet(new PrimitiveSuperSet)
{
  mPrimitiveSuperSet->insert(inSet);
}


/*!
 *  \brief Construct a new GP system.
 *  \param inSet Handle to the primitive set to use.
 *  \param inContextAllocator Context allocator to use.
 */
GP::System::System(GP::PrimitiveSet::Handle inSet,
                   Beagle::Context::Alloc::Handle inContextAllocator) :
  Beagle::System(inContextAllocator),
  mPrimitiveSuperSet(new PrimitiveSuperSet)
{
  mPrimitiveSuperSet->insert(inSet);
}


/*!
 *  \brief Construct a new GP system.
 *  \param inSet Handle to the primitive set to use.
 *  \param inContextAllocator Context allocator to use.
 *  \param inRandomizer Handle to the randomizer used.
 *  \param inRegister Handle to the register used.
 *  \param inLogger Handle to the logger used.
 */
GP::System::System(GP::PrimitiveSet::Handle inSet,
                   Beagle::Context::Alloc::Handle inContextAllocator,
                   Randomizer::Handle inRandomizer,
                   Register::Handle inRegister,
                   Logger::Handle     inLogger) :
  Beagle::System(inContextAllocator, inRandomizer, inRegister, inLogger),
  mPrimitiveSuperSet(new PrimitiveSuperSet)
{
  mPrimitiveSuperSet->insert(inSet);
}


/*!
 *  \brief Initialize the system.
 *  \param ioArgc Number of elements on the command-line.
 *  \param ioArgv Element on the command-line.
 */
void GP::System::initialize(int& ioArgc, char** ioArgv)
{
  // Initializing the system.
  Beagle_LogDetailedM(
    getLogger(),
    "system", "Beagle::System",
    "Initializing the system"
  );
  mLogger->initialize(*this);
  mRandomizer->initialize(*this);
  mRegister->initialize(*this);
  mPrimitiveSuperSet->initialize(*this);

  // Parsing the command-line.
  Beagle_LogTraceM(
    (*mLogger),
    "system", "Beagle::System",
    "Parsing the command-line"
  );
  mRegister->parseCommandLine(*this, ioArgc, ioArgv);
}


/*!
 *  \brief Initialize the system.
 *  \param inRegisterFileName Name of the filename containing the system configuration.
 */
void GP::System::initialize(std::string inRegisterFileName)
{
  // Initializing the system.
  Beagle_LogDetailedM(
    getLogger(),
    "system", "Beagle::System",
    "Initializing the system"
  );
  mLogger->initialize(*this);
  mRandomizer->initialize(*this);
  mRegister->initialize(*this);
  mPrimitiveSuperSet->initialize(*this);

  // Reading configuration file.
  if(inRegisterFileName.empty() == false) {
    Beagle_LogTraceM(
      (*mLogger),
      "system", "Beagle::System",
      "Reading system configuration file"
    );
    mRegister->readParametersFile(inRegisterFileName, *this);
  }
}


