/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/GP/src/Tree.cpp
 *  \brief  Implementation of the type GP::Tree.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/07/01 23:59:00 $
 */

#include "beagle/GP.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a GP node.
 *  \param inPrimitive Handle to the primitive refered by the node.
 *  \param inSubTreeSize Sub-tree size, including actual node.
 */
GP::Node::Node(GP::Primitive::Handle inPrimitive, unsigned int inSubTreeSize) :
  mPrimitive(inPrimitive),
  mSubTreeSize(inSubTreeSize)
{ }


/*!
 *  \brief Compare equality of two GP tree's nodes.
 *  \param inRightNode Right node to be compare to.
 *  \return True if nodes are identical, false if not.
 */
bool GP::Node::operator==(const GP::Node& inRightNode) const
{
  return (mPrimitive == inRightNode.mPrimitive) && (mSubTreeSize == inRightNode.mSubTreeSize);
}


/*!
 *  \brief Construct a GP tree of the size given.
 *  \param inSize Size of the tree.
 */
GP::Tree::Tree(unsigned int inSize) :
  std::vector<Node>(inSize)
{ }


/*!
 *  \brief Return number of nodes of GP tree.
 *  \return Number of nodes of GP tree.
 */
unsigned int GP::Tree::getSize() const
{
  return size();
}


/*!
 *  \brief Get the maximum depth of a sub-tree.
 *  \param inNodeIndex Index of the node root to the sub-tree.
 *  \return Depth to the given sub-tree.
 *  \throw Beagle::AssertException If the node index given is to out-of-bound.
 */
unsigned int GP::Tree::getTreeDepth(unsigned int inNodeIndex) const
{
  Beagle_BoundCheckAssertM(inNodeIndex, 0, size()-1);
  unsigned int lDepth = 1;
  unsigned int lChildNodeIndex = inNodeIndex + 1;
  for(unsigned int i=0; i<(*this)[inNodeIndex].mPrimitive->getNumberArguments(); i++) {
    unsigned int lChildDepth = getTreeDepth(lChildNodeIndex);
    lDepth = maxOf<unsigned int>(lDepth, lChildDepth+1);
    lChildNodeIndex += (*this)[lChildNodeIndex].mSubTreeSize;
  }
  return lDepth;
}


/*!
 *  \brief Interpret the GP tree.
 *  \param outResult Datum containing the result of the interpretation.
 *  \param ioContext GP evolutionary context.
 *  \throw Beagle::ObjectException When tree is empty or not in contextual individual.
 *  \throw Beagle::AssertException When the contextual individual is a NULL pointer.
 */
void GP::Tree::interpret(GP::Datum& outResult, GP::Context& ioContext)
{
  if(empty()) throw Beagle_ObjectExceptionM("Could not interpret, tree is empty!");
  GP::Individual::Handle lIndiv = ioContext.getIndividualHandle();
  Beagle_NonNullPointerAssertM(lIndiv.getPointer());
  unsigned int lTreeIndex = 0;
  for(; lTreeIndex < lIndiv->size(); lTreeIndex++) {
    if(this == (*lIndiv)[lTreeIndex].getPointer()) break;
  }
  if(lTreeIndex == lIndiv->size())
    throw Beagle_ObjectExceptionM("Interpreted tree is not in the actual individual of the context!");
    
  Tree::Handle lOldTreeHandle = ioContext.getGenotypeHandle();
  unsigned int lOldTreeIndex  = ioContext.getGenotypeIndex();
  ioContext.setGenotypeIndex(lTreeIndex);
  ioContext.setGenotypeHandle(Handle(this));
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "gptree", "Beagle::GP::Tree",
    std::string("Interpreting the ")+uint2ordinal(lTreeIndex+1)+
    std::string(" tree of the ")+uint2ordinal(ioContext.getIndividualIndex()+1)+
    std::string(" individual")
  );
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "gptree", "Beagle::GP::Tree",
    std::string("The individual is: ")+ioContext.getIndividual().serialize()
  );  
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "gptree", "Beagle::GP::Tree",
    std::string("Executing the tree root node \"")+
    (*this)[0].mPrimitive->getName()+"\""
  );
  ioContext.pushCallStack(0);  
  (*this)[0].mPrimitive->execute(outResult, ioContext);
  ioContext.popCallStack();
  Beagle_LogDebugM(
    ioContext.getSystem().getLogger(),
    "gptree", "Beagle::GP::Tree",
    std::string("Result of execution: ")+outResult.serialize()
  );
  ioContext.setGenotypeIndex(lOldTreeIndex);
  ioContext.setGenotypeHandle(lOldTreeHandle);
}


/*!
 *  \brief Compare the equality of two GP trees.
 *  \param inRightObj Right tree to be compare to tha actual one.
 *  \return True if the trees are identical, false if not.
 */
bool GP::Tree::isEqual(const Object& inRightObj) const
{
  const GP::Tree& lRightTree = castObjectT<const GP::Tree&>(inRightObj);
  if(size() != lRightTree.size()) return false;
  return std::equal(begin(), end(), lRightTree.begin());
}


/*!
 *  \brief Read a GP tree from an XML node.
 *  \param inNode XML node to use to read the tree from.
 *  \param ioContext Context to use to read the tree.
 */
void GP::Tree::readWithContext(XMLNode::Handle& inNode, Beagle::Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Genotype"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Genotype> expected!");

  std::string lType = inNode->getTagAttribute("type");
  if(lType.empty())
    throw Beagle_IOExceptionNodeM(*inNode, "GP tree type of the genotype is not present!");
  if(lType != "gptree")
    throw Beagle_IOExceptionNodeM(*inNode, "type of genotype mismatch, expected \"gptree\"!");
  
  XMLNode::Handle lChild = inNode->getFirstChild();
  if((lChild==NULL) || (lChild->getType()!=XMLNode::eTag))
    throw Beagle_IOExceptionNodeM(*lChild, "expected a XML tag for the GP tree!");

  GP::Context& lGPContext = castObjectT<GP::Context&>(ioContext);
  clear();
  readSubTree(lChild, lGPContext);
}


/*!
 *  \brief Read a GP subtree from a XML node.
 *  \param inNode XMLNode to read GP tree from.
 *  \param ioContext Context to use to map the node to the appropriate primitive.
 *  \return Read subtree size.
 *  \throw Beagle::InternalException When the the tree format is incorrect.
 */
unsigned int GP::Tree::readSubTree(XMLNode::Handle& inNode, GP::Context& ioContext)
{
  if(inNode->getType() != XMLNode::eTag)
    throw Beagle_IOExceptionNodeM(*inNode, "tag expected!");

  GP::PrimitiveSet::Handle lPrimitiveSet =
    ioContext.getSystem().getPrimitiveSuperSet()[ioContext.getGenotypeIndex()];
  GP::Primitive::Handle lPrimitive = lPrimitiveSet->getPrimitiveByName(inNode->getTagName());
  if(lPrimitive == NULL) {
    std::ostringstream lOSS;
    lOSS << "no primitive named \"" <<  inNode->getTagName();
    lOSS << "\" found in the primitive set";
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  unsigned int lNodeIdx = size();
  push_back(Node(lPrimitive->giveReference(ioContext), 0));
  (*this)[lNodeIdx].mPrimitive->read(inNode);
  
  unsigned int lSubTreeSize = 1;
  unsigned int lNbrGPChild = 0;
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      lSubTreeSize += readSubTree(lChild, ioContext);
      ++lNbrGPChild;
    }
  }  
  if(lNbrGPChild != (*this)[lNodeIdx].mPrimitive->getNumberArguments()) {
    std::ostringstream lOSS;
    lOSS << "number of arguments stated by the GP primitives (";
    lOSS << (*this)[lNodeIdx].mPrimitive->getNumberArguments();
    lOSS << ") mismatch the number of arguments read for the GP node (";
    lOSS << lNbrGPChild << ")!";
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  (*this)[lNodeIdx].mSubTreeSize = lSubTreeSize;  
  return lSubTreeSize;
}


/*!
 *  \brief Set the context to the node given.
 *  \param inNodeIndex Index of the node to use to set the context.
 *  \param ioContext Evolutionary context to set to the given node.
 *  \throw Beagle::AssertException If the tree is in a bad state.
 */
void GP::Tree::setContextToNode(unsigned int inNodeIndex, GP::Context& ioContext) const
{
  Beagle_AssertM(inNodeIndex < size());
  ioContext.emptyCallStack();
  unsigned int lIndex = 0;
  ioContext.pushCallStack(lIndex);
  while(lIndex < inNodeIndex) {
    unsigned int lChildIndex = lIndex + 1;
    unsigned int lArgument=0;
    for(; lArgument<(*this)[lIndex].mPrimitive->getNumberArguments(); lArgument++) {
      if((lChildIndex+(*this)[lChildIndex].mSubTreeSize) > inNodeIndex) break;
      lChildIndex += (*this)[lChildIndex].mSubTreeSize;
    }
    Beagle_AssertM(lArgument < (*this)[lIndex].mPrimitive->getNumberArguments());
    Beagle_AssertM(lChildIndex < size());
    lIndex = lChildIndex;
    ioContext.pushCallStack(lIndex);
  }
  Beagle_AssertM(lIndex == inNodeIndex);
}


/*!
 *  \brief Validate the topology of a sub-tree the a given node.
 *  \param inNodeIndex Index in the tree of the root node of the sub-tree to validate.
 *  \param ioContext Evolutionary context.
 *  \throw Beagle::AssertException When the evolutionary context is badly configured.
 */
bool GP::Tree::validateSubTree(unsigned int inNodeIndex, GP::Context& ioContext) const
{
  Beagle_AssertM(inNodeIndex < size());
  Beagle_AssertM(inNodeIndex == ioContext.getCallStackTop());
  if((*this)[inNodeIndex].mPrimitive->validate(ioContext) == false) return false;
  unsigned int lChildNodeIndex = inNodeIndex + 1;
  for(unsigned int i=0; i<(*this)[inNodeIndex].mPrimitive->getNumberArguments(); i++) {
    ioContext.pushCallStack(lChildNodeIndex);
    bool lChildValidation = validateSubTree(lChildNodeIndex, ioContext);
    ioContext.popCallStack();
    if(lChildValidation == false) return false;
    lChildNodeIndex += (*this)[lChildNodeIndex].mSubTreeSize;
  }
  return true;
}


/*!
 *  \brief Write a GP tree into a XML streamer.
 *  \param ioStreamer XML streamer to write the tree into.
 */
void GP::Tree::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Genotype");
  ioStreamer.insertAttribute("type", "gptree");
  ioStreamer.insertAttribute("size", uint2str(size()));
  ioStreamer.insertAttribute("depth", uint2str(getTreeDepth()));
  unsigned int lSizeSubTree = writeSubTree(ioStreamer, 0, false);
  Beagle_AssertM(lSizeSubTree == size());
  ioStreamer.closeTag(false);
}


/*!
 *  \brief Write a GP tree into a XML streamer with indentation.
 *  \param ioStreamer XML streamer to write the tree into.
 */
void GP::Tree::writeIndented(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Genotype");
  ioStreamer.insertAttribute("type", "gptree");
  ioStreamer.insertAttribute("size", uint2str(size()));
  ioStreamer.insertAttribute("depth", uint2str(getTreeDepth()));
  unsigned int lSizeSubTree = writeSubTree(ioStreamer, 0, true);
  Beagle_AssertM(lSizeSubTree == size());
  ioStreamer.closeTag(true);
}


/*!
 *  \brief Write a GP sub-tree into a XML streamer.
 *  \param ioStreamer XML streamer to write the tree into.
 *  \param inN Index of the actual subtree root node in the vector.
 *  \param inIndent True: indent sub-tree writing.
 *  \throw Beagle::AssertException When sub-tree size doesn't match.
 */
unsigned int GP::Tree::writeSubTree(XMLStreamer& ioStreamer,
                                    unsigned int inN,
                                    bool inIndent) const {
  Beagle_AssertM(inN < size());
  unsigned int lNumberArguments = (*this)[inN].mPrimitive->getNumberArguments();
  ioStreamer.openTag((*this)[inN].mPrimitive->getName(), inIndent);
  (*this)[inN].mPrimitive->writeContent(ioStreamer);
  unsigned int lSubTreeSize = 1;
  for(unsigned int i=0; i<lNumberArguments; ++i) {
    lSubTreeSize += writeSubTree(ioStreamer, (lSubTreeSize+inN), inIndent);
  }
  Beagle_AssertM(lSubTreeSize == (*this)[inN].mSubTreeSize);
  ioStreamer.closeTag(inIndent);
  return lSubTreeSize;
}


