/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/BreederNode.cpp
 *  \brief  Source code of class BreederNode.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct breeder node.
 *  \param inBreederOp Breeder operator associated to node.
 *  \param inFirstChild First breeder node child to actual node.
 *  \param inNextSibbling Next sibbling breeder node to actual node.
 */
BreederNode::BreederNode(PointerT<BreederOp,Operator::Handle> inBreederOp,
                         PointerT<BreederNode,Object::Handle> inFirstChild,
                         PointerT<BreederNode,Object::Handle> inNextSibbling) :
  mBreederOp(inBreederOp),
  mChild(inFirstChild),
  mSibbling(inNextSibbling)
{ }
                         

/*!
 *  \brief Initialize breeder node structure.
 *  \param ioSystem Evolutionary system to use to initialize the system.
 */
void BreederNode::initialize(System& ioSystem)
{
  if((mBreederOp != NULL) && (mBreederOp->isInitialized() == false)) {
    Beagle_LogTraceM(
      ioSystem.getLogger(),
      "breeder", "Beagle::BreederNode",
      std::string("Initializing operator \"") + mBreederOp->getName() + "\""
    );
    mBreederOp->initialize(ioSystem);
    mBreederOp->setInitialized();
  }
  if(mChild != NULL) mChild->initialize(ioSystem);
  if(mSibbling != NULL) mSibbling->initialize(ioSystem);
}


/*!
 *  \brief Call post initialize hooks of breeder node structure.
 *  \param ioSystem Evolutionary system to use to post-init the system.
 */
void BreederNode::postInit(System& ioSystem)
{
  if((mBreederOp != NULL) && (mBreederOp->isPostInitialized() == false)) {
    Beagle_LogTraceM(
      ioSystem.getLogger(),
      "breeder", "Beagle::BreederNode",
      std::string("Calling post-initialization hook of operator \"") +
      mBreederOp->getName() + "\""
    );
    mBreederOp->postInit(ioSystem);
    mBreederOp->setPostInitialized();
  }
  if(mChild != NULL) mChild->postInit(ioSystem);
  if(mSibbling != NULL) mSibbling->postInit(ioSystem);
}


/*!
 *  \brief Reading a breeder node with method read is undefined. Use readWithMap instead.
 */
void BreederNode::read(XMLNode::Handle&)
{
  throw Beagle_UndefinedMethodInternalExceptionM("read","BreederNode");
}


/*!
 *  \brief Read a breeder tree for XML tree.
 *  \param inNode XML node to use to read breeder node.
 *  \param inOpMap Operator map to use to read breeder node.
 */
void BreederNode::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  setBreederOp(NULL);
  setFirstChild(NULL);
  setNextSibbling(NULL);
  
  if(inNode->getType() == XMLNode::eTag) {
    OperatorMap::iterator lIterOp = inOpMap.find(inNode->getTagName());
    if(lIterOp == inOpMap.end()) {
      std::ostringstream lOSS;
      lOSS << "Operator \"" << inNode->getTagName() << "\" is not a breeder operator!";
      throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
    }
    try {
      mBreederOp = castHandleT<BreederOp>(lIterOp->second);
    }
    catch(BadCastException&) {
      std::ostringstream lOSS;
      lOSS << "Operator \"" << inNode->getTagName() << "\" is not a breeder operator!";
      throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
    }
    mBreederOp = castHandleT<BreederOp>(mBreederOp->giveReference());
    mBreederOp->readWithMap(inNode, inOpMap);
  }
  
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      OperatorMap::iterator lIterOp = inOpMap.find(lChild->getTagName());
      if(lIterOp == inOpMap.end()) continue;
      
      PointerT<BreederOp,Operator::Handle> lOp;
      try {
        lOp = castHandleT<BreederOp>(lIterOp->second);
      }
      catch(BadCastException&) {
        std::ostringstream lOSS;
        lOSS << "Operator \"" << lChild->getTagName() << "\" is not a breeder operator!";
        throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
      }
      lOp = castHandleT<BreederOp>(lOp->giveReference());
      if(getFirstChild() == NULL) {
        setFirstChild(new BreederNode(lOp));
        getFirstChild()->readWithMap(lChild, inOpMap);
      }
      else {
        PointerT<BreederNode,Object::Handle> lBreederChild;
        for(lBreederChild = getFirstChild(); lBreederChild->getNextSibbling() != NULL;
            lBreederChild = lBreederChild->getNextSibbling());
        lBreederChild->setNextSibbling(new BreederNode(lOp));
        lBreederChild->getNextSibbling()->readWithMap(lChild, inOpMap);
      }
    }
  }
}


/*!
 *  \brief Write breeder operator into XML streamer.
 *  \param ioStreamer XML streamer to write breeder operator into.
 */
void BreederNode::write(XMLStreamer& ioStreamer) const
{
  if(mBreederOp != NULL) {
    ioStreamer.openTag(mBreederOp->getName());
    mBreederOp->writeContent(ioStreamer);
  }
  for(BreederNode::Handle lChild=getFirstChild(); lChild != NULL;
      lChild=lChild->getNextSibbling()) {
    lChild->write(ioStreamer);
  }
  if(mBreederOp != NULL) ioStreamer.closeTag();
}

