/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/ConfigurationDumper.cpp
 *  \brief  Source code of class ConfigurationDumper.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/07/01 23:59:01 $
 */

#include "beagle/Beagle.hpp"

#include <cstdio>
#include <cstring>
#include <algorithm>
#include <fstream>

using namespace Beagle;


/*!
 *  \brief Construct a configuration file dumper.
 *  \param inSystem Reference to the associated system.
 *  \param inEvolver Reference to the associated evolver.
 *  \param inName File name value.
 */
ConfigurationDumper::ConfigurationDumper(System& inSystem, Evolver& inEvolver, std::string inName) :
  String(inName),
  mSystem(inSystem),
  mEvolver(inEvolver)
{ }


/*!
 *  \brief Read the file name and then dump the configuration in it.
 *  \param inNode XML node to read the file name from.
 */
void ConfigurationDumper::read(XMLNode::Handle& inNode)
{
  if(inNode == NULL) return;
  if(inNode->getType() != XMLNode::eString)
    throw Beagle_IOExceptionNodeM(*inNode, "not a string with a filename");
  getWrappedValue() = inNode->getValue();
  if(getWrappedValue().empty()) return;

  std::string lFilenameBak = getWrappedValue() + "~";
  std::remove(lFilenameBak.c_str());
  std::rename(getWrappedValue().c_str(), lFilenameBak.c_str());
  mSystem.getRegister().deleteEntry("ec.conf.dump");
  mSystem.getRegister().deleteEntry("ec.conf.file");
  std::ofstream lOFStream(getWrappedValue().c_str());
  std::cerr << "A typical configuration file named \"";
  std::cerr << getWrappedValue() << "\" is created." << std::endl << std::flush;
  XMLStreamer lStreamer(lOFStream);
  lStreamer.insertHeader();
  lStreamer.openTag("Beagle");
  lStreamer.insertAttribute("version", BEAGLE_VERSION);
  lStreamer.insertSpecialTag("!-- Created by a configuration dump --");
  lStreamer.insertSpecialTag("!-- Evolver: configuration of the algorithm --");
  mEvolver.write(lStreamer);
  lStreamer.insertSpecialTag("!-- Register: parameters of the evolution --");
  mSystem.getRegister().write(lStreamer);
  lStreamer.closeTag();
  lOFStream << std::flush;
  lOFStream.close();
  std::cerr << "Exiting..." << std::endl << std::flush;
  exit(0);
}

