/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/CrossoverOp.cpp
 *  \brief  Source code of class CrossoverOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <vector>
#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a plain new crossover operator.
 *  \param inMatingPbName Mating probability parameter name used in register.
 *  \param inName Name of the crossover operator.
 */
CrossoverOp::CrossoverOp(std::string inMatingPbName, std::string inName) :
  BreederOp(inName),
  mMatingProbaName(inMatingPbName)
{ }


/*!
 *  \brief Apply the crossover operation on a breeding pool, returning a mated individual.
 *  \param inBreedingPool Breeding pool to use for the crossover operation.
 *  \param inChild Node handle associated to child node in the breeder tree.
 *  \param ioContext Evolutionary context of the crossover operation.
 *  \return Mated individual.
 */
Individual::Handle CrossoverOp::breed(Individual::Bag& inBreedingPool,
                                      BreederNode::Handle inChild,
                                      Context& ioContext)
{
  Context::Handle lContext2 =
    castObjectT<Context*>(ioContext.getSystem().getContextAllocator().clone(ioContext));
  
  Beagle_NonNullPointerAssertM(inChild);
  Beagle_NonNullPointerAssertM(inChild->getBreederOp());
  Individual::Handle lIndiv1 = inChild->getBreederOp()->breed(inBreedingPool,
                                                              inChild->getFirstChild(), 
                                                              ioContext);

  Beagle_NonNullPointerAssertM(inChild->getNextSibbling());
  Beagle_NonNullPointerAssertM(inChild->getNextSibbling()->getBreederOp());                                                            
  Individual::Handle lIndiv2 =
    inChild->getNextSibbling()->getBreederOp()->breed(inBreedingPool,
                                                      inChild->getNextSibbling()->getFirstChild(),
                                                      *lContext2);
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "crossover", "Beagle::CrossoverOp",
    std::string("Mating the ")+uint2ordinal(ioContext.getIndividualIndex()+1)+
    std::string(" individual with the ")+uint2ordinal(lContext2->getIndividualIndex()+1)+
    " individual"
  );

  if((lIndiv1 != NULL) && (lIndiv2 != NULL)) {
    bool lMated = mate(*lIndiv1, ioContext, *lIndiv2, *lContext2);
    if(lMated && (lIndiv1->getFitness() != NULL)) lIndiv1->getFitness()->setInvalid();
  }
  
  return lIndiv1;
}


/*!
 *  \return Return selection probability of breeder operator.
 *  \param inChild Child node in the breeder tree.
 */
float CrossoverOp::getBreedingProba(BreederNode::Handle inChild)
{
  return mMatingProba->getWrappedValue();
}


/*!
 *  \brief Initialize the crossover operator.
 *  \param ioSystem Reference to the system to use for the initialization.
 */
void CrossoverOp::initialize(System& ioSystem)
{
  BreederOp::initialize(ioSystem);
  
  if(ioSystem.getRegister().isRegistered(mMatingProbaName)) {
    mMatingProba = castHandleT<Float>(ioSystem.getRegister().getEntry(mMatingProbaName));
  } else {
    mMatingProba = new Float(0.5);
    Register::Description lDescription(
      "Individual crossover probability",
      "Float",
      "0.5",
      "Single individual crossover probability for a generation."
    );
    ioSystem.getRegister().addEntry(mMatingProbaName, mMatingProba, lDescription);
  }
}


/*!
 *  \brief Apply the crossover operation on the deme.
 *  \param ioDeme Current deme of individuals to mate.
 *  \param ioContext Context of the evolution.
 */
void CrossoverOp::operate(Deme& ioDeme, Context& ioContext)
{
  Beagle_ValidateParameterM(mMatingProba->getWrappedValue()>=0.0, mMatingProbaName, "<0");
  
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "crossover", "Beagle::CrossoverOp",
    std::string("Mating individuals of the ")+
    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
  );
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "crossover", "Beagle::CrossoverOp",
    std::string("Mating individuals with probability ")+
    dbl2str(mMatingProba->getWrappedValue())
  );
  
  Context::Handle lContext2 =
    castObjectT<Context*>(ioContext.getSystem().getContextAllocator().clone(ioContext));
  Individual::Handle lOldIndividualHandle = ioContext.getIndividualHandle();
  unsigned int lOldIndividualIndex = ioContext.getIndividualIndex();

  std::vector<unsigned int> lMateVector;
  for(unsigned int i=0; i<ioDeme.size(); i++) {
    if(ioContext.getSystem().getRandomizer().rollUniform() <= mMatingProba->getWrappedValue()) {
      lMateVector.push_back(i);
    }
  }

  std::random_shuffle(lMateVector.begin(), lMateVector.end(),
                      ioContext.getSystem().getRandomizer());
  if((lMateVector.size() % 2) != 0) lMateVector.pop_back();
               
  for(unsigned int j=0; j<lMateVector.size(); j+=2) {
    unsigned int lFirstMate  = lMateVector[j];
    unsigned int lSecondMate = lMateVector[j+1];
    ioContext.setIndividualIndex(lFirstMate);
    ioContext.setIndividualHandle(ioDeme[lFirstMate]);
    lContext2->setIndividualIndex(lSecondMate);
    lContext2->setIndividualHandle(ioDeme[lSecondMate]);
    
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "crossover", "Beagle::CrossoverOp",
      std::string("Mating the ")+uint2ordinal(lFirstMate+1)+
      std::string(" individual with the ")+uint2ordinal(lSecondMate+1)+" individual"
    );

    bool lMated = mate(*ioDeme[lFirstMate], ioContext, *ioDeme[lSecondMate], *lContext2);
    if(lMated && (ioDeme[lFirstMate]->getFitness() != NULL))
      ioDeme[lFirstMate]->getFitness()->setInvalid();
    if(lMated && (ioDeme[lSecondMate]->getFitness() != NULL))
      ioDeme[lSecondMate]->getFitness()->setInvalid();
  }
  
  ioContext.setIndividualIndex(lOldIndividualIndex);
  ioContext.setIndividualHandle(lOldIndividualHandle);
}


/*!
 *  \brief Read a crossover operator from XML tree.
 *  \param inNode XML node to use to read crossover operator.
 *  \param inOpMap Operator map to use to read crossover operator.
 */
void CrossoverOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string mMatingProbaReadName = inNode->getTagAttribute("matingpb");
  if(mMatingProbaReadName.empty() == false) mMatingProbaName = mMatingProbaReadName;
}


/*!
 *  \brief Write crossover operator content into XML streamer.
 *  \param ioStreamer XML streamer to write crossover operator into.
 */
void CrossoverOp::writeContent(XMLStreamer& ioStreamer) const
{
  ioStreamer.insertAttribute("matingpb", mMatingProbaName);
}


  
