/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Deme.cpp
 *  \brief  Source code of class Deme.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.10 $
 *  $Date: 2004/07/27 20:41:39 $
 */

#include "beagle/Beagle.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a deme for an individual allocator.
 *  \param inIndAlloc Handle to the allocator of individuals.
 */
Deme::Deme(Individual::Alloc::Handle inIndAlloc) :
   Individual::Bag(inIndAlloc),
   mHOFAlloc(new HallOfFame::Alloc),
   mHallOfFame(new HallOfFame(inIndAlloc)),
   mMigrationBuffer(new Individual::Bag(inIndAlloc)),
   mStatsAlloc(new Stats::Alloc),
   mStats(new Stats)
{ }

 
/*!
 *  \brief Construct a deme for an individual allocator and a statistics allocator.
 *  \param inIndAlloc Handle to the allocator of individuals.
 *  \param inStatsAlloc Statistics allocator.
 *  \param inHOFAlloc Hall-of-fame allocator.
 *  \param inN Size of deme.
 */
Deme::Deme(Individual::Alloc::Handle inIndAlloc,
           Stats::Alloc::Handle inStatsAlloc,
           HallOfFame::Alloc::Handle inHOFAlloc,
           size_type inN) :
   Individual::Bag(inIndAlloc, inN),
   mHOFAlloc(inHOFAlloc),
   mHallOfFame(castObjectT<HallOfFame*>(inHOFAlloc->allocate())),
   mMigrationBuffer(new Individual::Bag(inIndAlloc)),
   mStatsAlloc(inStatsAlloc),
   mStats(castObjectT<Stats*>(inStatsAlloc->allocate()))
{
  mHallOfFame->setIndivAlloc(inIndAlloc);
}

  
/*!
 *  \brief Construct a deme with a genotype allocator.
 *  \param inGenotypeAlloc Genotype allocator.
 */
Deme::Deme(Genotype::Alloc::Handle inGenotypeAlloc) :
  Individual::Bag(new Individual::Alloc(inGenotypeAlloc, new FitnessSimple::Alloc)),
  mHOFAlloc(new HallOfFame::Alloc),
  mHallOfFame(new HallOfFame(new Individual::Alloc(inGenotypeAlloc, new FitnessSimple::Alloc))),
  mMigrationBuffer(new Individual::Bag(new Individual::Alloc(inGenotypeAlloc,
                                                             new FitnessSimple::Alloc))),
  mStatsAlloc(new Stats::Alloc),
  mStats(new Stats)
{ }


/*!
 *  \brief Construct a deme with a genotype allocator and a fitness allocator.
 *  \param inGenotypeAlloc Genotype allocator.
 *  \param inFitnessAlloc Fitness allocator.
 */
Deme::Deme(Genotype::Alloc::Handle inGenotypeAlloc, Fitness::Alloc::Handle inFitnessAlloc) :
  Individual::Bag(new Individual::Alloc(inGenotypeAlloc, inFitnessAlloc)),
  mHOFAlloc(new HallOfFame::Alloc),
  mHallOfFame(new HallOfFame(new Individual::Alloc(inGenotypeAlloc, inFitnessAlloc))),
  mMigrationBuffer(new Individual::Bag(new Individual::Alloc(inGenotypeAlloc, inFitnessAlloc))),
  mStatsAlloc(new Stats::Alloc),
  mStats(new Stats)
{ }


/*!
 *  \brief Copy construct a deme.
 *  \param inOriginalDeme Original deme to use as original of the actual deme.
 */
Deme::Deme(const Deme& inOriginalDeme) :
  mHallOfFame(new HallOfFame(inOriginalDeme.mHallOfFame->getIndivAlloc()))
{
  (*this) = inOriginalDeme;
}


/*!
 *  \brief Copy a deme.
 *  \param inOriginalDeme Original deme to use as original of the actual deme.
 */
Deme& Deme::operator=(const Deme& inOriginalDeme)
{
  if(this == &inOriginalDeme) return *this;
  Individual::Bag::operator=(inOriginalDeme);
  mHOFAlloc = inOriginalDeme.mHOFAlloc;
  mHallOfFame = castObjectT<HallOfFame*>(mStatsAlloc->clone(*inOriginalDeme.mHallOfFame));
  mMigrationBuffer->copyData(*inOriginalDeme.mMigrationBuffer);
  mStatsAlloc = inOriginalDeme.mStatsAlloc;
  mStats = castObjectT<Stats*>(mStatsAlloc->clone(*inOriginalDeme.mStats));
  return *this;
}


/*!
 *  \brief Copy a deme and clone the individuals it haves.
 *  \param inOrigContainer Original deme to copy.
 *  \throw BadCastException If the deme received as argument have no type allocator.
 */
void Deme::copyData(const Container& inOrigContainer)
{
  const Deme& lOrigDeme = castObjectT<const Deme&>(inOrigContainer);
  if(lOrigDeme.getTypeAlloc() == NULL) {
    std::string lMessage = "The copyData() method must be call only with as argument a deme";
    lMessage += " that have a type allocator!";
    throw Beagle_InternalExceptionM(lMessage);
  }
  if(this == &lOrigDeme) return;
  (*this) = lOrigDeme;
  clear();
  for(size_type i=0; i<lOrigDeme.size(); ++i) {
    Beagle_NonNullPointerAssertM(lOrigDeme[i]);
    push_back(getTypeAlloc()->cloneData(*lOrigDeme[i]));
  }
}


/*!
 *  \brief Reading a deme with method read is undefined. Use readWithContext instead.
 */
void Deme::read(XMLNode::Handle&)
{
  throw Beagle_UndefinedMethodInternalExceptionM("read","Deme");
}


/*!
 *  \brief Read a deme from an XML node.
 *  \param inNode XML node to read the deme from.
 *  \param ioContext Evolutionary context to use to read the deme.
 *  \throw Beagle::IOException If the format is not respected.
 */
void Deme::readWithContext(XMLNode::Handle& inNode, Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Deme"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Deme> expected!");
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Stats") mStats->read(lChild);
      else if(lChild->getTagName() == "HallOfFame") mHallOfFame->readWithContext(lChild, ioContext);
      else if(lChild->getTagName() == "MigrationBuffer") readMigrationBuffer(lChild, ioContext);
      else if(lChild->getTagName() == "Population") readPopulation(lChild, ioContext);
    }
  }
}


/*!
 *  \brief Read a deme migration buffer from an XML node.
 *  \param inNode XML node to read the deme migration buffer from.
 *  \param ioContext Evolutionary context to use to read the deme migration buffer.
 *  \throw Beagle::IOException If the format is not respected.
 */
void Deme::readMigrationBuffer(XMLNode::Handle& inNode, Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "MigrationBuffer"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <MigrationBuffer> expected!");
  XMLNode::Handle lChild = inNode->getFirstChild();
  Individual::Bag::size_type lSize;
  for(lSize = 0; lChild != NULL; ++lSize) lChild = lChild->getNextSibling();
  if((mMigrationBuffer->getTypeAlloc()==NULL) && (lSize > mMigrationBuffer->size())) {
    std::ostringstream lMessageOSS;
    lMessageOSS << "Deme migration buffer size (" << lSize << ") is bigger than the actual size (";
    lMessageOSS << mMigrationBuffer->size() << "), and there is no type allocator for resizing!";
    throw Beagle_IOExceptionNodeM(*inNode, lMessageOSS.str());
  }
  mMigrationBuffer->resize(lSize);
  lChild = inNode->getFirstChild();
  unsigned int lOldIndivIndex = ioContext.getIndividualIndex();
  Individual::Handle lOldIndivHandle = ioContext.getIndividualHandle();
  for(Individual::Bag::size_type j=0; lChild != NULL; ++j) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Individual") {
        Beagle_NonNullPointerAssertM((*mMigrationBuffer)[j]);
        ioContext.setIndividualIndex(j);
        ioContext.setIndividualHandle((*mMigrationBuffer)[j]);
        (*mMigrationBuffer)[j]->readWithContext(lChild, ioContext);
        lChild = lChild->getNextSibling();
      }
    }
  }
  ioContext.setIndividualHandle(lOldIndivHandle);
  ioContext.setIndividualIndex(lOldIndivIndex);
}


/*!
 *  \brief Read a deme population from an XML node.
 *  \param inNode XML node to read the deme population from.
 *  \param ioContext Evolutionary context to use to read the deme population.
 *  \throw Beagle::IOException If the format is not respected.
 */
void Deme::readPopulation(XMLNode::Handle& inNode, Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Population"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Population> expected!");
  XMLNode::Handle lChild = inNode->getFirstChild();
  size_type lSize;
  for(lSize = 0; lChild != NULL; ++lSize) lChild = lChild->getNextSibling();
  if((getTypeAlloc()==NULL) && (lSize > size())) {
    std::ostringstream lMessageOSS;
    lMessageOSS << "Deme population size (" << lSize << ") is bigger than the actual size (";
    lMessageOSS << size() << "), and there is no type allocator for resizing!";
    throw Beagle_IOExceptionNodeM(*inNode, lMessageOSS.str());
  }
  resize(lSize);
  lChild = inNode->getFirstChild();
  unsigned int lOldIndivIndex = ioContext.getIndividualIndex();
  Individual::Handle lOldIndivHandle = ioContext.getIndividualHandle();
  for(size_type j=0; lChild != NULL; ++j) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Individual") {
        Beagle_NonNullPointerAssertM((*this)[j]);
        ioContext.setIndividualIndex(j);
        ioContext.setIndividualHandle((*this)[j]);
        (*this)[j]->readWithContext(lChild, ioContext);
        lChild = lChild->getNextSibling();
      }
    }
  }
  ioContext.setIndividualHandle(lOldIndivHandle);
  ioContext.setIndividualIndex(lOldIndivIndex);
}


/*!
 *  \brief Write the actual deme in a XML streamer.
 *  \param ioStreamer XML streamer to write the deme into.
 */
void Deme::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Deme");
  mStats->write(ioStreamer);
  mHallOfFame->write(ioStreamer);
  writeMigrationBuffer(ioStreamer);
  writePopulation(ioStreamer);
  ioStreamer.closeTag();
}

/*!
 *  \brief Write the actual deme migration buffer in a XML streamer.
 *  \param ioStreamer XML streamer to write the deme migration buffer into.
 */
void Deme::writeMigrationBuffer(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("MigrationBuffer");
  ioStreamer.insertAttribute("size", uint2str(mMigrationBuffer->size()));
  for(Individual::Bag::size_type i=0; i<mMigrationBuffer->size(); ++i) {
    Beagle_NonNullPointerAssertM((*mMigrationBuffer)[i]);
    (*mMigrationBuffer)[i]->write(ioStreamer);
  }
  ioStreamer.closeTag();
}


/*!
 *  \brief Write the actual deme population in a XML streamer.
 *  \param ioStreamer XML streamer to write the deme population into.
 */
void Deme::writePopulation(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Population");
  ioStreamer.insertAttribute("size", uint2str(size()));
  for(size_type i=0; i<size(); ++i) {
    Beagle_NonNullPointerAssertM((*this)[i]);
    (*this)[i]->write(ioStreamer);
  }
  ioStreamer.closeTag();
}

