/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/EvaluationOp.cpp
 *  \brief  Source code of class EvaluationOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <algorithm>
#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a new evaluation operator.
 *  \param inName Name of the operator.
 */
EvaluationOp::EvaluationOp(std::string inName) :
  BreederOp(inName)
{ }


/*!
 *  \brief Apply the evaluation operation on a breeding pool, returning a evaluated bred individual.
 *  \param inBreedingPool Breeding pool to use for the breeding operation.
 *  \param inChild Node handle associated to child node in the breeder tree.
 *  \param ioContext Evolutionary context of the breeding operation.
 *  \return Evaluated bred individual.
 */
Individual::Handle EvaluationOp::breed(Individual::Bag& inBreedingPool,
                                       BreederNode::Handle inChild,
                                       Context& ioContext)
{
  Beagle_NonNullPointerAssertM(inChild);

  Deme& lDeme = *ioContext.getDemeHandle();
  if(lDeme.getStats()->isValid()) {
    ioContext.setProcessedDeme(0);
    if((ioContext.getGeneration()!=0) && (lDeme.getStats()->existItem("total-processed"))) {
      ioContext.setTotalProcessedDeme((unsigned int)lDeme.getStats()->getItem("total-processed"));
    }
    else ioContext.setTotalProcessedDeme(0);
    lDeme.getStats()->setInvalid();

    if(ioContext.getDemeIndex()==0) {
      Stats& lVivaStats = *ioContext.getVivarium().getStats();
      ioContext.setProcessedVivarium(0);
      if((ioContext.getGeneration()!=0) && (lVivaStats.existItem("total-processed"))) {
        ioContext.setTotalProcessedVivarium((unsigned int)lVivaStats.getItem("total-processed"));
      }
      else ioContext.setTotalProcessedVivarium(0);
      lVivaStats.setInvalid();
    }
  }

  Beagle_NonNullPointerAssertM(inChild);
  Beagle_NonNullPointerAssertM(inChild->getBreederOp());
  Individual::Handle lBredIndividual =
    inChild->getBreederOp()->breed(inBreedingPool, inChild->getFirstChild(), ioContext);
    
  if((lBredIndividual->getFitness()==NULL) || (lBredIndividual->getFitness()->isValid()==false)) {
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "evaluation", "Beagle::EvaluationOp",
      "Evaluating the fitness of a new bred individual"
    );

    lBredIndividual->setFitness(evaluate(*lBredIndividual, ioContext));
    lBredIndividual->getFitness()->setValid();

    ioContext.setProcessedDeme(ioContext.getProcessedDeme()+1);
    ioContext.setTotalProcessedDeme(ioContext.getTotalProcessedDeme()+1);
    ioContext.setProcessedVivarium(ioContext.getProcessedVivarium()+1);
    ioContext.setTotalProcessedVivarium(ioContext.getTotalProcessedVivarium()+1);

    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "evaluation", "Beagle::EvaluationOp",
      std::string("The individual fitness value is: ")+
      lBredIndividual->getFitness()->serialize()
    );

    if(mDemeHOFSize->getWrappedValue() > 0) {
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "evaluation", "Beagle::EvaluationOp",
        "Updating the deme hall-of-fame"
      );
      lDeme.getHallOfFame().updateWithIndividual(mDemeHOFSize->getWrappedValue(),
                                                 *lBredIndividual, ioContext);
    }
    if(mVivaHOFSize->getWrappedValue() > 0) {
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "evaluation", "Beagle::EvaluationOp",
        "Updating the vivarium hall-of-fame"
      );
      ioContext.getVivarium().getHallOfFame().updateWithIndividual(mVivaHOFSize->getWrappedValue(),
                                                                   *lBredIndividual, ioContext);
    }
  }
  
  return lBredIndividual;
}


/*!
 *  \return Return selection probability of breeder operator.
 *  \param inChild Child node in the breeder tree.
 */
float EvaluationOp::getBreedingProba(BreederNode::Handle inChild)
{
  Beagle_NonNullPointerAssertM(inChild);
  Beagle_NonNullPointerAssertM(inChild->getBreederOp());
  return inChild->getBreederOp()->getBreedingProba(inChild->getFirstChild());
}


/*!
 *  \brief Initialize the parameters evaluation operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void EvaluationOp::initialize(System& ioSystem)
{
  BreederOp::initialize(ioSystem);
  
  if(ioSystem.getRegister().isRegistered("ec.hof.vivasize")) {
    mVivaHOFSize =
      castHandleT<UInt>(ioSystem.getRegister().getEntry("ec.hof.vivasize"));
  } else {
    mVivaHOFSize = new UInt(1);
    std::string lLongDescript = "Number of individuals kept in vivarium's hall-of-fame ";
    lLongDescript += "(best individuals so far). Note that a hall-of-fame contains only ";
    lLongDescript += "copies of the best individuals so far and is not used by the evolution ";
    lLongDescript += "process.";
    Register::Description lDescription(
      "Vivarium's hall-of-fame size",
      "UInt",
      "1",
      lLongDescript
    );
    ioSystem.getRegister().addEntry("ec.hof.vivasize", mVivaHOFSize, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ec.hof.demesize")) {
    mDemeHOFSize =
      castHandleT<UInt>(ioSystem.getRegister().getEntry("ec.hof.demesize"));
  } else {
    mDemeHOFSize = new UInt(0);
    std::string lLongDescript = "Number of individuals kept in each deme's hall-of-fame ";
    lLongDescript += "(best individuals so far). Note that a hall-of-fame contains only ";
    lLongDescript += "copies of the best individuals so far and is not used by the evolution ";
    lLongDescript += "process.";
    Register::Description lDescription(
      "Demes' hall-of-fame size",
      "UInt",
      "0",
      lLongDescript
    );
    ioSystem.getRegister().addEntry("ec.hof.demesize", mDemeHOFSize, lDescription);
  }
}


/*!
 *  \brief Apply the evaluation process on the invalid individuals of the deme.
 *  \param ioDeme Deme to process.
 *  \param ioContext Context of the evolution.
 */
void EvaluationOp::operate(Deme& ioDeme, Context& ioContext)
{
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "evaluation", "Beagle::EvaluationOp",
    std::string("Evaluating the individuals fitness of the ")+
    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
  );

  Individual::Handle lOldIndividualHandle = ioContext.getIndividualHandle();
  unsigned int lOldIndividualIndex = ioContext.getIndividualIndex();    

  ioContext.setProcessedDeme(0);
  if((ioContext.getGeneration()!=0) && (ioDeme.getStats()->existItem("total-processed"))) {
    ioContext.setTotalProcessedDeme((unsigned int)ioDeme.getStats()->getItem("total-processed"));
  }
  else ioContext.setTotalProcessedDeme(0);
  ioDeme.getStats()->setInvalid();

  if(ioContext.getDemeIndex()==0) {
    Stats& lVivaStats = *ioContext.getVivarium().getStats();
    ioContext.setProcessedVivarium(0);
    if((ioContext.getGeneration()!=0) && (lVivaStats.existItem("total-processed"))) {
      ioContext.setTotalProcessedVivarium((unsigned int)lVivaStats.getItem("total-processed"));
    }
    else ioContext.setTotalProcessedVivarium(0);
    lVivaStats.setInvalid();
  }
    
  for(unsigned int i=0; i<ioDeme.size(); i++) {
    if((ioDeme[i]->getFitness() == NULL) ||
       (ioDeme[i]->getFitness()->isValid() == false)) {

      Beagle_LogVerboseM(   
        ioContext.getSystem().getLogger(),
        "evaluation", "Beagle::EvaluationOp",
        std::string("Evaluating the fitness of the ")+uint2ordinal(i+1)+
        " individual"
      );

      ioContext.setIndividualIndex(i);
      ioContext.setIndividualHandle(ioDeme[i]);

      ioDeme[i]->setFitness(evaluate(*ioDeme[i], ioContext));
      ioDeme[i]->getFitness()->setValid();
      
      ioContext.setProcessedDeme(ioContext.getProcessedDeme()+1);
      ioContext.setTotalProcessedDeme(ioContext.getTotalProcessedDeme()+1);
      ioContext.setProcessedVivarium(ioContext.getProcessedVivarium()+1);
      ioContext.setTotalProcessedVivarium(ioContext.getTotalProcessedVivarium()+1);    
  
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "evaluation", "Beagle::EvaluationOp",
        std::string("The individual\'s fitness is: ")+
        ioDeme[i]->getFitness()->serialize()
      );
    }
  }
  
  ioContext.setIndividualIndex(lOldIndividualIndex);
  ioContext.setIndividualHandle(lOldIndividualHandle);
  
  if(mDemeHOFSize->getWrappedValue() > 0) {
    Beagle_LogDetailedM(
      ioContext.getSystem().getLogger(),
      "evaluation", "Beagle::EvaluationOp",
      "Updating the deme's hall-of-fame"
    );
    ioDeme.getHallOfFame().updateWithDeme(mDemeHOFSize->getWrappedValue(), ioDeme, ioContext);
    ioDeme.getHallOfFame().log(Logger::eVerbose, ioContext);
  }

  if(mVivaHOFSize->getWrappedValue() > 0) {
    Beagle_LogDetailedM(
      ioContext.getSystem().getLogger(),
      "evaluation", "Beagle::EvaluationOp",
      "Updating the vivarium's hall-of-fame"
    );
    ioContext.getVivarium().getHallOfFame().updateWithDeme(mVivaHOFSize->getWrappedValue(),
                                                           ioDeme, ioContext);
    ioContext.getVivarium().getHallOfFame().log(Logger::eVerbose, ioContext);
  }
}


/*!
 *  \brief Test the fitness of a given individual.
 *  \param inIndividual Handle to the individual to test.
 *  \param ioSystem Handle to the system to use to test the individual.
 *  \par Note:
 *    This method is provided as a mean to test some individuals after an evolution.
 */
Fitness::Handle EvaluationOp::test(Individual::Handle inIndividual, System::Handle ioSystem)
{
  Beagle_LogInfoM(
    ioSystem->getLogger(),
    "evaluation", "Beagle::EvaluationOp",
    std::string("Testing the following individual: ")+inIndividual->serialize()
  );
    
  Context::Alloc::Handle lContextAlloc =
    castHandleT<Context::Alloc>(ioSystem->getContextAllocatorHandle());
  Context::Handle lContext = castHandleT<Context>(lContextAlloc->allocate());
  lContext->setSystemHandle(ioSystem);
  lContext->setIndividualHandle(inIndividual);
  Fitness::Handle lFitness = evaluate(*inIndividual, *lContext);

  Beagle_LogInfoM(
    ioSystem->getLogger(),
    "evaluation", "Beagle::EvaluationOp",
    std::string("New fitness of the individual: ")+lFitness->serialize()
  );
    
  return lFitness;
}



