/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/FitnessMultiObj.cpp
 *  \brief  Source code of class FitnessMultiObj.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a multiobjective fitness object.
 *  \brief inSize Size of the multiobjective fitness value.
 *  \brief inValue Default initial value of the multiobjective fitness measures.
 */
FitnessMultiObj::FitnessMultiObj(unsigned int inSize, float inValue) :
  Fitness((inSize != 0)),
  std::vector<float>(inSize, inValue)
{ }


/*!
 *  \brief Construct a valid multiobjective fitness object.
 *  \param inFitness Value of the fitness.
 */
FitnessMultiObj::FitnessMultiObj(const std::vector<float>& inFitness) :
  Fitness(true),
  std::vector<float>(inFitness)
{ }


/*!
 *  \brief Return scaling factors, 1.0 for all objectives by default.
 *  \return Scaling factor of the dimensions of the objective space.
 */
std::vector<float> FitnessMultiObj::getScalingFactors() const
{
  return std::vector<float>(size(), 1.0);
}


/*!
 *  \brief Test whether the actual multiobjective fitness is Pareto-dominated.
 *  \param inRightFitness Second multiobjective fitness to test domination on.
 *  \return True if the actual multiobjective measure is dominated, false if not.
 */
bool FitnessMultiObj::isDominated(const FitnessMultiObj& inRightFitness) const
{
  if(isValid() != inRightFitness.isValid()) return false;
  if((isValid()==false) && (inRightFitness.isValid()==false)) return false;
  size_type lMinSize = minOf(size(), inRightFitness.size());
  if(lMinSize == 0) return false;
  bool lNotEqual = false;
  for(size_type i=0; i<lMinSize; ++i) {
    if((*this)[i] > inRightFitness[i]) return false;
    else if((*this)[i] < inRightFitness[i]) lNotEqual = true;
  }
  return lNotEqual;
}


/*!
 *  \brief Test equality of two multiobjective fitness measures.
 *  \param inRightObj Right measure to compare.
 *  \return True if the fitness are equal, false if not.
 */
bool FitnessMultiObj::isEqual(const Object& inRightObj) const
{
  const FitnessMultiObj& lRightFitness = castObjectT<const FitnessMultiObj&>(inRightObj);
  if(isValid() != lRightFitness.isValid()) return false;
  if((isValid()==false) && (lRightFitness.isValid()==false)) return true;
  if(size() != lRightFitness.size()) return false;
  for(size_type i=0; i<size(); ++i) {
    if((*this)[i] != lRightFitness[i]) return false;
  }
  return true;
}


/*!
 *  \brief Test order of two multiobjective fitness measures, from first to last objective.
 *  \param inRightObj Right measure to compare.
 *  \return True if the actual fitness is less from the right one, false if not.
 */
bool FitnessMultiObj::isLess(const Object& inRightObj) const
{
  const FitnessMultiObj& lRightFitness = castObjectT<const FitnessMultiObj&>(inRightObj);
  if(isValid() != lRightFitness.isValid()) return false;
  if((isValid()==false) && (lRightFitness.isValid()==false)) return false;
  size_type lMinSize = minOf(size(), lRightFitness.size());
  for(size_type i=0; i<lMinSize; ++i) {
    if((*this)[i] < lRightFitness[i]) return true;
    else if((*this)[i] > lRightFitness[i]) return false;
  }
  return false;
}


/*!
 *  \brief Read a multiobjective fitness from a XML node.
 *  \param inNode XML to use to read the fitness value.
 */
void FitnessMultiObj::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Fitness"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Fitness> expected!");

  std::string lValid = inNode->getTagAttribute("valid");
  if(lValid.empty() || (lValid == "yes")) {
    std::string lType = inNode->getTagAttribute("type");
    if(lType != "multiobj")
      throw Beagle_IOExceptionNodeM((*inNode), "fitness type mismatch!");

    clear();
    for(XMLNode::Handle lChild=inNode->getFirstChild();
        lChild!=NULL; lChild=lChild->getNextSibling()) {
      if((lChild->getType() != XMLNode::eTag) || (lChild->getTagName() != "Obj")) continue;
      XMLNode::Handle lChild2 = lChild->getFirstChild();
      if(lChild2 == NULL)
        throw Beagle_IOExceptionNodeM((*lChild), "needed a float value in the <Obj> tag!");
      if(lChild2->getType() != XMLNode::eString)
        throw Beagle_IOExceptionNodeM((*lChild2), "needed a float value in the <Obj> tag!");
      push_back(str2dbl(lChild2->getValue()));  
    }
    setValid();
  }
  else if(lValid == "no") setInvalid();
  else throw Beagle_IOExceptionNodeM((*inNode), "bad fitness validity value!");
}


/*!
 *  \brief Write a multiobjective fitness into an XML streamer.
 *  \param ioStreamer Streamer to use to write the fitness value.
 */
void FitnessMultiObj::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Fitness");
  ioStreamer.insertAttribute("type", "multiobj");
  if(isValid()) {
    ioStreamer.insertAttribute("size", uint2str(size()));
    for(size_type i=0; i<size(); ++i) {
      ioStreamer.openTag("Obj", false);
      ioStreamer.insertFloat((*this)[i]);
      ioStreamer.closeTag(false);
    }
  }
  else ioStreamer.insertAttribute("valid", "no"); 
  ioStreamer.closeTag(false);
}




