/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/FitnessSimple.cpp
 *  \brief  Source code of class FitnessSimple.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <cmath>
#include <cfloat>
#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Default construct a simple fitness object.
 */
FitnessSimple::FitnessSimple() :
  Fitness(false)
{ }


/*!
 *  \brief Construct a valid simple fitness object.
 *  \param inFitness Value of the fitness.
 */
FitnessSimple::FitnessSimple(float inFitness) :
  Fitness(true)
{
  setValue(inFitness);
}


/*!
 *  \brief Test equality of two simple fitness measure.
 *  \param inRightObj Right measure to compare.
 *  \return True if the fitness are equal, false if not.
 */
bool FitnessSimple::isEqual(const Object& inRightObj) const
{
  const FitnessSimple& lRightFitness = castObjectT<const FitnessSimple&>(inRightObj);
  if(isValid() != lRightFitness.isValid()) return false;
  if((isValid()==false) && (lRightFitness.isValid()==false)) return true;
  return mFitness == lRightFitness.mFitness;
}


/*!
 *  \brief Test order of two simple fitness measure.
 *  \param inRightObj Right measure to compare.
 *  \return True if the actual fitness is less from the right one, false if not.
 */
bool FitnessSimple::isLess(const Object& inRightObj) const
{
  const FitnessSimple& lRightFitness = castObjectT<const FitnessSimple&>(inRightObj);
  if(isValid() != lRightFitness.isValid()) return false;
  if((isValid()==false) && (lRightFitness.isValid()==false)) return false;
  return mFitness < lRightFitness.mFitness;
}


/*!
 *  \brief Read a simple fitness from a XML node.
 *  \param inNode XML  to use to read the fitness value.
 */
void FitnessSimple::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Fitness"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Fitness> expected!");

  std::string lValid = inNode->getTagAttribute("valid");
  if(lValid.empty() || (lValid == "yes")) {
    std::string lType = inNode->getTagAttribute("type");
    if(lType != "simple")
      throw Beagle_IOExceptionNodeM((*inNode), "fitness type mismatch!");

    XMLNode::Handle lChild = inNode->getFirstChild();
    if(lChild == NULL)
      throw Beagle_IOExceptionNodeM(*inNode, "no fitness value present!");
    if(lChild->getType() != XMLNode::eString)
      throw Beagle_IOExceptionNodeM(*inNode, "no fitness value present!");
    mFitness = str2dbl(lChild->getValue());
    setValid();
  }
  else if(lValid == "no") setInvalid();
  else throw Beagle_IOExceptionNodeM((*inNode), "bad fitness validity value!");
}


/*!
 *  \brief Set the fitness value and validity.
 *  \param inFitness New fitness value.
 */
void FitnessSimple::setValue(float inFitness)
{
  if(isFinite(inFitness)) mFitness = inFitness;
  else mFitness = FLT_MIN;
  setValid();
}


/*!
 *  \brief Write a simple fitness into an XML streamer.
 *  \param ioStreamer Streamer to use to write the fitness value.
 */
void FitnessSimple::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Fitness");
  ioStreamer.insertAttribute("type", "simple");
  if(isValid()) ioStreamer.insertFloat(mFitness);
  else ioStreamer.insertAttribute("valid", "no");
  ioStreamer.closeTag(false);
}

