/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/IOException.cpp
 *  \brief  Implementation of class IOException.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct the IOException using given values.
 *  \param inMessage     Message explaining the exception.
 *  \param inFileName    Name of the file where the error state had been detected.
 *  \param inLineNumber  Line number in the filename.
 */
IOException::IOException(std::string  inMessage,
                         std::string  inFileName,
                         unsigned int inLineNumber) :
  TargetedException(inMessage,inFileName,inLineNumber)
{ }


/*!
 *  \brief Construct the IOException using given values.
 *  \param inXMLNode     XML node in which the exception occured.
 *  \param inMessage     Message explaining the exception.
 *  \param inFileName    Name of the file where the error state had been detected.
 *  \param inLineNumber  Line number in the filename.
 */
IOException::IOException(const XMLNode& inXMLNode,
                         std::string  inMessage,
                         std::string  inFileName,
                         unsigned int inLineNumber) :
  TargetedException("",inFileName,inLineNumber)
{
  std::ostringstream lOSS;
  lOSS << "Error with XML node \"" << inXMLNode.getValue();
  lOSS << "\": " << inMessage;
  setMessage(lOSS.str());
}


/*!
 *  \brief Construct the IOException using given values.
 *  \param inXMLNode     XML node in which the exception occured.
 *  \param inTokenizer   Reference to the tokenizer related to the exception.
 *  \param inMessage     Message explaining the exception.
 *  \param inFileName    Name of the file where the error state had been detected.
 *  \param inLineNumber  Line number in the filename.
 */
IOException::IOException(const XMLNode& inXMLNode,
                         const Tokenizer& inTokenizer,
                         std::string  inMessage,
                         std::string  inFileName,
                         unsigned int inLineNumber) :
  TargetedException("",inFileName,inLineNumber)
{
  std::ostringstream lOSS;
  if(inTokenizer.getStreamName() != "") {
    lOSS << "Error when reading file \"" << inTokenizer.getStreamName() << "\", at line ";
  }
  else {
    lOSS << "Error when reading line ";
  }
  lOSS << inTokenizer.getLineNumber() << ", \"" << inXMLNode.getValue() << "\": " << inMessage;
  setMessage(lOSS.str());
}


/*!
 *  \brief  Get the name of the exception.
 *  \return Name of the exception.
 */
const char* IOException::getExceptionName() const throw()
{
  return "Beagle::IOException";
}

