/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/IntegerVector.cpp
 *  \brief  Source code of class IntegerVector.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.6 $
 *  $Date: 2004/10/23 23:38:13 $
 */

#include "beagle/Beagle.hpp"

#include <cfloat>
#include <climits> 
#include <string>
#include <sstream>

using namespace Beagle;


/*!
 * \brief Construct the population size parameter.
 * \param inSize Number of demes of the population.
 * \param inValue Size of the demes.
 */
IntegerVector::IntegerVector(unsigned int inSize, unsigned int inValue) :
  std::vector<unsigned int>(inSize, inValue)
{ }


/*!
 *  \brief  Test if an vector of integer is equal to another.
 *  \param  inRightObj Vector of integer to be compare to.
 *  \return True if actual vector is equal to right vector, false if not.
 */
bool IntegerVector::isEqual(const Object& inRightObj) const
{
  const IntegerVector& lRightIntVect = castObjectT<const IntegerVector&>(inRightObj);
  if(size() != lRightIntVect.size()) return false;
  const std::vector<unsigned int>& lThisVector  = *this;
  const std::vector<unsigned int>& lRightVector = lRightIntVect;
  return (lThisVector == lRightVector);
}


/*!
 *  \brief  Test if an vector of integer is less than another.
 *  \param  inRightObj Vector of integer to be compare to.
 *  \return True if actual vector is less than the right vector, false if not.
 */
bool IntegerVector::isLess(const Object& inRightObj) const
{
  const IntegerVector& lRightIntVect = castObjectT<const IntegerVector&>(inRightObj);
  const std::vector<unsigned int>& lThisVector  = *this;
  const std::vector<unsigned int>& lRightVector = lRightIntVect;
  return (lThisVector < lRightVector);
}


/*!
 *  \brief Read the integer vector from XML node.
 *  \param inNode XML node from which the integer vector is read.
 *  \throw IOException If a reading error occurs.
 */
void IntegerVector::read(XMLNode::Handle& inNode)
{
  if(inNode->getType() != XMLNode::eString)
    throw Beagle_IOExceptionNodeM(*inNode, "expected string to read integer vector!");
  resize(0);
  std::istringstream lISS(inNode->getValue());
  while(lISS.good()) {
    unsigned int lValue=UINT_MAX;
    lISS >> lValue;
    if(lValue==UINT_MAX) break;
    push_back(lValue);
    if(lISS.good()==false) break;
    int lDelim=lISS.get();
    if((lISS.good()==false) || (lDelim==-1)) break;
  }
}


/*!
 *  \brief Write the integer vector into a XML output stream.
 *  \param ioStreamer XML streamer in which the integer vector is written.
 */
void IntegerVector::write(XMLStreamer& ioStreamer) const
{
  std::ostringstream lOSS;
  for(size_type i=0; i<size(); i++) {
    lOSS << (*this)[i];
    if(i != (size()-1)) lOSS << '/';
  }
  ioStreamer.insertString(lOSS.str());
}

