/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Logger.cpp
 *  \brief  Implementation of class Logger.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */
 
#include "beagle/Beagle.hpp"

#include <ctime>
#include <iomanip>
#include <fstream>

using namespace Beagle;


/*!
 *  \brief Construct a logger.
 */
Logger::Logger() :
  mInitialized(false)
{ }


/*!
 *  \brief Initialize logger by registering its parameters.
 *  \param ioSystem Reference to the system to use for initialization.
 */
void Logger::initialize(System& ioSystem)
{
  Beagle_LogDetailedM(
    ioSystem.getLogger(),
    "logger", "Beagle::Logger",
    "Initializing logger"
  );
  
  if(ioSystem.getRegister().isRegistered("lg.log.level")) {
    mLogLevel = castHandleT<UInt>(ioSystem.getRegister().getEntry("lg.log.level"));
  } else {
    mLogLevel = new UInt(eInfo);
    std::string lLongDescrip("Log level used during output generation. ");
    lLongDescrip += "Log levels available are: (0) no log, (1) basic logs, (2) stats, ";
    lLongDescrip += "(3) general informations, (4) details on operations, ";
    lLongDescrip += "(5) trace of the algorithms, (6) verbose, ";
    lLongDescrip += "(7) debug (enabled only in full debug mode).";
    Register::Description lDescription(
      "Log level",
      "UInt",
      "3",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.log.level", mLogLevel, lDescription);
  }
}


/*!
 *  \brief Show greeting message in log file.
 *  \param ioSystem Reference to the system to used.
 */
void Logger::postInit(System& ioSystem)
{
  mInitialized = true;

  Beagle_LogBasicM(
    (*this),
    "logger", "Beagle::Logger",
    std::string("Open BEAGLE, version ")+BEAGLE_VERSION
  );

  switch(mLogLevel->getWrappedValue()) {
    case eNothing:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 0 (no message)"
      );
      break;
    }
    case eBasic:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 1 (essential informations)"
      );
      break;
    }
    case eStats:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 2 (evolution statistics)"
      );
      break;
    }
    case eInfo:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 3 (general informations)"
      );
      break;
    }
    case eDetailed:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 4 (details on operations)"
      );
      break;
    }
    case eTrace:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 5 (trace of the algorithms)"
      );
      break;
    }
    case eVerbose:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 6 (details on everything)"
      );
#ifdef BEAGLE_NDEBUG
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Log level 6 (verbose) is unused as optimization mode is enabled"
      );
#endif // BEAGLE_NDEBUG
      break;
    }
    case eDebug:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Logger initialized, using log level 7 (debug)"
      );
#ifdef BEAGLE_NDEBUG
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Log level 7 (debug) is unused as optimization mode is enabled"
      );
#endif // BEAGLE_NDEBUG
#ifndef BEAGLE_FULL_DEBUG
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        "Log level 7 (debug) is unused as full debug mode is disabled"
      );
#endif // BEAGLE_FULL_DEBUG
      break;
    }
    default:
    {
      Beagle_LogBasicM(
        (*this),
        "logger", "Beagle::Logger",
        std::string("Logger initialized, using log level ")+
        uint2str(mLogLevel->getWrappedValue())+
        " (unused)"
      );
      break;
    }
  }

  for(std::list<Message>::const_iterator lIter = mBuffer.begin();
      lIter != mBuffer.end(); ++lIter) {
    if(lIter->mLogLevel <= mLogLevel->getWrappedValue()) {
      outputMessage(lIter->mLogLevel, lIter->mType, lIter->mClass, lIter->mMessage);
    }
  }
  mBuffer.clear();
}







