/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/LoggerXML.cpp
 *  \brief  Implementation of class LoggerXML.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.6 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <cstdio>
#include <iomanip>
#include <fstream>

using namespace Beagle;

/*!
 *  \brief Construct a new LoggerXML.
 */
LoggerXML::LoggerXML() :
  mActualFileName(""),
  mLogOutStream(NULL),
  mTerminated(false)
{ }


/*!
 *  \brief Copy construct a LoggerXML is undefined.
 */
LoggerXML::LoggerXML(const LoggerXML& inLoggerXML)
{
  throw Beagle_UndefinedMethodInternalExceptionM("LoggerXML(const LoggerXML&)", "LoggerXML");
}


/*!
 *  \brief Destruct a LoggerXML.
 */
LoggerXML::~LoggerXML()
{
  terminate();
}


/*!
 *  \brief Copy a LoggerXML is undefined.
 */
LoggerXML& LoggerXML::operator=(const LoggerXML& inLoggerXML)
{
  throw Beagle_UndefinedMethodInternalExceptionM("operator=","LoggerXML");
}


/*!
 *  \brief Initialize logger by registering its parameters.
 *  \param ioSystem Reference to the system to use for initialization.
 */
void LoggerXML::initialize(System& ioSystem)
{
  Logger::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("lg.file.name")) {
    mLogFileName = castHandleT<String>(ioSystem.getRegister().getEntry("lg.file.name"));
  } else {
    mLogFileName = new String("beagle.log");
    std::string lLongDescrip("Filename in which messages are outputed. ");
    lLongDescrip += "An empty string file name means no output is done to a file.";
    Register::Description lDescription(
      "Log filename",
      "String",
      "\"beagle.log\"",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.file.name", mLogFileName, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("lg.console.enabled")) {
    mConsoleEnabled = castHandleT<Bool>(ioSystem.getRegister().getEntry("lg.console.enabled"));
  } else {
    mConsoleEnabled = new Bool(true);
    std::string lLongDescrip("Flag whether logging is done to the console (STDOUT).");
    Register::Description lDescription(
      "Console logging switch",
      "Bool",
      "1",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.console.enabled", mConsoleEnabled, lDescription);
  }
  
  if(ioSystem.getRegister().isRegistered("lg.show.level")) {
    mShowLevel = castHandleT<Bool>(ioSystem.getRegister().getEntry("lg.show.level"));
  } else {
    mShowLevel = new Bool(false);
    std::string lLongDescrip("Flag whether logging level in outputed in the logs.");
    Register::Description lDescription(
      "Show level in logs",
      "Bool",
      "0",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.show.level", mShowLevel, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("lg.show.type")) {
    mShowType = castHandleT<Bool>(ioSystem.getRegister().getEntry("lg.show.type"));
  } else {
    mShowType = new Bool(false);
    std::string lLongDescrip("Flag whether message type is outputed in the logs.");
    Register::Description lDescription(
      "Show message type in logs",
      "Bool",
      "0",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.show.type", mShowType, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("lg.show.class")) {
    mShowClass = castHandleT<Bool>(ioSystem.getRegister().getEntry("lg.show.class"));
  } else {
    mShowClass = new Bool(false);
    std::string lLongDescrip("Flag whether class name is outputed in the logs.");
    Register::Description lDescription(
      "Show class name in logs",
      "Bool",
      "0",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("lg.show.class", mShowClass, lDescription);
  }
}


/*!
 *  \brief Write message to the output device.
 *  \param inLevel Log level of the message
 *  \param inType Type of the message to log.
 *  \param inClass Class associated to the message.
 *  \param inMessage Message to log in the output device.
 */
void LoggerXML::outputMessage(unsigned int inLevel,
                              std::string inType,
                              std::string inClass,
                              std::string inMessage)
{
  if(mTerminated)
    throw Beagle_RunTimeExceptionM("Can't log in a terminated XML logger!");
  
  if(mActualFileName != mLogFileName->getWrappedValue()) {
    mActualFileName = mLogFileName->getWrappedValue();
    if(mStreamerFile != NULL) {
      mStreamerFile->closeTag();
      mStreamerFile->closeTag();
      *mLogOutStream << std::endl;
      mStreamerFile = NULL;     
    }
    if(mLogOutStream) {
      mLogOutStream->close();
      delete mLogOutStream;
      mLogOutStream = NULL;
    }
    if(mLogFileName->getWrappedValue().empty() == false) {
      std::string lFilenameBak = mLogFileName->getWrappedValue() + "~";
      std::remove(lFilenameBak.c_str());
      std::rename(mLogFileName->getWrappedValue().c_str(), lFilenameBak.c_str());
      mLogOutStream = new std::ofstream(mLogFileName->getWrappedValue().c_str());
      mStreamerFile = new XMLStreamer(*mLogOutStream);
      mStreamerFile->insertHeader();
      mStreamerFile->openTag("Beagle");
      mStreamerFile->insertAttribute("version", BEAGLE_VERSION);
      mStreamerFile->openTag("Logger");
    }
  }
  if(mStreamerFile != NULL) {
    mStreamerFile->openTag("Log");
    if(mShowLevel->getWrappedValue()) mStreamerFile->insertAttribute("level", uint2str(inLevel));
    if(mShowType->getWrappedValue())  mStreamerFile->insertAttribute("type",  inType);
    if(mShowClass->getWrappedValue()) mStreamerFile->insertAttribute("class", inClass);
    mStreamerFile->insertString(inMessage);
    mStreamerFile->closeTag(false);
  }
  
  if(mConsoleEnabled->getWrappedValue()) {
    if(mStreamerConsole == NULL) {
      mStreamerConsole = new XMLStreamer(std::cout);
      mStreamerConsole->insertHeader();
      mStreamerConsole->openTag("Beagle");
      mStreamerConsole->insertAttribute("version", BEAGLE_VERSION); 
      mStreamerConsole->openTag("Logger");
    }
    mStreamerConsole->openTag("Log");
    if(mShowLevel->getWrappedValue()) mStreamerConsole->insertAttribute("level", uint2str(inLevel));
    if(mShowType->getWrappedValue())  mStreamerConsole->insertAttribute("type",  inType);
    if(mShowClass->getWrappedValue()) mStreamerConsole->insertAttribute("class", inClass);
    mStreamerConsole->insertString(inMessage);
    mStreamerConsole->closeTag(false);
  }
  else if(mStreamerConsole != NULL) {
    mStreamerConsole->closeTag();
    mStreamerConsole->closeTag();
    std::cout << std::endl;
    mStreamerConsole = NULL;
  }
}


/*!
 *  \brief Write Beagle object to the output device.
 *  \param inLevel Log level of the message
 *  \param inType Type of the message to log.
 *  \param inClass Class associated to the message.
 *  \param inObject Beagle object to log in the output device.
 */
void LoggerXML::outputObject(unsigned int inLevel,
                             std::string inType,
                             std::string inClass,
                             const Object& inObject)
{
  if(mTerminated)
    throw Beagle_RunTimeExceptionM("Can't log in a terminated XML logger!");
  
  if(mActualFileName != mLogFileName->getWrappedValue()) {
    mActualFileName = mLogFileName->getWrappedValue();
    if(mStreamerFile != NULL) {
      mStreamerFile->closeTag();
      mStreamerFile->closeTag();
      *mLogOutStream << std::endl;
      mStreamerFile = NULL;
    }
    if(mLogOutStream) {
      mLogOutStream->close();
      delete mLogOutStream;
      mLogOutStream = NULL;
    }
    if(mLogFileName->getWrappedValue().empty() == false) {
      std::string lFilenameBak = mLogFileName->getWrappedValue() + "~";
      std::remove(lFilenameBak.c_str());
      std::rename(mLogFileName->getWrappedValue().c_str(), lFilenameBak.c_str());
      mLogOutStream = new std::ofstream(mLogFileName->getWrappedValue().c_str());
      mStreamerFile = new XMLStreamer(*mLogOutStream);
      mStreamerFile->insertHeader();
      mStreamerFile->openTag("Beagle");
      mStreamerFile->insertAttribute("version", BEAGLE_VERSION); 
      mStreamerFile->openTag("Logger");
    }
  }
  if(mStreamerFile != NULL) {
    mStreamerFile->openTag("Log");
    if(mShowLevel->getWrappedValue()) mStreamerFile->insertAttribute("level", uint2str(inLevel));
    if(mShowType->getWrappedValue())  mStreamerFile->insertAttribute("type",  inType);
    if(mShowClass->getWrappedValue()) mStreamerFile->insertAttribute("class", inClass);
    inObject.write(*mStreamerFile);
    mStreamerFile->closeTag();
  }

  if(mConsoleEnabled->getWrappedValue()) {
    if(mStreamerConsole == NULL) {
      mStreamerConsole = new XMLStreamer(std::cout);
      mStreamerConsole->insertHeader();
      mStreamerConsole->openTag("Beagle");
      mStreamerConsole->insertAttribute("version", BEAGLE_VERSION);  
      mStreamerConsole->openTag("Logger");
    }
    mStreamerConsole->openTag("Log");
    if(mShowLevel->getWrappedValue()) mStreamerConsole->insertAttribute("level", uint2str(inLevel));
    if(mShowType->getWrappedValue())  mStreamerConsole->insertAttribute("type",  inType);
    if(mShowClass->getWrappedValue()) mStreamerConsole->insertAttribute("class", inClass);
    inObject.write(*mStreamerConsole);
    mStreamerConsole->closeTag();
  }
  else if(mStreamerConsole != NULL) {
    mStreamerConsole->closeTag();
    mStreamerConsole->closeTag();
    std::cout << std::endl;
    mStreamerConsole = NULL;
  }
}


/*!
 *  \brief Terminate XML logger operations.
 */
void LoggerXML::terminate()
{
  if(mTerminated == false) {
    mTerminated = true;
    if(mStreamerFile != NULL) {
      mStreamerFile->closeTag();
      mStreamerFile->closeTag();
      *mLogOutStream << std::endl;
      mStreamerFile = NULL;
    }
    if(mStreamerConsole != NULL) {
      mStreamerConsole->closeTag();
      mStreamerConsole->closeTag();
      std::cout << std::endl;
      mStreamerConsole = NULL;
    }
    if(mLogOutStream) {
      mLogOutStream->close();
      delete mLogOutStream;
      mLogOutStream = NULL;
    }
  }
}

