/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Matrix.cpp
 *  \brief  Source code of class Matrix.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.10 $
 *  $Date: 2004/10/28 17:58:48 $
 */

#include "beagle/Beagle.hpp"

#include <cmath>
#include <cfloat>
#include <float.h>
#include <algorithm>

using namespace Beagle;

          
/*!
 *  \brief Construct a matrix of size (inNumberRows,inNumberColumns) using number value inValue.
 *  \param inNumberRows Number of rows in matrix.
 *  \param inNumberColumns Number of columns in matrix.
 *  \param inValue Default value of data in matrix.
 */
Matrix::Matrix(unsigned int inNumberRows, unsigned int inNumberColumns, double inValue) :
  std::vector<double>((inNumberRows*inNumberColumns), inValue),
  mRows(inNumberRows)
{ }
  

/*!
 *  \brief Construct a (n,1) matrix from a STL vector of doubles.
 *  \param inDoubleVector STL vector used to build the matrix.
 */
Matrix::Matrix(const std::vector<double>& inDoubleVector) :
  std::vector<double>(inDoubleVector),
  mRows(inDoubleVector.size())
{ }


/*!
 *  \brief Assign a STL double vector to a matrix.
 *  \param inRightVector STL double vector assigned to matrix.
 */
Matrix& Matrix::operator=(const std::vector<double>& inRightVector)
{
  if(this==&inRightVector) return *this;
  std::vector<double>::operator=(inRightVector);
  mRows=size();
  return *this;
}


/*!
 *  \brief Add together two matrices and return result.
 *  \param inRightMatrix Matrix to add.
 *  \return Resulting matrix.
 *  \throw Beagle::RunTimeException If matrices sizes mismatch.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator += whenever possible.
 */
Matrix Matrix::operator+(const Matrix& inRightMatrix) const
{
  Matrix lTmp = *this;
  return lTmp += inRightMatrix;
}


/*!
 *  \brief Add a matrix to the actual one and assign result to actual matrix.
 *  \param inRightMatrix Matrix to add to the actual one.
 *  \return Reference to actual matrix matrix.
 *  \throw Beagle::RunTimeException If matrices sizes mismatch.
 */
Matrix& Matrix::operator+=(const Matrix& inRightMatrix)
{
  if((mRows!=inRightMatrix.mRows) || (size()!=inRightMatrix.size()))
    throw Beagle_RunTimeExceptionM("Could not add matrices, size mismatch!");
  for(unsigned int i=0; i<size(); ++i) (*this)[i] += inRightMatrix[i];
  return *this;
}


/*!
 *  \brief Add a scalar to the actual matrix values and return result.
 *  \param inRightScalar Scalar to add.
 *  \return Resulting matrix.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator += whenever possible.
 */
Matrix Matrix::operator+(double inRightScalar) const
{
  Matrix lTmp = *this;
  return lTmp += inRightScalar;
}


/*!
 *  \brief Add a scalar to the actual matrix values and assign result to the matrix.
 *  \param inRightScalar Scalar to add.
 *  \return Reference to actual matrix.
 */
Matrix& Matrix::operator+=(double inRightScalar)
{
  for(unsigned int i=0; i<size(); ++i) (*this)[i] += inRightScalar;
  return *this;
}


/*!
 *  \brief Subtract two matrices and return result.
 *  \param inRightMatrix Matrix to subtract from the actual one.
 *  \return Resulting matrix.
 *  \throw Beagle::RunTimeException If matrices sizes mismatch.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator -= whenever possible.
 */
Matrix Matrix::operator-(const Matrix& inRightMatrix) const
{
  Matrix lTmp = *this;
  return lTmp -= inRightMatrix;
}


/*!
 *  \brief Subtract a matrix from the actual one and assign result.
 *  \param inRightMatrix Matrix to subtract.
 *  \return Reference to actual matrix.
 *  \throw Beagle::RunTimeException If matrices sizes mismatch.
 */
Matrix& Matrix::operator-=(const Matrix& inRightMatrix)
{
  if((mRows!=inRightMatrix.mRows) || (size()!=inRightMatrix.size()))
    throw Beagle_RunTimeExceptionM("Could not subtract matrices, size mismatch!");
  for(unsigned int i=0; i<size(); ++i) (*this)[i] -= inRightMatrix[i];
  return *this;
}


/*!
 *  \brief Subtract a scalar from the actual matrix and return result.
 *  \param inRightScalar Scalar to subtract.
 *  \return Resulting matrix.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator -= whenever possible.
 */
Matrix Matrix::operator-(double inRightScalar) const
{
  Matrix lTmp = *this;
  return lTmp -= inRightScalar;
}


/*!
 *  \brief Subtract a scalar from the actual matrix and assign result.
 *  \param inRightScalar Scalar to subtract.
 *  \return Reference to actual matrix matrix.
 */
Matrix& Matrix::operator-=(double inRightScalar)
{
  for(unsigned int i=0; i<size(); ++i) (*this)[i] -= inRightScalar;
  return *this;
}


/*!
 *  \brief Multiply two matrices and return result.
 *  \param inRightMatrix Right matrix to multiply with actual one.
 *  \return Resulting matrix.
 *  \throw Beagle::RunTimeException If sizes are incompatible.
 */
Matrix Matrix::operator*(const Matrix& inRightMatrix) const
{
  const unsigned int lColsLeft = getNumberColumns();
  const unsigned int lColsRight = inRightMatrix.getNumberColumns();
  if(lColsLeft != inRightMatrix.mRows)
    throw Beagle_RunTimeExceptionM("Could not multiply matrices, incompatible size!");
  Matrix lResult(mRows, lColsRight, 0.0);
  for(unsigned int j=0; j<lColsRight; ++j) {
    for(unsigned int i=0; i<mRows; ++i) {
      for(unsigned int k=0; k<lColsLeft; ++k) lResult(i,j) += ((*this)(i,k) * inRightMatrix(k,j));
    }
  }
  return lResult;
}


/*!
 *  \brief Multiply a matrix with a scalar and return result.
 *  \param inRightScalar Scalar to multiply to matrix.
 *  \return Resulting matrix.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator *= whenever possible.
 */
Matrix Matrix::operator*(double inRightScalar) const
{
  Matrix lTmp;
  return lTmp *= inRightScalar;
}


/*!
 *  \brief Subtract scalar from the actual matrix.
 *  \param inRightScalar Scalar to subtract.
 *  \return Reference to actual matrix.
 */
Matrix& Matrix::operator*=(double inRightScalar)
{
  for(unsigned int i=0; i<size(); ++i) (*this)[i] *= inRightScalar;
  return *this;
}


/*!
 *  \brief Divide a matrix with a scalar and return result.
 *  \param inRightScalar Scalar to divide the matrix.
 *  \return Resulting matrix.
 *  \warning This operator is highly inefficient as it results in several temporary copies of
 *    resulting matrix. Use operator /= whenever possible.
 */
Matrix Matrix::operator/(double inRightScalar) const
{
  Matrix lTmp;
  return lTmp /= inRightScalar;
}


/*!
 *  \brief Divide the actual matrix with a scalar and assign result to actual matrix.
 *  \param inRightScalar Scalar used to divide.
 *  \return Reference to actual divided matrix.
 */
Matrix& Matrix::operator/=(double inRightScalar)
{
  for(unsigned int i=0; i<size(); ++i) (*this)[i] /= inRightScalar;
  return *this;
}


/*!
 *  \brief  Compute determinant of a square matrix.
 *  \return Determinant of the matrix.
 *  \throw  Beagle::RunTimeException If the matrix is not square.
 */
double Matrix::computeDeterminant() const
{
  if(isSquare() == false)
    throw Beagle_RunTimeExceptionM("Could not get determinant from a non-square matrix!");
  Matrix lDecompose;
  std::vector<unsigned int> lIndexes(mRows);
  int lD;
  decomposeLU(lDecompose, lIndexes, lD);
  double lDet = lD;
  for(unsigned int i=0; i<mRows; ++i) lDet *= lDecompose(i,i);
  return lDet;
}


/*!
 *  \brief Compute eigenvalues and eigenvectors of symetric square matrix.
 *  \param outEigenValues Eigenvalues computed from the matrix.
 *  \param outEigenVectors Eigenvectors computed from the matrix.
 *  \throw Beagle::RunTimeException If the actual matrix is not square or non-symetric.
 */
void Matrix::computeEigenSystem(Vector& outEigenValues, Matrix& outEigenVectors) const
{
  if(size() == 0) {
    outEigenValues.resize(0);
    outEigenVectors.resize(0,0);
    return;
  }
  
  // Assert the matrix is square and symetric.
  if(isSquare()==false)
    throw Beagle_RunTimeExceptionM("Can't compute eigensystem from non-square matrix!");
  for(unsigned int i=1; i<mRows; ++i) {
    for(unsigned int j=0; j<i; ++j) {
      if((*this)(i,j) != (*this)(j,i))
        throw Beagle_RunTimeExceptionM("Can't compute eigensystem from non-symetric matrix!");
    }
  }
  outEigenValues.resize(mRows);
  outEigenVectors.resize(mRows,mRows);
  if(size()==0) return;    

  // Computer eigenvectors/eigenvalues using Triagonal QL method
  Vector lE(mRows);
  tred2(outEigenValues, lE, outEigenVectors);
  tql2(outEigenValues, lE, outEigenVectors);
  
  // Sort by eigenvalues.
  for(unsigned int j=0; j<outEigenValues.size(); ++j) {
    double lMax=outEigenValues[j];
    unsigned int lMaxArg=j;
    for(unsigned int l=(j+1); l<outEigenValues.size(); ++l) {
      if(outEigenValues[l] > lMax) {
        lMax=outEigenValues[l];
        lMaxArg=l;
      }
    }
    if(lMaxArg!=j) {
      for(unsigned int r=0; r<outEigenVectors.mRows; ++r) {
        double lTmp = outEigenVectors(r,j);
        outEigenVectors(r,j) = outEigenVectors(r,lMaxArg);
        outEigenVectors(r,lMaxArg) = lTmp;
      }
      double lTmp = outEigenValues[j];
      outEigenValues[j] = outEigenValues[lMaxArg];
      outEigenValues[lMaxArg] = lTmp;
    }
  }
}


/*!
 *  \brief Extract a subrange in [inFirstRow,inLastRow) and [inFirstColumn,inLastColumn).
 *  \param outSubMatrix Resulting submatrix extracted.
 *  \param inFirstRow First row index of the range to extract (inclusive).
 *  \param inLastRow Last row index of the range to extract (exclusive).
 *  \param inFirstColumn First column index of the range to extract (inclusive).
 *  \param inLastColumn Last column index of the range to extract (exclusive).
 *  \throw Beagle::AssertException If rows or columns indexes given are out-of-bound or if
 *           indexes of last row/columns is smaller or equal to the first one.
 */
void Matrix::extract(Matrix& outSubMatrix, unsigned int inFirstRow, unsigned int inLastRow,
                     unsigned int inFirstColumn, unsigned int inLastColumn) const
{
  const unsigned int lCols=getNumberColumns();
  Beagle_BoundCheckAssertM(inFirstRow+1,1,mRows);
  Beagle_BoundCheckAssertM(inLastRow,1,mRows);
  Beagle_BoundCheckAssertM(inFirstColumn+1,1,lCols);
  Beagle_BoundCheckAssertM(inLastColumn,1,lCols);  
  Beagle_AssertM(inLastRow > inFirstRow);
  Beagle_AssertM(inLastColumn > inFirstColumn);
  outSubMatrix.resize((inLastRow-inFirstRow), (inLastColumn-inFirstColumn));
  for(unsigned int j=inFirstColumn; j<inLastColumn; ++j) {
    for(unsigned int i=inFirstRow; i<inLastRow; ++i) outSubMatrix(i,j) = (*this)(i,j);
  }
}


/*!
 *  \brief Get maximum value of matrix.
 *  \return Maximum value of matrix.
 */
double Matrix::getMaximum() const
{
  Beagle_AssertM(size()!=0);
  return *(std::max_element<const_iterator>(begin(), end()));
}


/*!
 *  \brief Get maximum value of matrix along with the index of the element.
 *  \param outArgMaxRow Row index of maximum element.
 *  \param outArgMaxCol Column index of the maximum element.
 *  \return Maximum value of matrix.
 */
double Matrix::getMaximum(unsigned int& outArgMaxRow, unsigned int& outArgMaxCol) const
{
  Beagle_AssertM(size()!=0);
  double lMaxElement=(*this)(0,0);
  outArgMaxRow=outArgMaxCol=0;
  const unsigned int lCols=getNumberColumns();
  for(unsigned int i=0; i<mRows; ++i) {
    for(unsigned int j=0; j<lCols; ++j) {
      const double lValue=(*this)(i,j);
      if(lValue>lMaxElement) {
        lMaxElement=lValue;
        outArgMaxRow=i;
        outArgMaxCol=j;
      }
    }
  }
  return lMaxElement;
}


/*!
 *  \brief Get minimum value of matrix.
 *  \return Minimum value of matrix.
 */
double Matrix::getMinimum() const
{
  Beagle_AssertM(size()!=0);
  return *(std::min_element<const_iterator>(begin(), end()));
}


/*!
 *  \brief Get minimum value of matrix along with the index of the element.
 *  \param outArgMinRow Row index of minimum element.
 *  \param outArgMinCol Column index of the minimum element.
 *  \return Minimum value of matrix.
 */
double Matrix::getMinimum(unsigned int& outArgMinRow, unsigned int& outArgMinCol) const
{
  Beagle_AssertM(size()!=0);
  double lMinElement=(*this)(0,0);
  outArgMinRow=outArgMinCol=0;
  const unsigned int lCols=getNumberColumns();
  for(unsigned int i=0; i<mRows; ++i) {
    for(unsigned int j=0; j<lCols; ++j) {
      const double lValue=(*this)(i,j);
      if(lValue<lMinElement) {
        lMinElement=lValue;
        outArgMinRow=i;
        outArgMinCol=j;
      }
    }
  }
  return lMinElement;
}


/*!
 *  \brief Compute the inverse of the matrix and assign it to current one.
 *  \throw Beagle::RunTimeException If the matrix is not square.
 */
void Matrix::inverse()
{
  if(isSquare()==false) throw Beagle_RunTimeExceptionM("Could not inverse a non-square matrix!");
  if(size()==0) return;
  const unsigned int lCols=getNumberColumns();
  Matrix lDecompose;
  std::vector<unsigned int> lIndexes(mRows);
  int lD;
  decomposeLU(lDecompose, lIndexes, lD);
  setIdentity(mRows);
  Matrix lB(mRows, 1);
  for(unsigned int j=0; j<lCols; ++j) {
    for(unsigned int i=0; i<mRows; ++i) lB(i,0) = (*this)(i,j);
    lDecompose.backSubstituteLU(lIndexes, lB);
    for(unsigned int i=0; i<mRows; ++i) (*this)(i,j) = lB(i,0);
  }
}


/*!
 *  \brief Determine the equality of two matrices.
 *  \param inRightObj Right-hand side matrix to compare equality with.
 *  \return False: if the matrices are of different size or not equal for one or more elements,
 *            True: otherwize.
 */
bool Matrix::isEqual(const Object& inRightObj) const
{
  const Matrix& lRightMatrix = castObjectT<const Matrix&>(inRightObj);
  if((mRows!=lRightMatrix.mRows) || (size()!=lRightMatrix.size())) return false;
  const std::vector<double>& lLeft = *this;
  const std::vector<double>& lRight = lRightMatrix;
  return (lLeft == lRight);
}


/*!
 *  \brief Read matrix for XML node.
 *  \param inNode XML node to read matrix from.
 *
 *  Format of matrix written is simply a text string composed of the floating-point numbers,
 *  each separated by one character delimiter. Numbers are listed row by row, from uppermost
 *  to lowermost row. In-row numbers can be separated by any delimiters except whitespaces,
 *  ';', '.', and number characters. Rows themselves are delimited by character ';'. Within
 *  these specifications, the MATLAB matrix format is valid (i.e. '1,0;0,1' for a 2x2 identity
 *  matrix).
 *
 */
void Matrix::read(XMLNode::Handle& inNode)
{
  if(inNode->getType() != XMLNode::eString)
    throw Beagle_IOExceptionNodeM(*inNode, "expected string to read matrix!");
  std::vector<double>::resize(0);
  mRows=0;
  std::istringstream lISS(inNode->getValue());
  std::vector<double> lTmpVec;
  while(lISS.good()) {
    double lValue=DBL_MAX;
    lISS >> lValue;
    if(lValue==DBL_MAX) break;
    lTmpVec.push_back(lValue);
    int lDelim=lISS.get();
    if((lISS.good()==false) || (lDelim==-1)) break;
    if(lDelim==int(';')) {
      if(mRows==0) mRows=lTmpVec.size();
      else if(mRows!=lTmpVec.size()) {
        std::ostringstream lOSS;
        lOSS << "Bad number of columns (" << lTmpVec.size() << ") in matrix row " << size()/mRows;
        lOSS << " read XML node. Expected " << mRows << " columns read.";
        throw Beagle_IOExceptionMessageM(lOSS.str());
      }
      insert(end(), lTmpVec.begin(), lTmpVec.end());
      lTmpVec.resize(0);
    }
  }
  if(lTmpVec.size() != 0) {
    if(mRows==0) mRows=lTmpVec.size();
    else if(mRows!=lTmpVec.size()) {
      std::ostringstream lOSS;
      lOSS << "Bad number of columns (" << lTmpVec.size() << ") in matrix row " << size()/mRows;
      lOSS << " read XML node. Expected " << mRows << " columns read.";
      throw Beagle_IOExceptionMessageM(lOSS.str());
    }
    insert(end(), lTmpVec.begin(), lTmpVec.end());
  }
  transpose();
}


/*!
 *  \brief Change the size of the matrix, pad with given value if necessary.
 *  \param inNumberRows New number of rows in matrix.
 *  \param inNumberCols New number of columns in matrix.
 *  \param inValue Default value of added numbers in matrix.
 */
void Matrix::resize(unsigned int inNumberRows, unsigned int inNumberCols, double inValue)
{
  const unsigned int lSizeOrig = size();
  const unsigned int lSizeNew = inNumberRows * inNumberCols;
  const unsigned int lCols = getNumberColumns();
  if((lSizeOrig==0) || (lSizeNew==0)) {
    std::vector<double>::resize(lSizeNew, inValue);
  }
  else if((inNumberRows>=mRows) && (inNumberCols>=lCols)) {
    std::vector<double>::resize(lSizeNew, inValue);
    for(unsigned int j=0; j<inNumberCols; ++j) {
      const unsigned int lC=(inNumberCols-1)-j;
      for(unsigned int i=0; i<inNumberRows; ++i) {
        const unsigned int lR=(inNumberRows-1)-i;
        (*this)[(lC*inNumberRows)+lR] = ((lR<mRows)&&(lC<lCols)) ? (*this)[(lC*mRows)+lR] : inValue;
      }
    }
  }
  else {
    Matrix lTmp(*this);
    std::vector<double>::resize(lSizeNew, inValue);
    for(unsigned int j=0; j<inNumberCols; ++j) {
      for(unsigned int i=0; i<inNumberRows; ++i) {
        (*this)[(j*inNumberRows)+i] = ((i<mRows)&&(j<lCols)) ? lTmp[(j*mRows)+i] : inValue;
      }
    }
  }
  mRows = inNumberRows;
}


/*!
 *  \brief Set as an identity matrix of (inSize,inSize) dimension.
 *  \param inSize Number of rows and columns of the identity matrix.
 */
void Matrix::setIdentity(unsigned int inSize)
{
  resize(inSize, inSize);
  for(unsigned int j=0; j<inSize; ++j) {
    for(unsigned int i=0; i<inSize; ++i) (*this)(i,j) = (i==j) ? 1.0 : 0.0;
  }
}


/*!
 *  \brief Transpose matrix.
 */
void Matrix::transpose()
{
  Matrix lTmp(*this);
  mRows = getNumberColumns();
  const unsigned int lCols = getNumberColumns();
  for(unsigned int j=0; j<lCols; ++j) {
    for(unsigned int i=0; i<mRows; ++i) (*this)(i,j) = lTmp(j,i);
  }
}


/*!
 *  \brief Write matrix into XML output stream.
 *  \param ioStreamer XML output stream.
 */
void Matrix::write(XMLStreamer& ioStreamer) const
{
  std::ostringstream lOSS;
  const unsigned int lCols = getNumberColumns();
  for(unsigned int i=0; i<mRows; ++i) {
    for(unsigned int j=0; j<lCols; ++j) {
      lOSS << (*this)(i,j);
      if(j!=(lCols-1)) lOSS << ',';
    }
    if(i!=(mRows-1)) lOSS << ';';
  }
  ioStreamer.insertString(lOSS.str());
}


/*!
 *  \brief Back Substitution for L-U decomposition.
 *  \param inIndexes Vector of indexes.
 *  \param ioMatrixB B matrix.
 */
void Matrix::backSubstituteLU(const std::vector<unsigned int>& inIndexes,
                              Matrix& ioMatrixB) const
{
  const unsigned int lCols=getNumberColumns();
  unsigned int lII = UINT_MAX;
  for(unsigned int i=0; i<mRows; ++i) {
    double lSum = ioMatrixB(inIndexes[i],0);
    ioMatrixB(inIndexes[i],0) = ioMatrixB(i,0);
    if(lII != UINT_MAX) for(unsigned int j=lII; j<i; ++j) lSum -= (*this)(i,j) * ioMatrixB(j,0);
    else if(lSum!=0.0) lII = i;
    ioMatrixB(i,0) = lSum;
  }
  for(unsigned int i=0; i<mRows; ++i) {
    const unsigned int lR=(mRows-1)-i;
    double lSum = ioMatrixB(lR,0);
    for(unsigned int j=(lR+1); j<lCols; ++j) lSum -= (*this)(lR,j) * ioMatrixB(j,0);
    ioMatrixB(lR,0) = lSum / (*this)(lR,lR);
  }
}


/*!
 *  \brief  L-U decomposition.
 *  \param  outDecompose Matrix resulting from L-U decomposition.
 *  \param  outIndexes Vector of indexes.
 *  \param  outD D value.
 *  \return Resulting matrix after L-U decomposition.
 *  \throw  Beagle::RunTimeException If a singular matrix is used for L-U decomposition.
 */
void Matrix::decomposeLU(Matrix& outDecompose,
                         std::vector<unsigned int>& outIndexes,
                         int& outD) const
{
  const unsigned int lCols=getNumberColumns();
  outD = 1;
  outDecompose = *this;
  std::vector<double> lScales;
  scaleLU(lScales);
  for(unsigned int j=0; j<lCols; ++j) {
    for(unsigned int i=0; i<j; ++i) {
      double lSum = outDecompose(i,j);
      for(unsigned int k=0; k<i; ++k) lSum -= outDecompose(i,k) * outDecompose(k,j);
      outDecompose(i,j) = lSum;
    }
    double lMax = 0.0;
    unsigned int l;
    for(unsigned int i=j; i<mRows; ++i) {
      double lSum = outDecompose(i,j);
      for(unsigned int k=0; k<j; ++k) lSum -= outDecompose(i,k) * outDecompose(k,j);
      outDecompose(i,j) = lSum;
      double lDummy = lScales[i] * std::fabs(lSum);
      if(lDummy >= lMax) {
        l = i;
        lMax = lDummy;
      }
    }
    if(j != l) {
      for(unsigned int k=0; k<lCols; ++k) {
        double lTmp=outDecompose(l,k);
        outDecompose(l,k)=outDecompose(j,k);
        outDecompose(j,k)=lTmp;
      }
      outD = -outD;
      lScales[l] = lScales[j];
    }
    outIndexes[j] = l;
    if(outDecompose(j,j)==0.0) outDecompose(j,j) = 1e-20;
    if(j != (lCols-1)) {
      double lDummy = 1.0 / outDecompose(j,j);
      for(unsigned int i=(j+1); i<mRows; ++i) outDecompose(i,j) *= lDummy;
    }
  }
}


/*!
 *  \brief Get scaling value of matrix for L-U decomposition.
 *  \param outScales Resulting scaling values.
 *  \throw Beagle::RunTimeException If a singular matrix is used for L-U decomposition.
 */
void Matrix::scaleLU(std::vector<double>& outScales) const
{
  outScales.resize(getNumberColumns());
  for(unsigned int i=0; i<mRows; ++i) {
    double lMax=0.0;
    for(unsigned int j=0; j<outScales.size(); ++j) {
      const double lTmp=std::fabs((*this)(i,j));     
      if(lTmp>lMax) lMax=lTmp;
    }
    if(lMax==0.0) throw Beagle_RunTimeExceptionM("Singular matrix, can't do L-U decomposition!");
    outScales[i] = 1.0 / lMax;
  }
}


/*!
 *  \brief sqrt(a^2 + b^2) without under/overflow.
 *  \param a First value.
 *  \param b Second value.
 *  \return Hypothenuse of a and b.
 *  \note Used internally by method tql2.
 */
double Matrix::hypot(double a, double b) const
{
  double r;
  if(std::fabs(a) > std::fabs(b)) {
    r = b/a;
    r = std::fabs(a)*std::sqrt(1+r*r);
  }
  else if(b != 0) {
    r = a/b;
    r = std::fabs(b)*std::sqrt(1+r*r);
  }
  else {
    r = 0.0;
  }
  return r;
}

  
/*!
 *  \brief Triagonalize matrix for computing eigensystem using QL method.
 *  \param d Real part of eigenvalues computed from the matrix.
 *  \param e Imaginary part of eigenvalues computed from the matrix.
 *  \param V Eigenvectors computed from the matrix.
 *
 *  This method is derived from procedure tql2 of the Java package JAMA,
 *  which is itself derived from the Algol procedures tql2, by
 *  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
 *  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
 *  Fortran subroutine in EISPACK.
 */
void Matrix::tql2(Vector& d, Vector& e, Matrix& V) const
{
  const unsigned int n=mRows;
  for(unsigned int i=1; i<n; i++) {
    e[i-1] = e[i];
  }
  e[n-1] = 0.0;

  double f = 0.0;
  double tst1 = 0.0;
  double eps = std::pow(2.0,-52.0);
  for(unsigned int l=0; l<n; l++) {

    // Find small subdiagonal element
    tst1 = maxOf(tst1, std::abs(d[l]) + std::abs(e[l]));
    unsigned int m=l;
    while(m<n) {
      if(std::abs(e[m]) <= eps*tst1) {
        break;
      }
      m++;
    }

    // If m == l, d[l] is an eigenvalue,
    // otherwise, iterate.
    if(m > l) {
      unsigned int iter = 0;
      do {
        iter = iter + 1;  // (Could check iteration count here.)

        // Compute implicit shift
        double g = d[l];
        double p = (d[l+1] - g) / (2.0 * e[l]);
        double r = hypot(p,1.0);
        if(p < 0) {
          r = -r;
        }
        d[l] = e[l] / (p + r);
        d[l+1] = e[l] * (p + r);
        double dl1 = d[l+1];
        double h = g - d[l];
        for(unsigned int i=l+2; i<n; i++) {
          d[i] -= h;
        }
        f = f + h;

        // Implicit QL transformation.
        p = d[m];
        double c = 1.0;
        double c2 = c;
        double c3 = c;
        double el1 = e[l+1];
        double s = 0.0;
        double s2 = 0.0;
        for(unsigned int i=m-1; i>=l; i--) {
          c3 = c2;
          c2 = c;
          s2 = s;
          g = c * e[i];
          h = c * p;
          r = hypot(p,e[i]);
          e[i+1] = s * r;
          s = e[i] / r;
          c = p / r;
          p = c * d[i] - s * g;
          d[i+1] = h + s * (c * g + s * d[i]);

          // Accumulate transformation.
          for(unsigned int k=0; k<n; k++) {
            h = V(k,i+1);
            V(k,i+1) = s * V(k,i) + c * h;
            V(k,i) = c * V(k,i) - s * h;
          }
          
          if(i==0) break;
        }
        p = -s * s2 * c3 * el1 * e[l] / dl1;
        e[l] = s * p;
        d[l] = c * p;

        // Check for convergence.
      } while (std::abs(e[l]) > eps*tst1);
    }
    d[l] = d[l] + f;
    e[l] = 0.0;
  }
}


/*!
 *  \brief Diagonalize matrix for computing eigensystem using QL method.
 *  \param d Real part of eigenvalues computed from the matrix.
 *  \param e Imaginary part of eigenvalues computed from the matrix.
 *  \param V Eigenvectors computed from the matrix.
 *
 *  This method is derived from procedure tred2 of the Java package JAMA,
 *  which is itself derived from the Algol procedures tred2, by
 *  Bowdler, Martin, Reinsch, and Wilkinson, Handbook for
 *  Auto. Comp., Vol.ii-Linear Algebra, and the corresponding
 *  Fortran subroutine in EISPACK.
 */
void Matrix::tred2(Vector& d, Vector& e, Matrix& V) const
{
  const unsigned int n=mRows;
  V = *this;
  for(unsigned int j=0; j<n; ++j) {
    d[j] = V(n-1,j);
  }

  // Householder reduction to tridiagonal form.
  for(unsigned int i=(n-1); i>0; --i) {

    // Scale to avoid under/overflow.
    double scale = 0.0;
    double h = 0.0;
    for(unsigned int k=0; k<i; ++k) {
      scale += std::abs(d[k]);
    }
    if(scale == 0.0) {
      e[i] = d[i-1];
      for(unsigned int j=0; j<i; ++j) {
        d[j] = V(i-1,j);
        V(i,j) = 0.0;
        V(j,i) = 0.0;
      }
    }
    else {
      // Generate Householder vector.
      for(unsigned int k=0; k<i; ++k) {
        d[k] /= scale;
        h += d[k] * d[k];
      }
      double f = d[i-1];
      double g = std::sqrt(h);
      if(f > 0.0) {
        g = -g;
      }
      e[i] = scale * g;
      h = h - f * g;
      d[i-1] = f - g;
      for(unsigned int j=0; j<i; j++) {
        e[j] = 0.0;
      }

      // Apply similarity transformation to remaining columns.
      for(unsigned int j=0; j<i; j++) {
        f = d[j];
        V(j,i) = f;
        g = e[j] + V(j,j) * f;
        for(unsigned int k=j+1; k<=(i-1); k++) {
          g += V(k,j) * d[k];
          e[k] += V(k,j) * f;
        }
        e[j] = g;
      }
      f = 0.0;
      for(unsigned int j=0; j<i; j++) {
        e[j] /= h;
        f += e[j] * d[j];
      }
      double hh = f / (h + h);
      for(unsigned int j=0; j<i; j++) {
        e[j] -= hh * d[j];
      }
      for(unsigned int j=0; j<i; j++) {
        f = d[j];
        g = e[j];
        for(unsigned int k=j; k<=(i-1); k++) {
          V(k,j) -= (f * e[k] + g * d[k]);
        }
        d[j] = V(i-1,j);
        V(i,j) = 0.0;
      }
    }
    d[i] = h;
  }

  // Accumulate transformations.
  for(unsigned int i=0; i<(n-1); i++) {
    V(n-1,i) = V(i,i);
    V(i,i) = 1.0;
    double h = d[i+1];
    if(h!=0.0) {
      for(unsigned int k=0; k<=i; k++) {
        d[k] = V(k,i+1) / h;
      }
      for(unsigned int j=0; j<=i; j++) {
        double g = 0.0;
        for(unsigned int k=0; k<=i; k++) {
          g += V(k,i+1) * V(k,j);
        }
        for(unsigned int k=0; k<=i; k++) {
          V(k,j) -= g * d[k];
        }
      }
    }
    for(unsigned int k=0; k<=i; k++) {
      V(k,i+1) = 0.0;
    }
  }
  for(unsigned int j=0; j<n; j++) {
    d[j] = V(n-1,j);
    V(n-1,j) = 0.0;
  }
  V(n-1,n-1) = 1.0;
  e[0] = 0.0;
}

