/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/MilestoneReadOp.cpp
 *  \brief  Source code of class MilestoneReadOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <fstream>

#ifdef BEAGLE_HAVE_LIBZ
#include "gzstream.h"
#endif // BEAGLE_HAVE_LIBZ

using namespace Beagle;


/*!
 *  \brief Construct a reprise checking operator.
 *  \param inName Name of the operator.
 */
MilestoneReadOp::MilestoneReadOp(std::string inName) :
  Operator(inName),
  mMilestoneName(NULL)
{ }


/*!
 *  \brief Initialize the reprise checking operator.
 *  \param ioSystem System to use to initialize the operator.
 */
void MilestoneReadOp::initialize(System& ioSystem)
{
  if(ioSystem.getRegister().isRegistered("ms.restart.file")) {
    mMilestoneName =
      castHandleT<String>(ioSystem.getRegister().getEntry("ms.restart.file"));
  } else {
    mMilestoneName = new String("");
    std::string lMessage = "Name of the milestone file from which the evolution should ";
    lMessage += "be restarted. An empty string means no restart.";
    Register::Description lDescription(
      "Milestone restart filename",
      "String",
      "\"\"",
      lMessage
    );
    ioSystem.getRegister().addEntry("ms.restart.file", mMilestoneName, lDescription);
  }
}


/*!
 *  \brief Apply the read milestone operation.
 *  \param ioDeme Actual deme of the evolution.
 *  \param ioContext Context of the evolution.
 */
void MilestoneReadOp::operate(Deme& ioDeme, Context& ioContext)
{
  if(mMilestoneName->getWrappedValue().empty()) return;
  Beagle_LogBasicM(
    ioContext.getSystem().getLogger(),
    "milestone", "Beagle::MilestoneReadOp",
    std::string("Reading milestone file \"")+
    mMilestoneName->getWrappedValue()+"\""
  );
  readMilestone(mMilestoneName->getWrappedValue(), ioContext);
  if(ioContext.getDemeIndex() == (ioContext.getVivarium().size()-1)) {
    ioContext.setGeneration(ioContext.getGeneration()+1);
    ioContext.setDemeIndex(0);
  }
  else ioContext.setDemeIndex(ioContext.getDemeIndex()+1);
}


/*!
 *  \brief Read an evolution milestone.
 *  \param inFilename  Filename of the milestone.
 *  \param ioContext  Context of the evolution,
 *  \throw IOException If a read error occured.
 */
void MilestoneReadOp::readMilestone(std::string inFilename, Context& ioContext)
{ 
  XMLParser lParser;
  try {
#ifdef BEAGLE_HAVE_LIBZ
    igzstream lIGZStream(inFilename.c_str());
    lParser.parse(lIGZStream, inFilename.c_str());
#else // BEAGLE_HAVE_LIBZ
    std::ifstream lIFStream(inFilename.c_str());
    lParser.parse(lIFStream, inFilename.c_str());
#endif // BEAGLE_HAVE_LIBZ
  } catch(IOException& inExcept) {
    std::ostringstream lOSS;
    lOSS << "the evolution milestone file given is invalid: " << inExcept.what();
    throw Beagle_IOExceptionMessageM(lOSS.str());
  }
  for(XMLNode::Handle lNode=lParser.getRootNode();
      lNode!=NULL; lNode=lNode->getNextSibling()) {
    if((lNode->getType()==XMLNode::eTag) && (lNode->getTagName()=="Beagle")) {
      for(XMLNode::Handle lChild=lNode->getFirstChild();
          lChild!=NULL; lChild=lChild->getNextSibling()) {
        if(lChild->getType()==XMLNode::eTag) {
          if(lChild->getTagName() == "Milestone") {
            for(XMLNode::Handle lChild2=lChild->getFirstChild();
                lChild2!=NULL; lChild2=lChild2->getNextSibling()) {
              if(lChild2->getType()==XMLNode::eTag) {
                if(lChild2->getTagName() == "Generation") {
                  if(lChild2->getFirstChild()->getType() != XMLNode::eString) {
                    throw Beagle_IOExceptionNodeM(*lChild2, "generation value is not present!");
                  }
                  ioContext.setGeneration(str2uint(lChild2->getFirstChild()->getValue()));
                }
                else if(lChild2->getTagName() == "Deme") {
                  if(lChild2->getFirstChild()->getType() != XMLNode::eString) {
                    throw Beagle_IOExceptionNodeM(*lChild2, "deme index is not present!");
                  }
                  ioContext.setDemeIndex(str2uint(lChild2->getFirstChild()->getValue()));
                }
              }
            }
          }
          else if(lChild->getTagName() == "Evolver") {
            Beagle_LogTraceM(
              ioContext.getSystem().getLogger(),
              "milestone", "Beagle::MilestoneReadOp",
              "Reading evolver part of milestone"
            );
            ioContext.getEvolver().read(lChild);
          }
          else if(lChild->getTagName() == "Register") {
            Beagle_LogTraceM(
              ioContext.getSystem().getLogger(),
              "milestone", "Beagle::MilestoneReadOp",
              "Reading register part of milestone"
            );
            ioContext.getSystem().getRegister().readWithSystem(lChild, ioContext.getSystem());
          }
          else if(lChild->getTagName() == "Vivarium") {
            Beagle_LogTraceM(
              ioContext.getSystem().getLogger(),
              "milestone", "Beagle::MilestoneReadOp",
              "Reading population part of milestone"
            );
            ioContext.getVivarium().readWithContext(lChild, ioContext);
          }
        }
      }   
    }
  }
}

