/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/NSGA2Op.cpp
 *  \brief  Source code of class NSGA2Op.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/10/26 19:04:58 $
 */

#include "beagle/Beagle.hpp"

#include <cfloat>
#include <float.h>     // To get FLT_MAX on Borland C++ Builder
#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Build NSGA2 replacement strategy operator.
 *  \param inName Name of the NSGA2 operator.
 */
NSGA2Op::NSGA2Op(std::string inName) :
  ReplacementStrategyOp(inName)
{ }


/*!
 *  \brief Apply NSGA2 multiobjective selection operator as a replacement strategy.
 *  \param ioDeme Deme on which selection operator is applied.
 *  \param ioContext Evolutionary context.
 */
void NSGA2Op::applyAsReplacementStrategy(Deme& ioDeme, Context& ioContext)
{
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    std::string("Processing using NSGA2 replacement strategy the ")+
    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
  );
  Beagle_LogObjectM(
    ioContext.getSystem().getLogger(),
    Logger::eTrace,
    "replacement-strategy", "Beagle::GenerationalOp",
    (*this)
  );

  // Generate a new generation of individuals, merged with the actual one.
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    std::string("Generating ")+uint2str(ioDeme.size())+" offsprings using breeder tree"
  );
  std::vector<double> lRoulette;
  buildRoulette(lRoulette, ioContext);
  Individual::Bag lOffsprings(ioDeme);
  for(unsigned int i=0; i<ioDeme.size(); ++i) {
    unsigned int lIndexBreeder = rollRoulette(lRoulette, ioContext);
    BreederNode::Handle lSelectedBreeder=getRootNode();
    for(unsigned int j=0; j<lIndexBreeder; ++j)
      lSelectedBreeder=lSelectedBreeder->getNextSibbling();
    Beagle_NonNullPointerAssertM(lSelectedBreeder);
    Beagle_NonNullPointerAssertM(lSelectedBreeder->getBreederOp());
    Individual::Handle lBredIndiv =
      lSelectedBreeder->getBreederOp()->breed(ioDeme, lSelectedBreeder->getFirstChild(), ioContext);
    Beagle_NonNullPointerAssertM(lBredIndiv);
    lOffsprings.push_back(lBredIndiv);
  }

  // Fast non-dominated sorting, followed by insertion of the first Pareto fronts.
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    "Applying fast non-dominated sorting on the union of parents and offsprings individual"
  );
  std::vector< std::vector<unsigned int> > lParetoFronts;
  sortFastND(lParetoFronts, ioDeme.size(), lOffsprings, ioContext);
  unsigned int lIndexDeme=0;
  for(unsigned int j=0; j<(lParetoFronts.size()-1); ++j) {
    for(unsigned int k=0; k<lParetoFronts[j].size(); ++k) {
      ioDeme[lIndexDeme++] = lOffsprings[lParetoFronts[j][k]];
    }
  }

  // Insertion of the last Pareto front, using crowding distance
  Individual::Bag lLastFrontIndiv;
  for(unsigned int l=0; l<lParetoFronts.back().size(); ++l) {
    lLastFrontIndiv.push_back(lOffsprings[lParetoFronts.back()[l]]);
  }
  std::vector< std::pair<double,unsigned int> > lDistances;
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    std::string("Computing crowding distance on the ")+uint2ordinal(lParetoFronts.size())+
    std::string(" Pareto front, which is made of ")+uint2str(lParetoFronts.back().size())+
    " individuals"
  );
  evalCrowdingDistance(lDistances, lLastFrontIndiv);
  for(unsigned int m=0; lIndexDeme<ioDeme.size(); ++m) {
    ioDeme[lIndexDeme++] = lLastFrontIndiv[lDistances[m].second];
  }
}


/*!
 *  \brief Apply NSGA2 multiobjective selection operator as a standard operator.
 *  \param ioDeme Deme on which selection operator is applied.
 *  \param ioContext Evolutionary context.
 */
void NSGA2Op::applyAsStandardOperator(Deme& ioDeme, Context& ioContext)
{
  // Fast non-dominated sorting, followed by insertion of the first Pareto fronts.
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    "Applying fast non-dominated sorting on the whole population"
  );
  std::vector< std::vector<unsigned int> > lParetoFronts;
  const unsigned int lDesiredPopSize = (*mPopSize)[ioContext.getDemeIndex()];
  Individual::Bag lSortedPop(ioDeme);
  sortFastND(lParetoFronts, lDesiredPopSize, lSortedPop, ioContext);
  unsigned int lIndexDeme=0;
  for(unsigned int j=0; j<(lParetoFronts.size()-1); ++j) {
    for(unsigned int k=0; k<lParetoFronts[j].size(); ++k) {
      ioDeme[lIndexDeme++] = lSortedPop[lParetoFronts[j][k]];
    }
  }

  // Insertion of the last Pareto front, using crowding distance
  Individual::Bag lLastFrontIndiv;
  for(unsigned int l=0; l<lParetoFronts.back().size(); ++l) {
    lLastFrontIndiv.push_back(lSortedPop[lParetoFronts.back()[l]]);
  }
  std::vector< std::pair<double,unsigned int> > lDistances;
  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    std::string("Computing crowding distance on the ")+uint2ordinal(lParetoFronts.size())+
    std::string(" Pareto front, which is made of ")+uint2str(lParetoFronts.back().size())+
    " individuals"
  );
  evalCrowdingDistance(lDistances, lLastFrontIndiv);
  for(unsigned int m=0; lIndexDeme<lDesiredPopSize; ++m) {
    ioDeme[lIndexDeme++] = lLastFrontIndiv[lDistances[m].second];
  }
  ioDeme.resize(lDesiredPopSize);
}


/*!
 *  \brief Evaluate crowding distance of a pool of individuals.
 *  \param outDistances Evaluated crowding distance and associated indexes.
 *  \param inIndividualPool Pool of individuals to evaluate distance on.
 */
void NSGA2Op::evalCrowdingDistance(std::vector< std::pair<double,unsigned int> >& outDistances,
                                   const Individual::Bag& inIndividualPool) const
{
  outDistances.clear();
  if(inIndividualPool.size() == 0) return;
  outDistances.resize(inIndividualPool.size());
  std::vector< CrowdingPair > lCrowdingSet(inIndividualPool.size());
  for(unsigned int i=0; i<inIndividualPool.size(); ++i) {
    outDistances[i].first = 0.0;
    outDistances[i].second = i;
    lCrowdingSet[i].first = i;
    lCrowdingSet[i].second = castHandleT<FitnessMultiObj>(inIndividualPool[i]->getFitness());
  }
  unsigned int lNumberObjectives = lCrowdingSet.back().second->size();
  std::vector<float> lScalingFactors = lCrowdingSet.back().second->getScalingFactors();
  for(unsigned int j=0; j<lNumberObjectives; ++j) {
    std::sort(lCrowdingSet.begin(), lCrowdingSet.end(), IsLessCrowdingPairPredicate(j));
    outDistances[lCrowdingSet.front().first].first = DBL_MAX;
    outDistances[lCrowdingSet.back().first].first  = DBL_MAX;
    for(unsigned int k=1; k<(lCrowdingSet.size()-1); ++k) {
      if(outDistances[lCrowdingSet[k].first].first < DBL_MAX) {
        double lNeighborDist = (*lCrowdingSet[k+1].second)[j] - (*lCrowdingSet[k-1].second)[j];
        outDistances[lCrowdingSet[k].first].first += (lNeighborDist * lScalingFactors[j]);
      }
    }
  }
  std::sort(outDistances.begin(), outDistances.end(),
            std::greater< std::pair<double,unsigned int> >());
}

/*!
 *  \brief Initialize the NSGA2 operator.
 *  \param ioSystem Reference to the system to use for the initialization.
 */
void NSGA2Op::initialize(System& ioSystem)
{
  ReplacementStrategyOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered("ec.pop.size")) {
    mPopSize = castHandleT<IntegerVector>(ioSystem.getRegister().getEntry("ec.pop.size"));
  } else {
    mPopSize = new IntegerVector(1,100);
    std::string lLongDescrip("Number of demes and size of each deme of the population. ");
    lLongDescrip += "The format of an IntegerVector is S1/S2/.../Sn, where Si is the ith value. ";
    lLongDescrip += "The size of the IntegerVector is the number of demes present in the ";
    lLongDescrip += "vivarium, while each value of the vector is the size of the corresponding ";
    lLongDescrip += "deme.";
    Register::Description lDescription(
      "Vivarium and demes sizes",
      "IntegerVector",
      "100",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("ec.pop.size", mPopSize, lDescription);
  }
}


/*!
 *  \brief Apply the NSGA2 multiobjective selection operation on a deme in the given context.
 *  \param ioDeme Reference to the deme on which the operation takes place.
 *  \param ioContext Evolutionary context of the operation.
 */
void NSGA2Op::operate(Deme& ioDeme, Context& ioContext)
{
  if(getRootNode()==NULL) applyAsStandardOperator(ioDeme,ioContext);
  else applyAsReplacementStrategy(ioDeme,ioContext);
}


/*!
 *  \brief Post-initialize hook for the NSGA operator.
 *  \param ioSystem Evolutionary system.
 */
void NSGA2Op::postInit(System& ioSystem)
{
  if(ioSystem.getLogger().getLogLevel() >= Logger::eInfo) {
    if(ioSystem.getRegister().isRegistered("ec.hof.vivasize")) {
      UInt::Handle lVivaHOFSize =
        castHandleT<UInt>(ioSystem.getRegister().getEntry("ec.hof.vivasize"));
      if(lVivaHOFSize->getWrappedValue() > 0) {
        Beagle_LogInfoM(
          ioSystem.getLogger(),
          "replacement-strategy", "Beagle::NSGA2Op",
          std::string("Warning: the vivarium hall-of-fame size (parameter \"ec.hof.vivasize\") ")+
          std::string("is non-zero; the hall-of-fame is not meaningful in a multiobjective ")+
          std::string("EA context")
        );
      }
    }
    if(ioSystem.getRegister().isRegistered("ec.hof.demesize")) {
      UInt::Handle lVivaHOFSize =
        castHandleT<UInt>(ioSystem.getRegister().getEntry("ec.hof.demesize"));
      if(lVivaHOFSize->getWrappedValue() > 0) {
        Beagle_LogInfoM(
          ioSystem.getLogger(),
          "replacement-strategy", "Beagle::NSGA2Op",
          std::string("Warning: the demes hall-of-fame size (parameter \"ec.hof.demesize\") ")+
          std::string("is non-zero; the hall-of-fame is not meaningful in a multiobjective ")+
          std::string("EA context")
        );
      }
    }
  }
}


/*!
 *  \brief Do fast non-dominated sort and evaluate Pareto fronts of data up to given size.
 *  \param outParetoFronts Pareto fronts from the fast ND sort.
 *  \param inSortStop Number of individuals on the Pareto fronts needed to stop the sort.
 *  \param inIndividualPool Pool of individuals to get next Pareto front of.
 *  \param ioContext Evolutionary context.
 */
void NSGA2Op::sortFastND(std::vector< std::vector<unsigned int> >& outParetoFronts,
                         unsigned int inSortStop,
                         const Individual::Bag& inIndividualPool,
                         Context& ioContext) const
{
  outParetoFronts.clear();
  if(inSortStop == 0) return;
  outParetoFronts.resize(1);
  unsigned int lParetoSorted = 0;
  // N: Number of dominating individuals
  std::vector<unsigned int> lN(inIndividualPool.size(), 0);
  // S: Set of dominated individuals
  std::vector< std::vector<unsigned int> > lS(inIndividualPool.size());

  // First pass to get first Pareto front and domination sets
  for(unsigned int i=0; i<inIndividualPool.size(); ++i) {
    FitnessMultiObj::Handle lFitI = castHandleT<FitnessMultiObj>(inIndividualPool[i]->getFitness());
    for(unsigned int j=(i+1); j<inIndividualPool.size(); ++j) {
      FitnessMultiObj::Handle lFitJ =
        castHandleT<FitnessMultiObj>(inIndividualPool[j]->getFitness());
      if(lFitJ->isDominated(*lFitI)) {
        lS[i].push_back(j);  // Add index j to dominated set of i
        ++lN[j];             // Increment domination counter of j
      }
      else if(lFitI->isDominated(*lFitJ)) {
        lS[j].push_back(i);  // Add index i to dominated set of j
        ++lN[i];             // Increment domination counter of i
      }
    }
    if(lN[i] == 0) {         // If i is non-dominated
      outParetoFronts.back().push_back(i);
      ++lParetoSorted;
    }
  }

  Beagle_LogVerboseM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::NSGA2Op",
    uint2ordinal(1)+std::string(" Pareto front consists of")+
    uint2str(outParetoFronts.back().size())+" individuals"
  );
  
  // Continue ranking until individual pool is empty or stop criterion reach
  while((lParetoSorted < inIndividualPool.size()) && (lParetoSorted < inSortStop)) {
    unsigned int lIndexLastFront = outParetoFronts.size() - 1;
    outParetoFronts.resize(outParetoFronts.size() + 1);
    for(unsigned int k=0; k<outParetoFronts[lIndexLastFront].size(); ++k) {
      unsigned int lIndexPk = outParetoFronts[lIndexLastFront][k];
      for(unsigned int l=0; l<lS[lIndexPk].size(); ++l) {
        unsigned int lIndexSl = lS[lIndexPk][l];
        if(--lN[lIndexSl] == 0) {
          outParetoFronts.back().push_back(lIndexSl);
          ++lParetoSorted;
        }
      }
    }
    Beagle_LogVerboseM(
      ioContext.getSystem().getLogger(),
      "replacement-strategy", "Beagle::NSGA2Op",
      uint2ordinal(outParetoFronts.size())+std::string(" Pareto front consists of")+
      uint2str(outParetoFronts.back().size())+" individuals"
    );
  }
}


