/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Operator.cpp
 *  \brief  Source code of class Operator.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <string>

using namespace Beagle;


/*!
 *  \brief Construct a new operator.
 *  \param inName Name of the operator.
 */
Operator::Operator(std::string inName) :
  mName(inName),
  mInitialized(false),
  mPostInitialized(false)
{ }


/*!
 *  \brief Give a reference to the current operator.
 *  \return Handle to the refered operator.
 */
PointerT<Operator,Object::Handle> Operator::giveReference()
{
  return Operator::Handle(this);
}


/*!
 *  \brief Reading a operator with method read is undefined. Use readWithMap instead.
 */
void Operator::read(XMLNode::Handle&)
{
  throw Beagle_UndefinedMethodInternalExceptionM("read","Operator");
}


/*!
 *  \brief Read an operator from an input stream.
 *  \param inNode XML node from which the operator is read.
 *  \param inOpMap Operator map to use to read operator.
 *  \throw IOException If a reading error occurs.
 */
void Operator::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != mName)) {
    std::ostringstream lOSS;
    lOSS << "tag <" << mName << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
}


/*!
 *  \brief Write an operator into an output stream.
 *  \param ioStreamer Output stream in which the operator is written.
 */
void Operator::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag(mName);
  ioStreamer.closeTag(false);
}

