/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/OversizeOp.cpp
 *  \brief  Source code of class OversizeOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.3 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <cmath>
#include <algorithm>

using namespace std;
using namespace Beagle;


/*!
 *  \brief Build oversize replacement strategy operator.
 *  \param inOversizeRatioName Oversize ratio parameter name used in the register.
 *  \param inName Name of the oversize operator.
 */
OversizeOp::OversizeOp(std::string inOversizeRatioName, std::string inName) :
  ReplacementStrategyOp(inName),
  mOversizeRatioName(inOversizeRatioName)
{ }


/*!
 *  \brief Initialize the operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void OversizeOp::initialize(System& ioSystem)
{
  ReplacementStrategyOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered(mOversizeRatioName)) {
    mOversizeRatio = castHandleT<Float>(ioSystem.getRegister().getEntry(mOversizeRatioName));
  } else {
    mOversizeRatio = new Float((float)7.0);
    Register::Description lDescription(
      "Oversizing ratio",
      "Float",
      "7.0",
      std::string("Oversizing ratio, which mean how much bigger is the child population from ") +
      std::string("the parent population.")
    );
    ioSystem.getRegister().addEntry(mOversizeRatioName, mOversizeRatio, lDescription);
  }
}


/*!
 *  \brief Apply the oversize replacement strategy operation on a deme.
 *  \param ioDeme Reference to the deme on which the operation takes place.
 *  \param ioContext Evolutionary context of the operation.
 */
void OversizeOp::operate(Deme& ioDeme, Context& ioContext)
{
  Beagle_NonNullPointerAssertM(getRootNode());
  Beagle_ValidateParameterM(mOversizeRatio->getWrappedValue() >= 1.0,
                            mOversizeRatioName,
                            "The oversize ratio must be higher or equal to 1.0.");

  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::OversizeOp",
    std::string("Using oversize replacement strategy to process the ")+
    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
  );
  Beagle_LogObjectM(
    ioContext.getSystem().getLogger(),
    Logger::eTrace,
    "replacement-strategy", "Beagle::OversizeOp",
    (*this)
  );

  std::vector<double> lRoulette;
  buildRoulette(lRoulette, ioContext);

  unsigned int lLambda =
    (unsigned int)ceil(mOversizeRatio->getWrappedValue()*float(ioDeme.size()));
  Individual::Bag lOffsprings;
  for(unsigned int i=0; i<lLambda; ++i) {
    unsigned int lIndexBreeder = rollRoulette(lRoulette, ioContext);
    BreederNode::Handle lSelectedBreeder=getRootNode();
    for(unsigned int j=0; j<lIndexBreeder; ++j)
      lSelectedBreeder=lSelectedBreeder->getNextSibbling();
    Beagle_NonNullPointerAssertM(lSelectedBreeder);
    Beagle_NonNullPointerAssertM(lSelectedBreeder->getBreederOp());
    Individual::Handle lBredIndiv =
      lSelectedBreeder->getBreederOp()->breed(ioDeme, lSelectedBreeder->getFirstChild(), ioContext);
    Beagle_NonNullPointerAssertM(lBredIndiv);
    lOffsprings.push_back(lBredIndiv);
  }
  ioDeme.insert(ioDeme.end(), lOffsprings.begin(), lOffsprings.end());
}


/*!
 *  \brief Read a OversizeOp operator for XML tree.
 *  \param inNode XML node to use to read OversizeOp operator.
 *  \param inOpMap Operator map to use to read OversizeOp operator.
 */
void OversizeOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string lOversizeRatioReadName = inNode->getTagAttribute("ratio_name");
  if(lOversizeRatioReadName.empty() == false) mOversizeRatioName = lOversizeRatioReadName;

  ReplacementStrategyOp::readWithMap(inNode, inOpMap);
}


/*!
 *  \brief Write OversizeOp operator into XML streamer.
 *  \param ioStreamer XML streamer to write OversizeOp operator into.
 */
void OversizeOp::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag(getName());
  ioStreamer.insertAttribute("ratio_name", mOversizeRatioName);
  if(getRootNode() != NULL) getRootNode()->write(ioStreamer);
  ioStreamer.closeTag();
}




