/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */
 
/*!
 *  \file   beagle/src/Randomizer.cpp
 *  \brief  Source code of class Randomizer.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/10/26 16:54:00 $
 */

#include "beagle/Beagle.hpp"

#include <cmath>

using namespace std;
using namespace Beagle;


/*!
 *  \brief Construct a randomizer.
 */
Randomizer::Randomizer() :
  mSeed(NULL),
  mState(NULL)
{ }


/*!
 *  \brief Initialize the randomizer. The initial seed value is the timer.
 *  \param ioSystem Reference to the system.
 */
void Randomizer::initialize(System& ioSystem)
{
  Beagle_LogDetailedM(
    ioSystem.getLogger(),
    "random", "Beagle::Randomizer",
    "Initializing randomizer"
  );
  
 if(ioSystem.getRegister().isRegistered("ec.rand.seed")) {
    mSeed = castHandleT<ULong>(ioSystem.getRegister().getEntry("ec.rand.seed"));
  } else {
    mSeed = new ULong(0);
    std::string lLongDescrip = "Randomizer seed. A zero value means that ";
    lLongDescrip += "the seed should be initialized using the current system time.";
    Register::Description lDescription(
      "Randomizer seed",
      "ULong",
      "0",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("ec.rand.seed", mSeed, lDescription);
  }

  if(ioSystem.getRegister().isRegistered("ec.rand.state")) {
    mState = castHandleT<ULong>(ioSystem.getRegister().getEntry("ec.rand.state"));
  } else {
    mState = new ULong(0);
    std::string lLongDescrip = "Actual randomizer internal state. ";
    lLongDescrip += "The state changes at every function call to the random number generator. ";
    lLongDescrip += "This parameter is useful to get the correct randomizer state when an ";
    lLongDescrip += "evolution is restarted from a milestone. The state must be set to 0 ";
    lLongDescrip += "before starting a new evolution.";
    Register::Description lDescription(
      "Randomizer internal state",
      "ULong",
      "0",
      lLongDescrip
    );
    ioSystem.getRegister().addEntry("ec.rand.state", mState, lDescription);
  }
}


/*!
 *  \brief Post-initializing the randomizer.
 *  \param ioSystem Evolutionary system.
 */
void Randomizer::postInit(System& ioSystem)
{
  Beagle_LogDetailedM(
    ioSystem.getLogger(),
    "random", "Beagle::Randomizer",
    "Post-initializing randomizer"
  );
  initRandomizer(0);
}


/*!
 *  \brief Generate a floating-point number following a Gaussian distribution.
 *  \param inMu Mean of the Gaussain distribution.
 *  \param inSigma Standard-error of the Gaussian distribution.
 */
double Randomizer::rollGaussian(double inMu, double inSigma)
{
  double lFact,lRSQ,lV1,lV2;
  do {
    lV1  = rollUniform(-1.0, 1.0);
    lV2  = rollUniform(-1.0, 1.0);
    lRSQ = (lV1*lV1) + (lV2*lV2);
  } while(lRSQ >= 1.0 || lRSQ == 0.0);
  lFact = sqrt(-2.0 * log(lRSQ) / lRSQ);
  return lV2*lFact*inSigma + inMu;
}
