/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*
 *  Part of this file include code from file mt19937ar.c, licensed under
 *  a BSD license. This code is available on-line at
 *  http://www.math.keio.ac.jp/~matumoto/MT2002/emt19937ar.html
 *
 *  A C-program for MT19937, with initialization improved 2002/1/26.
 *  Coded by Takuji Nishimura and Makoto Matsumoto.
 *
 *  Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
 *  All rights reserved.
 *
 *  Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 *
 *    1. Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *
 *    2. Redistributions in binary form must reproduce the above copyright
 *       notice, this list of conditions and the following disclaimer in the
 *       documentation and/or other materials provided with the distribution.
 *
 *    3. The names of its contributors may not be used to endorse or promote
 *       products derived from this software without specific prior written
 *       permission.
 *
 *  THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 *  "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 *  LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 *  A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 *  CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 *  EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 *  PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 *  PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
 *  LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
 *  NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 *  SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *  Any feedback is very welcome.
 *  http://www.math.keio.ac.jp/matumoto/emt.html
 *  email: matumoto@math.keio.ac.jp
 *
 */

/*!
 *  \file   beagle/src/RandomizerMersenne.cpp
 *  \brief  Source code of class RandomizerMersenne.
 *  \author Makoto Matsumoto
 *  \author Takuji Nishimura
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/07/27 20:41:39 $
 */

#include "beagle/Beagle.hpp"

#include <ctime>

using namespace Beagle;


/*!
 *  \brief Construct a mersenne twister randomizer.
 */
RandomizerMersenne::RandomizerMersenne() :
  mValidator(0),
  mNextMT(NULL),
  mLeftMT(0)
{
  for(unsigned int i=0; i<N; ++i) mStateMT[i] = 0;
}


/*!
 *  \brief Hash a seed from time and clock values.
 *  \param inTime Time value.
 *  \param inClock Clock value.
 *  \return Hash value.
 */
unsigned long RandomizerMersenne::hash(unsigned int inTime, unsigned int inClock)
{
  // Get a uint32 from t and c
  // Better than uint32(x) in case x is floating point in [0,1]
  // Based on code by Lawrence Kirby (fred@genesis.demon.co.uk)
  
  static unsigned long lDiffer = 0;  // guarantee time-based seeds will change
  unsigned long lH1 = 0;
  unsigned char *lP = (unsigned char*)&inTime;
  for(size_t i = 0; i<sizeof(inTime); ++i) {
    lH1 *= UCHAR_MAX + 2U;
    lH1 += lP[i];
  }
  unsigned long lH2 = 0;
  lP = (unsigned char*)&inClock;
  for(size_t j = 0; j<sizeof(inClock); ++j) {
    lH2 *= UCHAR_MAX + 2U;
    lH2 += lP[j];
  }
  return (lH1 + lDiffer++) ^ lH2;
}


/*!
 *  \brief Initialize the randomizer state.
 *  \param inState Value of the state. If 0, set to seed value (if seed 0, set to timer value).
 *  \note Based on function init_genrand of mt19937ar.c.
 */
void RandomizerMersenne::initRandomizer(unsigned long inState)
{
  inState &= 0xffffffffUL;
  if(inState == 0) {
    if(mSeed->getWrappedValue() == 0) {
#ifdef BEAGLE_HAVE_DEV_URANDOM
      unsigned long lReadSeed = 0;
      std::ifstream lDevURandomIFS("/dev/urandom", std::iostream::binary);
      if(lDevURandomIFS.good()) lDevURandomIFS.read((char*)&lReadSeed, sizeof(unsigned long));
      if(lDevURandomIFS.good() == false) lReadSeed = 0;
      lDevURandomIFS.close();
      if(lReadSeed==0) mStateMT[0] = *mSeed = (hash(std::time(NULL),std::clock()) & 0xffffffffUL);
      else mStateMT[0] = *mSeed = (lReadSeed & 0xffffffffUL);
#else // BEAGLE_HAVE_DEV_URANDOM
      mStateMT[0] = *mSeed = (hash(std::time(NULL),std::clock()) & 0xffffffffUL);
#endif // BEAGLE_HAVE_DEV_URANDOM
    }
    else mStateMT[0] = (mSeed->getWrappedValue() & 0xffffffffUL);
  }
  else mStateMT[0] = inState;
  
  // Initialize generator state with seed
  // See Knuth TAOCP Vol 2, 3rd Ed, p.106 for multiplier.
  // In previous versions, most significant bits (MSBs) of the seed affect
  // only MSBs of the state array.  Modified 9 Jan 2002 by Makoto Matsumoto.
  unsigned long *lS = mStateMT+1;
  unsigned long *lR = mStateMT;
  for(int i=1; i<N; ++i) {
    *lS++ = ( 1812433253UL * ( *lR ^ (*lR >> 30) ) + i ) & 0xffffffffUL;
    lR++;
  }
  reload();
}


/*!
 *  \brief Read the state of the randomizer from a XML node.
 *  \param inNode XML node to read the mersenne twister randomizer from.
 */
void RandomizerMersenne::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Randomizer"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Randomizer> expected!");
  
  std::string lTypeStr = inNode->getTagAttribute("type");
  if(lTypeStr != "Mersenne")
    throw Beagle_IOExceptionNodeM(*inNode,"type of randomizer is not Mersenne!");

  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild!=NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Seed") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL)
          throw Beagle_IOExceptionNodeM(*lChild,"missing value of randomizer seed!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing value of randomizer seed!");
        *mSeed = str2int(lChild2->getValue());
      }
      else if(lChild->getTagName() == "State") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL)
          throw Beagle_IOExceptionNodeM(*lChild,"missing value of randomizer state!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing value of randomizer state!");
        *mState = str2int(lChild2->getValue());
      }
      else if(lChild->getTagName() == "StateMT") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL) throw Beagle_IOExceptionNodeM(*lChild,"missing StateMT array!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing StateMT array!");
        std::istringstream lISS(lChild2->getValue());
        char lTmp = '\0';
        for(int i=0; i<(N-1); ++i) {
          lISS >> mStateMT[i] >> lTmp;
          if(lTmp != '/')
            throw Beagle_IOExceptionNodeM(*lChild2, "bad delimiter in StateMT enumeration!");
        }
        lISS >> mStateMT[N-1];
      }
      else if(lChild->getTagName() == "NextMT") {
        XMLNode::Handle lChild2 = lChild->getFirstChild();
        if(lChild2 == NULL) throw Beagle_IOExceptionNodeM(*lChild,"missing NextMT index!");
        if(lChild2->getType() != XMLNode::eString)
          throw Beagle_IOExceptionNodeM(*lChild2,"missing NextMT index!");
        mNextMT = mStateMT + str2int(lChild2->getValue());
      }
    }
  }
}


/*!
 *  \brief Reload the mersenne twister with new set of numbers.
 */
void RandomizerMersenne::reload()
{
  // Generate N new values in state
  // Made clearer and faster by Matthew Bellew (matthew.bellew@home.com)
  unsigned long *lP = mStateMT;
  for(int i=(N-M); i--; ++lP) *lP = twist(lP[M], lP[0], lP[1]);
  for(int i=M; --i; ++lP)     *lP = twist(lP[M-N], lP[0], lP[1]);
  *lP = twist(lP[M-N], lP[0], mStateMT[0]);
  mLeftMT = N;
  mNextMT = mStateMT;
  mValidator = mState->getWrappedValue() = *mNextMT;
}


/*!
 *  \brief Generate an positive integer following an uniform discrete distribution.
 *  \return Random positive integer.
 */
unsigned long RandomizerMersenne::rollInt()
{
  // Check if randomizer is still valid.
  if((mValidator == 0) || (mValidator != mState->getWrappedValue())) {
    initRandomizer(mState->getWrappedValue());
  }

  // Pull a 32-bit integer from the generator state.
  // Every other access function simply transforms the numbers extracted here.
  if(mLeftMT==0) reload();
  --mLeftMT;

  unsigned long lS = mValidator = mState->getWrappedValue() = *mNextMT++;
  lS ^= (lS >> 11);
  lS ^= (lS <<  7) & 0x9d2c5680UL;
  lS ^= (lS << 15) & 0xefc60000UL;
  lS ^= (lS >> 18);
  return lS;
}


/*!
 *  \brief Generate an positive integer following an uniform discrete distribution.
 *  \param inLower Lower bound of the distribution.
 *  \param inUpper Upper bound of the distribution.
 *  \return Random positive integer generated in [inLower, inUpper].
 */
unsigned long RandomizerMersenne::rollInteger(unsigned long inLower, unsigned long inUpper)
{
  Beagle_AssertM(inLower <= inUpper);
  if((inLower==0) && (inUpper==ULONG_MAX)) return rollInt();  
  if(inLower == inUpper) return inLower;
  
  // Find which bits are used in n
  // Optimized by Magnus Jonsson (magnus@smartelectronix.com)
  unsigned long lUsed = (inUpper - inLower);
  lUsed |= lUsed >> 1;
  lUsed |= lUsed >> 2;
  lUsed |= lUsed >> 4;
  lUsed |= lUsed >> 8;
  lUsed |= lUsed >> 16;

  // Draw numbers until one is found in [0,n]
  unsigned long lX;
  do {
    lX = rollInt() & lUsed;  // toss unused bits to shorten search
  } while(lX > (inUpper - inLower));

  return inLower + lX;
}

 
/*!
 *  \brief Generate a number following a uniform distribution in the interval [inLower,inUpper).
 *  \param inLower Lower bound of the uniform distribution.
 *  \param inUpper Upper bound of the uniform distribution.
 *  \return Random number generated in [inLower,inUpper).
 */
double RandomizerMersenne::rollUniform(double inLower, double inUpper)
{
  Beagle_AssertM(inLower <= inUpper);
  unsigned long lY = rollInt();   // lY in [0,2^32].
  return (double(lY) * (inUpper-inLower) / 4294967296.0) + inLower;
}

/*!
 *  \brief Write the state of the randomizer into a Beagle XML streamer.
 *  \param ioStreamer XML streamer to write the randomizer into.
 */
void RandomizerMersenne::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Randomizer");
  ioStreamer.insertAttribute("type","Mersenne");
  ioStreamer.insertInteger(*mSeed, "Seed");
  ioStreamer.insertInteger(*mState, "State");
  std::ostringstream lOSS;
  for(int i=0; i<(N-1); ++i) {
    lOSS << mStateMT[i] << '/';
  }
  lOSS << mStateMT[N-1];
  ioStreamer.insertString(lOSS.str(), "StateMT");
  ioStreamer.insertInteger((mNextMT-mStateMT), "NextMT");
  ioStreamer.closeTag();
}

