/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Register.cpp
 *  \brief  Source code of class Register.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.10 $
 *  $Date: 2004/07/04 21:45:31 $
 */

#include "beagle/Beagle.hpp"

#include <cstdio>
#include <cstring>
#include <algorithm>
#include <fstream>
#include <iostream>

#ifdef BEAGLE_HAVE_LIBZ
#include "gzstream.h"
#endif // BEAGLE_HAVE_LIBZ

using namespace Beagle;


/*!
 *  \brief Default constructor of Register::Description.
 */
Register::Description::Description() :
  mBrief("[No brief]"),
  mType(""),
  mDefaultValue(""),
  mDescription("[No description]")
{ }


/*!
 *  \brief Construct a Register::Description instance.
 *  \param inBrief Brief desccription of the parameter.
 *  \param inType Type of the parameter contained.
 *  \param inDefaultValue Default value of the parameter.
 *  \param inDescription Long description of the parameter.
 */
Register::Description::Description(std::string inBrief,
                                   std::string inType,
                                   std::string inDefaultValue,
                                   std::string inDescription) :
  mBrief(inBrief),
  mType(inType),
  mDefaultValue(inDefaultValue),
  mDescription(inDescription)
{ }


/*!
 *  \brief Add a new description in the register.
 *  \param inTag Tag of the entry.
 *  \param inDescription Description of the entry.
 *  \throw RunTimeException If the entry is already described.
 */
void Register::addDescription(std::string inTag, const Description& inDescription)
{
  std::map<std::string, Description>::const_iterator lIterDescrip = mDescriptions.find(inTag);
  if(lIterDescrip != mDescriptions.end()) {
    std::string lMessage = "The entry \"";
    lMessage += inTag;
    lMessage += "\" is already described!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  mDescriptions[inTag] = inDescription;
}


/*!
 *  \brief Add a new entry in the register.
 *  \param inTag Tag of the entry.
 *  \param inEntry Value of the entry.
 *  \param inDescription Description of the entry.
 *  \throw RunTimeException If the entry is already in the register.
 */
void Register::addEntry(std::string inTag, Object::Handle inEntry,
                        const Register::Description& inDescription)
{
  const_iterator lIterObj = Map::find(inTag);
  if(lIterObj != end()) {
    std::string lMessage = "The entry \"";
    lMessage += inTag;
    lMessage += "\" is already in the register!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  Map::operator[](inTag) = inEntry;
  addDescription(inTag,inDescription);
}


/*!
 *  \brief Delete an entry of the register.
 *  \param inTag Tag of the entry.
 *  \return Smart pointer to the deleted object.
 *  \throw RunTimeException If the entry is not in the register.
 */
Object::Handle Register::deleteEntry(std::string inTag)
{
  iterator lIterObj = Map::find(inTag);
  if(lIterObj == end()) {
    std::string lMessage = "The entry \"";
    lMessage += inTag;
    lMessage += "\" is not registered!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  Object::Handle lObjHandle = lIterObj->second;
  Map::erase(lIterObj);
  std::map<std::string,Description>::iterator lIterDescrip = mDescriptions.find(inTag);
  if(lIterDescrip == mDescriptions.end()) {
    std::string lMessage = "No description for entry \"";
    lMessage += inTag;
    lMessage += "\"!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  mDescriptions.erase(lIterDescrip);
  return lObjHandle;
}
  

/*!
 *  \brief Get the description of the given entry.
 *  \param inTag Tag name of the entry.
 *  \throw RunTimeException If the tag is not registred.
 */
const Register::Description& Register::getDescription(std::string inTag) const
{
  std::map<std::string,Description>::const_iterator lIterDescrip = mDescriptions.find(inTag);
  if(lIterDescrip == mDescriptions.end()) {
    std::string lMessage = "No description for entry \"";
    lMessage += inTag;
    lMessage += "\"!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  return lIterDescrip->second;
}


/*!
 *  \brief Erase an argument from the command-line.
 *  \param inN Index of the argument to remove.
 *  \param ioArgc Number of elements on the command-line.
 *  \param ioArgv Elements of the command-line.
 *  \return Removed element of the command-line.
 */
char* Register::eraseArg(int inN, int& ioArgc, char** ioArgv) const
{
  Beagle_BoundCheckAssertM(inN, 1, (ioArgc-1));
  char *lTemp = ioArgv[inN];
  for(int i=inN; i<ioArgc-1; i++) ioArgv[i] = ioArgv[i+1];
  ioArgv[ioArgc-1] = NULL;
  ioArgc--;
  return lTemp;
}


/*!
 *  \brief Initialize the register.
 *  \param ioSystem System used.
 *  \param inConfigFileName Configuration file name.
 */
void Register::initialize(System& ioSystem, std::string inConfigFileName)
{
  Beagle_LogDetailedM(
    ioSystem.getLogger(),
    "register", "Beagle::Register",
    "Initializing register"
  );
  
  Description lUsageDescription(
    "Print program usage",
    "",
    "",
    "Shows the Open BEAGLE specific command-line usage."
  );
  addDescription("usage", lUsageDescription);

  Description lHelpDescription(
    "Print detailed help",
    "",
    "",
    "Shows the Open BEAGLE specific command-line usage and detailed parameter descriptions."
  );
  addDescription("help", lHelpDescription);

  if(isRegistered("ec.conf.file")) {
    mFileName = castHandleT<String>(getEntry("ec.conf.file"));
  }
  else {
    mFileName = new String(inConfigFileName);
    std::string lDefaultFileName = std::string("\"") + inConfigFileName + std::string("\"");
    std::string lLongDescripFN = "The name of a configuration file containing ";
    lLongDescripFN += "evolver and parameter values. A typical configuration file can ";
    lLongDescripFN += "be created with parameter \"ec.conf.dump\".";
    Description lFileNameDescription(
      "Configuration filename",
      "String",
      lDefaultFileName,
      lLongDescripFN
    );
    addEntry("ec.conf.file", mFileName, lFileNameDescription);
  }

  if(inConfigFileName.empty() == false) readParametersFile(inConfigFileName, ioSystem);
}


/*!
 *  \brief Interpret the command-line to extract the appropriated arguments.
 *  \param ioSystem System to use to log messages.
 *  \param ioArgc Number of elements of the command-line.
 *  \param ioArgv Elements of the command-line.
 */
void Register::interpretArgs(System& ioSystem, int& ioArgc, char** ioArgv)
{
  for(int i=1; i<ioArgc; i++) {
    char lPrefix[4] = "???";
    strncpy(lPrefix, ioArgv[i], 3);
    std::string lPrefixStr = lPrefix;
    if((lPrefixStr == "-OB") || (lPrefixStr == "-EC")) {
      if(lPrefixStr == "-EC") {
        Beagle_LogBasicM(
          ioSystem.getLogger(),
          "register", "Beagle::Register",
          std::string("The prefix -EC for Open BEAGLE command-line directives is DEPRECATED, ")+
          std::string("please use the prefix -OB instead.")
        );
      }
      char *lArg = eraseArg(i--,ioArgc,ioArgv)+3;
      while(true) {
        std::string lTag(lArg);

        char *lDelimiter = strchr(lArg, '=');
        if(lDelimiter == NULL) {
          if(lTag == "usage") {
            showUsage(ioArgv, std::cerr);
            exit(0);
          }
          else if(lTag == "help") {
            showHelp(ioArgv, std::cerr);
            exit(0);
          }
          else {
            std::cerr << "Bad command-line format!" << std::endl;
            showUsage(ioArgv, std::cerr);
            exit(1);
          }
        }

        *lDelimiter = '\0';
        lTag = lArg;

        iterator lIterTag = Map::find(lTag);
        if(lIterTag == end()) {
          std::cerr << "Bad command-line format!" << std::endl
                    << "The parameter \"" << lTag << "\" is not registred." << std::endl;
          showUsage(ioArgv, std::cerr);
          exit(1);
        }

        char *lNext = strchr(lDelimiter+1,',');
        if(lNext!=NULL) *lNext = '\0';
        std::string lStringValue(lDelimiter+1);
        Beagle_LogDetailedM(
          ioSystem.getLogger(),
          "register", "Beagle::Register",
          std::string("Parameter tag \"")+lTag+std::string("\" with associated value \"")+
          lStringValue+"\" read on the command-line"
        );
        
        try {
          std::istringstream lISStream(lStringValue);
          XMLParser lParser;
          lParser.parse(lISStream);
          XMLNode::Handle lRootNode = lParser.getRootNode();
          lIterTag->second->read(lRootNode);
        } catch(Exception& inException) {
          std::ostringstream lOSS;
          lOSS << "An exception of the type \"";
          lOSS <<  inException.getExceptionName();
          lOSS <<  "\" had been thrown during the command-line parsing. ";
          lOSS <<  "The reason is: ";
          lOSS <<  inException.getMessage();
          throw Beagle_RunTimeExceptionM(lOSS.str());
        }

        // If it is a configuration file name.
        if((lTag == "ec.conf.file") &&
           (lStringValue.empty()==false) &&
           (lStringValue != "")) {
          readParametersFile(lStringValue, ioSystem);
        }
        
        if(lNext != NULL) lArg = lNext+1;
        else break;
      }
    }
  }
}


/*!
 *  \brief To check whether a given tag-entry is registered.
 *  \param inTag Tag of the entry.
 *  \return True if an entry with the given tag is registred, false if not.
 */
bool Register::isRegistered(std::string inTag) const
{
  const_iterator lIterObj = Map::find(inTag);
  return (lIterObj != end());
}


/*!
 *  \brief Modify the smart pointer of an entry in the register.
 *  \param inTag Tag of the entry.
 *  \param inNewEntry New entry that will replace to old one.
 *  \return Handle to the old entry.
 *  \throw RunTimeException If the entry is not found in the register.
 */
Object::Handle Register::modifyEntry(std::string inTag, Object::Handle inNewEntry)
{
  iterator lIterObj = Map::find(inTag);
  if(lIterObj == end()) {
    std::string lMessage = "The entry \"";
    lMessage += inTag;
    lMessage += "\" is not registered!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  Object::Handle lObjHandle = lIterObj->second;
  lIterObj->second = inNewEntry;
  return lObjHandle;
}


/*!
 *  \brief Modify the description of an entry in the register.
 *  \param inTag Tag of the entry.
 *  \param inNewDescription New description that will replace to old one.
 *  \return Old description.
 *  \throw RunTimeException If the entry is not found in the register.
 */
Register::Description Register::modifyDescription(std::string inTag,
                                                  const Register::Description& inNewDescription)
{
  std::map<std::string,Description>::iterator lIterDescrip = mDescriptions.find(inTag);
  if(lIterDescrip == mDescriptions.end()) {
    std::string lMessage = "The entry \"";
    lMessage += inTag;
    lMessage += "\" is not registered!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  Description lDescription = lIterDescrip->second;
  lIterDescrip->second = inNewDescription;
  return lDescription;
}


/*!
 *  \brief Read the parameters from the command-line and files.
 *  \param ioSystem System to use to log messages.
 *  \param ioArgc Number of arguments on the command-line.
 *  \param ioArgv Arguments of the command-line.
 *  \throw Exception If something bad occured ;).
 */
void Register::parseCommandLine(System& ioSystem, int& ioArgc, char** ioArgv)
{
  if(ioSystem.getLogger().getLogLevel() >= (unsigned int)Logger::eInfo) {
    std::string lCommandLine = ioArgv[0];
    for(int i=1; i<ioArgc; ++i) lCommandLine += std::string(" ") + ioArgv[i];
    Beagle_LogInfoM(
      ioSystem.getLogger(),
      "register", "Beagle::Register",
      std::string("Parsing command-line \"")+lCommandLine+std::string("\"")
    );
  }
  std::string lFileName = ioArgv[0];
  if(lFileName.size() > 4) {
    std::string lSuffix(lFileName.end()-4, lFileName.end());
    if(lSuffix == ".exe") lFileName.resize(lFileName.size()-4);
  }
  lFileName += ".conf";
  std::ifstream lDefaultIS(lFileName.c_str());
  bool lGoodFile = lDefaultIS.good();
  lDefaultIS.close();
  if(lGoodFile) readParametersFile(lFileName, ioSystem);

  std::string::size_type lLTLibPos = lFileName.find(std::string(".libs/lt-"));
  if(lLTLibPos != std::string::npos) {
    std::string lFilenameNoLT(lFileName.begin(), lFileName.begin()+lLTLibPos);
    lFilenameNoLT.append(lFileName.begin()+lLTLibPos+9, lFileName.end());
    std::ifstream lDefaultISNoLT(lFilenameNoLT.c_str());
    bool lGoodFileNoLT = lDefaultISNoLT.good();
    lDefaultISNoLT.close();
    if(lGoodFileNoLT) readParametersFile(lFilenameNoLT, ioSystem);
  }

  std::string::size_type lLTPos = lFileName.find(std::string("lt-"));
  if(lLTPos != std::string::npos) {
    std::string lFilenameNoLT(lFileName.begin(), lFileName.begin()+lLTPos);
    lFilenameNoLT.append(lFileName.begin()+lLTPos+3, lFileName.end());
    std::ifstream lDefaultISNoLT(lFilenameNoLT.c_str());
    bool lGoodFileNoLT = lDefaultISNoLT.good();
    lDefaultISNoLT.close();
    if(lGoodFileNoLT) readParametersFile(lFilenameNoLT, ioSystem);
  }
      
  interpretArgs(ioSystem, ioArgc, ioArgv);
}


/*!
 *  \brief Read a register from a Beagle input stream.
 *  \param inNode XML node to read the register from.
 *  \throw Beagle::IOException If reading from the stream problems occured.
 */
void Register::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Register"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Register> expected!");
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if((lChild->getType() == XMLNode::eTag) && (lChild->getTagName() == "Entry")) {
      std::string lEntryKey = lChild->getTagAttribute("key");
      if(lEntryKey.empty())
        throw Beagle_IOExceptionNodeM(*lChild, "no key given for actual entry!");
      if(find(lEntryKey) == end()) continue;
      XMLNode::Handle lChild2 = lChild->getFirstChild();
      (*this)[lEntryKey]->read(lChild2);
    }
  }
}


/*!
 *  \brief Read parameters from a file.
 *  \param inFileName File name of the parameters to read.
 *  \param ioSystem Evolutionary system.
 */
void Register::readParametersFile(std::string inFileName, System& ioSystem)
{
  mFileName->setWrappedValue(inFileName);
#ifdef BEAGLE_HAVE_LIBZ
  igzstream lIFStream(inFileName.c_str());
#else // BEAGLE_HAVE_LIBZ
  std::ifstream lIFStream(inFileName.c_str());
#endif // BEAGLE_HAVE_LIBZ
  if(lIFStream.good() == false) {
    std::string lMessage = "The filename \"";
    lMessage += inFileName;
    lMessage += "\" doesn't refer to a valid parameter file!";
    throw Beagle_RunTimeExceptionM(lMessage);
  }
  Beagle_LogInfoM(
    ioSystem.getLogger(),
    "register", "Beagle::Register",
    std::string("Reading parameter file named \"")+inFileName+"\""
  );
  XMLParser lParser;
  lParser.parse(lIFStream, inFileName.c_str());
  lIFStream.close();
  for(XMLNode::Handle lNode=lParser.getRootNode();
      lNode!=NULL; lNode=lNode->getNextSibling()) {
    if((lNode->getType()==XMLNode::eTag) && (lNode->getTagName()=="Register")) {
      readWithSystem(lNode, ioSystem);
    }
    else if((lNode->getType() == XMLNode::eTag) && (lNode->getTagName() == "Beagle")) {
      for(XMLNode::Handle lChild = lNode->getFirstChild();
          lChild!=NULL; lChild=lChild->getNextSibling()) {
        if((lChild->getType()==XMLNode::eTag) && (lChild->getTagName()=="Register")) {
          readWithSystem(lChild, ioSystem);
        }
      }
    }
  }
}


/*!
 *  \brief Read a register from a Beagle input stream.
 *  \param inNode XML node to read the register from.
 *  \param ioSystem Reference to evolutionary system.
 *  \throw Beagle::IOException If reading from the stream problems occured.
 */
void Register::readWithSystem(XMLNode::Handle& inNode, System& ioSystem)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Register"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Register> expected!");
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if((lChild->getType() == XMLNode::eTag) && (lChild->getTagName() == "Entry")) {
      std::string lEntryKey = lChild->getTagAttribute("key");
      if(lEntryKey.empty())
        throw Beagle_IOExceptionNodeM(*lChild, "no key given for actual entry!");
      if(find(lEntryKey) == end()) {
        Beagle_LogBasicM(
          ioSystem.getLogger(),
          "register", "Beagle::Register",
          std::string("Warning: error while reading register, parameter named \"")+
          lEntryKey+"\" is not registered!"
        );
        continue;
      }
      XMLNode::Handle lChild2 = lChild->getFirstChild();
      (*this)[lEntryKey]->read(lChild2);
    }
  }
}


/*!
 *  \brief Display the program usage in the output stream.
 *  \param ioArgv Elements of the command-line.
 *  \param ioOs C++ output stream to write the usage into.
 */
void Register::showHelp(char** ioArgv, std::ostream& ioOs) const
{
  ioOs << "Open BEAGLE detailed help  (package: ";
  ioOs << BEAGLE_PACKAGE << ", version: " << BEAGLE_VERSION << ")" << std::endl << std::endl;
  ioOs << "usage: " << ioArgv[0] << " [-OBparameter=value] ..." << std::endl;
  ioOs << "       " << ioArgv[0] << " [-OBparam1=value1,param2=value2, ... ,paramN=valueN] ...";
  ioOs << std::endl << std::endl;

  ioOs << "Supported parameters:" << std::endl << std::endl;
  for(std::map<std::string,Description>::const_iterator lIter=mDescriptions.begin();
      lIter!=mDescriptions.end(); ++lIter) {
    ioOs << "  " << lIter->first;

    if(lIter->second.mType.empty() == false) {
      unsigned int lCountSize = lIter->first.size()+2;
      Beagle_AssertM(lCountSize <= 46);
      std::string lTmpStr1(46-lCountSize, ' ');
      ioOs << lTmpStr1;
      ioOs << "<" << lIter->second.mType << "> (def: ";
      ioOs << lIter->second.mDefaultValue << ")";
    }
    ioOs << std::endl;

    std::string lDescription = lIter->second.mDescription;
    wrapString(lDescription, 74);
    std::string lBlank("    ");
    lDescription = lBlank + lDescription;
    for(unsigned int i=0; i<lDescription.size(); i++) {
      if(lDescription[i] == '\n') lDescription.insert(i+1, "    ");
    }
    ioOs << lDescription << std::endl << std::endl;
  }
  ioOs << std::endl << std::flush;
}


/*!
 *  \brief Display the program help in the output stream.
 *  \param ioArgv Elements of the command-line.
 *  \param ioOs C++ output stream to write the usage into.
 */
void Register::showUsage(char** ioArgv, std::ostream& ioOs) const
{
  ioOs << "usage: " << ioArgv[0] << " [-OBparameter=value] ..." << std::endl;
  ioOs << "       " << ioArgv[0] << " [-OBparam1=value1,param2=value2, ... ,paramN=valueN] ...";
  ioOs << std::endl << std::endl;

  for(std::map<std::string,Description>::const_iterator lIter=mDescriptions.begin();
      lIter!=mDescriptions.end(); ++lIter) {
    ioOs << " -OB" << lIter->first;
    unsigned int lCountSize = lIter->first.size()+4;
    if(lIter->second.mType.empty() == false) {
      ioOs << "=<" << lIter->second.mType << ">";
      lCountSize += lIter->second.mType.size()+3;
    }
    Beagle_AssertM(lCountSize <= 46);
    std::string lTmpStr3(46-lCountSize, ' ');
    ioOs << lTmpStr3;
    ioOs << lIter->second.mBrief << std::endl;
  }
  ioOs << std::endl << std::flush;
}
  

/*!
 *  \brief Write a register into a Beagle output stream.
 *  \param ioStreamer XML streamer to read the register to.
 */
void Register::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Register");
  for(const_iterator lIter=begin(); lIter != end(); ++lIter) {
    ioStreamer.openTag("Entry");
    ioStreamer.insertAttribute("key", lIter->first);
    lIter->second->write(ioStreamer);
    ioStreamer.closeTag(false);
    std::map<std::string,Description>::const_iterator lDescription=
      mDescriptions.find(lIter->first);
    if(lDescription != mDescriptions.end()) {
      std::ostringstream lOSS;
      lOSS << "!-- " << lIter->first << " [";
      lOSS << lDescription->second.mType << "]: ";
      lOSS << lDescription->second.mDescription << " --";
      ioStreamer.insertSpecialTag(lOSS.str(), false);
    }
  }
  ioStreamer.closeTag();
}

