/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/ReplacementStrategyOp.cpp
 *  \brief  Implementation of class ReplacementStrategyOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <cmath>

using namespace Beagle;


/*!
 *  \brief Build replacement strategy operator.
 *  \param inName Name of the replacement strategy operator.
 */
ReplacementStrategyOp::ReplacementStrategyOp(std::string inName) :
  Operator(inName)
{ }


/*!
 *  \brief Build routing tables for breeding children.
 *  \param outRoulette Normalized probability tables.
 *  \param ioContext Evolutionary context used.
 */
void ReplacementStrategyOp::buildRoulette(std::vector<double>& outRoulette, Context& ioContext) const
{
  Beagle_NonNullPointerAssertM(getRootNode());
  outRoulette.clear();
  
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "replacement-strategy", "Beagle::ReplacementStrategyOp",
    "Building routing tables of the replacement strategy"
  );

  double lSumProba = 0.0;
  for(BreederNode::Handle lChild=getRootNode(); lChild != NULL;
      lChild=lChild->getNextSibbling()) {
    Beagle_NonNullPointerAssertM(lChild->getBreederOp());
    lSumProba += lChild->getBreederOp()->getBreedingProba(lChild->getFirstChild());
    outRoulette.push_back(lSumProba);
  }

  if(std::fabs(1.0 - outRoulette.back()) > 0.01) {
    Beagle_LogInfoM(
      ioContext.getSystem().getLogger(),
      "replacement-strategy", "Beagle::ReplacementStrategyOp",
      std::string("Sum of probabilities of breeder operators children to ") +
      std::string("replacement strategy named \"") + getName() +
      std::string("\" is different from 1.0 (value: ") + dbl2str(outRoulette.back()) +
      std::string(")")
    );
    Beagle_LogInfoM(
      ioContext.getSystem().getLogger(),
      "replacement-strategy", "Beagle::ReplacementStrategyOp",
      "All probabilities of breeder operators child normalized to get a sum equal to 1.0"
    );
  }
}

                                        
/*!
 *  \brief Initialize the operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void ReplacementStrategyOp::initialize(System& ioSystem)
{
  if(getRootNode() != NULL) getRootNode()->initialize(ioSystem);
}


/*!
 *  \brief Call post initialization hook of the operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void ReplacementStrategyOp::postInit(System& ioSystem)
{
  if(getRootNode() != NULL) getRootNode()->postInit(ioSystem);
}


/*!
 *  \brief Read a replacement strategy operator for XML tree.
 *  \param inNode XML node to use to read replacement strategy operator.
 *  \param inOpMap Operator map to use to read replacement strategy operator.
 */
void ReplacementStrategyOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }

  setRootNode(NULL);
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      OperatorMap::iterator lIterOp = inOpMap.find(lChild->getTagName());
      if(lIterOp == inOpMap.end()) continue;
      
      BreederOp::Handle lOp;
      try {
        lOp = castHandleT<BreederOp>(lIterOp->second);
      }
      catch(BadCastException&) {
        std::ostringstream lOSS;
        lOSS << "Operator \"" << lChild->getTagName() << "\" is not a breeder operator!";
        throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
      }
      lOp = castHandleT<BreederOp>(lOp->giveReference());

      if(getRootNode() == NULL) {
        setRootNode(new BreederNode(lOp));
        getRootNode()->readWithMap(lChild, inOpMap);
      }
      else {
        BreederNode::Handle lBreederChild;
        for(lBreederChild = getRootNode(); lBreederChild->getNextSibbling() != NULL;
            lBreederChild = lBreederChild->getNextSibbling());
        lBreederChild->setNextSibbling(new BreederNode(lOp));
        lBreederChild->getNextSibbling()->readWithMap(lChild, inOpMap);
      }
    }
  }
}


/*!
 *  \brief Select randomly an indice following the probability table.
 *  \param inRoulette Probability table to get an indice.
 *  \param ioContext Evolutionary context to use.
 */
unsigned int ReplacementStrategyOp::rollRoulette(const std::vector<double>& inRoulette,
                                                 Context& ioContext) const
{
  Beagle_AssertM(inRoulette.size() > 0);
  double lRandomNumber =
    ioContext.getSystem().getRandomizer().rollUniform(0.0, inRoulette.back());
  unsigned int lIndex=(inRoulette.size()-1);
  for(; lIndex > 0; --lIndex) {
    if(lRandomNumber > inRoulette[lIndex-1]) break;
  }
  return lIndex;
}


/*!
 *  \brief Write replacement strategy operator into XML streamer.
 *  \param ioStreamer XML streamer to write replacement strategy operator into.
 */
void ReplacementStrategyOp::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag(getName());
  for(BreederNode::Handle lNode=getRootNode(); lNode != NULL;
      lNode=lNode->getNextSibbling()) {
    lNode->write(ioStreamer);
  }
  ioStreamer.closeTag();
}

