/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/SelectRouletteOp.cpp
 *  \brief  Source code of class SelectRouletteOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */
 
#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a proportionnal roulette selection operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register.
 *  \param inName Name of the proportionnal roulette selection operator.
 */
SelectRouletteOp::SelectRouletteOp(std::string inReproProbaName, std::string inName) :
  SelectionOp(inReproProbaName, inName),
  mRouletteValid(false)
{ }


/*!
 *  \brief Build roulette for selection operation.
 *  \param inIndividuals Bag of individuals to use to build roulette.
 */
void SelectRouletteOp::buildRoulette(const Individual::Bag& inIndividuals)
{
  mRoulette.clear();
  double lSumProba = 0.0;
  for(Individual::Bag::size_type i=0; i<inIndividuals.size(); ++i) {
    Beagle_NonNullPointerAssertM(inIndividuals[i]);
    FitnessSimple::Handle lFitness =
      castHandleT<FitnessSimple>(inIndividuals[i]->getFitness());
    Beagle_AssertM(lFitness->getValue() >= 0.0);
    lSumProba += lFitness->getValue();
    mRoulette.push_back(lSumProba);
  }
}


/*!
 *  \brief Apply roulette selection operation.
 *  \param ioDeme Deme on which roulette selection is applied.
 *  \param ioContext Evolutionary context.
 */
void SelectRouletteOp::operate(Deme& ioDeme, Context& ioContext)
{
  buildRoulette(ioDeme);
  mRouletteValid = true;
  SelectionOp::operate(ioDeme, ioContext);
  mRouletteValid = false;
}


/*!
 *  \brief Select an individual using proportional roulette selection.
 *  \param ioPool Pool of individuals on which roulette selection is applied.
 *  \param ioContext Evolutionary context.
 *  \warning To use roulette selection, every individuals composing the population must have
 *    non-negative fitness value. Roulette selection is valid only with FitnessSimple object
 *    and derived (FitnessSimpleMin, FitnessKoza).
 */
unsigned int SelectRouletteOp::selectIndividual(Individual::Bag& ioPool, Context& ioContext)
{
  if(ioPool.size() == 0) return 0;
  if(mRouletteValid == false) buildRoulette(ioPool);
  double lRandomNumber =
    ioContext.getSystem().getRandomizer().rollUniform(0.0, mRoulette.back());
  unsigned int lIndex=(mRoulette.size()-1);
  for(; lIndex > 0; --lIndex) {
    if(lRandomNumber > mRoulette[lIndex-1]) break;
  }
  return lIndex;
}

