/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/SelectionOp.cpp
 *  \brief  Source code of class SelectionOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.6 $
 *  $Date: 2004/07/04 21:45:31 $
 */

#include "beagle/Beagle.hpp"

#include <vector>

using namespace Beagle;


/*!
 *  \brief Construct a new selection operator.
 *  \param inReproProbaName Reproduction probability parameter name used in register. 
 *  \param inName Name of the operator.
 */
SelectionOp::SelectionOp(std::string inReproProbaName, std::string inName) :
  BreederOp(inName),
  mReproProbaName(inReproProbaName)
{ }


/*!
 *  \brief Apply the breeding operation on a breeding pool, returning a bred individual.
 *  \param inBreedingPool Breeding pool to use for the breeding operation.
 *  \param inChild Node handle associated to child node in the breeder tree.
 *  \param ioContext Evolutionary context of the breeding operation.
 *  \return Bred individual.
 */
Individual::Handle SelectionOp::breed(Individual::Bag& inBreedingPool,
                                      BreederNode::Handle inChild,
                                      Context& ioContext)
{
  unsigned int lSelectedIndiv = selectIndividual(inBreedingPool, ioContext);
  Beagle_AssertM(lSelectedIndiv < inBreedingPool.size());
  Individual& lOriginalIndiv = *inBreedingPool[lSelectedIndiv];
  Individual::Handle lClonedIndiv =
    castHandleT<Individual>(ioContext.getDeme().getTypeAlloc()->cloneData(lOriginalIndiv));
  Beagle_NonNullPointerAssertM(lClonedIndiv);
  ioContext.setIndividualIndex(lSelectedIndiv);
  ioContext.setIndividualHandle(lClonedIndiv);
  return lClonedIndiv;
}


/*!
 *  \return Return selection probability of breeder operator.
 *  \param inChild Child node in the breeder tree.
 */
float SelectionOp::getBreedingProba(BreederNode::Handle inChild)
{
  return mReproductionProba->getWrappedValue();
}


/*!
 *  \brief Initialize the operator.
 *  \param ioSystem Reference to the evolutionary system.
 */
void SelectionOp::initialize(System& ioSystem)
{
  BreederOp::initialize(ioSystem);

  if(ioSystem.getRegister().isRegistered(mReproProbaName)) {
    mReproductionProba =
      castHandleT<Float>(ioSystem.getRegister().getEntry(mReproProbaName));
  } else {
    mReproductionProba = new Float(0.1f);
    Register::Description lDescription(
      "Reproduction probability",
      "Float",
      "0.1",
      std::string("Probability than an individual is reproducted as is, without modification. ")+
      std::string("This parameter is useful only in selection and initialization operators that ")+
      std::string("are composing a breeder tree.")
    );
    ioSystem.getRegister().addEntry(mReproProbaName, mReproductionProba, lDescription);
  }
}


/*!
 *  \brief Apply the selection operation on all the deme.
 *  \param ioDeme Deme to which the selection operation is applied.
 *  \param ioContext Evolutionary context.
 */
void SelectionOp::operate(Deme& ioDeme, Context& ioContext)
{
  Beagle_LogTraceM(
    ioContext.getSystem().getLogger(),
    "selection", "Beagle::SelectionOp",
    std::string("Applying selection on the ")+
    uint2ordinal(ioContext.getDemeIndex()+1)+" deme"
  );

  if(ioContext.getGeneration() == 0) {
    Beagle_LogTraceM(
      ioContext.getSystem().getLogger(),
      "selection", "Beagle::SelectionOp",
      std::string("Note that the reproduction probability parameter named \"") + mReproProbaName +
      std::string("\" is ignored when no breeder is set in the evolver, as it seems ") +
      " actually the case."
    );
  }
  
  std::vector<unsigned int> lIndices(ioDeme.size(), 0);
  for(unsigned int i=0; i<ioDeme.size(); i++) lIndices[selectIndividual(ioDeme,ioContext)]++;

  if(ioContext.getSystem().getLogger().getLogLevel() >= Logger::eVerbose) {
    for(unsigned int j=0; j<ioDeme.size(); ++j) {
      Beagle_LogVerboseM(
        ioContext.getSystem().getLogger(),
        "selection", "Beagle::SelectionOp",
        uint2ordinal(j+1)+std::string(" individual selected ")+
        uint2str(lIndices[j])+std::string(" times")
      );
    }
  }
  
  unsigned int lNextEmpty  = 0;
  unsigned int lNextFilled = 0;
  while((lNextFilled < ioDeme.size()) && (lIndices[lNextFilled] <= 1)) lNextFilled++;
  while(lNextFilled < ioDeme.size()) {
    while(lIndices[lNextFilled] > 1) {
      while(lIndices[lNextEmpty] != 0) lNextEmpty++;
      ioDeme.getTypeAlloc()->copyData(*ioDeme[lNextEmpty], *ioDeme[lNextFilled]);
      lIndices[lNextFilled]--;
      lIndices[lNextEmpty]++;
    }
    while((lNextFilled < ioDeme.size()) && (lIndices[lNextFilled] <= 1)) lNextFilled++;
  }
}


/*!
 *  \brief Read a selection operator from XML tree.
 *  \param inNode XML node to use to read selection operator.
 *  \param inOpMap Operator map to use to read selection operator.
 */
void SelectionOp::readWithMap(XMLNode::Handle& inNode, OperatorMap& inOpMap)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != getName())) {
    std::ostringstream lOSS;
    lOSS << "tag <" << getName() << "> expected!" << std::flush;
    throw Beagle_IOExceptionNodeM(*inNode, lOSS.str());
  }
  std::string mReproProbaReadName = inNode->getTagAttribute("repropb");
  if(mReproProbaReadName.empty() == false) mReproProbaName = mReproProbaReadName;
}


/*!
 *  \brief Write selection operator into XML streamer.
 *  \param ioStreamer XML streamer to write selection operator into.
 */
void SelectionOp::writeContent(XMLStreamer& ioStreamer) const
{
  ioStreamer.insertAttribute("repropb", mReproProbaName);
}
