/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Stats.cpp
 *  \brief  Source code of class Stats.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a new stats measure.
 *  \param inId Measure id.
 *  \param inAvg Average of the measure.
 *  \param inStd Standard deviation of the measure.
 *  \param inMax Maximum measure.
 *  \param inMin Minimum measure.
 */
Measure::Measure(std::string inId, double inAvg, double inStd, double inMax, double inMin) :
  mId(inId),
  mAvg(inAvg),
  mStd(inStd),
  mMax(inMax),
  mMin(inMin)
{ }


/*!
 *  \brief Construct new stats.
 *  \param inId Stats id.
 *  \param inGeneration Generation value.
 *  \param inPopSize Population size associated to the statistics.
 *  \param inValid If the stats are valid or not.
 */
Stats::Stats(std::string  inId,
             unsigned int inGeneration,
             unsigned int inPopSize,
             bool         inValid) :
  mId(inId),
  mGeneration(inGeneration),
  mPopSize(inPopSize),
  mValid(inValid)
{ }


/*!
 *  \brief Read stats from a Beagle XML node.
 *  \param inNode XML node to read the stats from.
 */
void Stats::read(XMLNode::Handle& inNode)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Stats"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Stats> expected!");

  std::string lValid = inNode->getTagAttribute("valid");
  if(lValid.empty() || (lValid == "yes")) {
    mValid = true;
        
    mId = inNode->getTagAttribute("id");
  
    std::string lGenerationStr = inNode->getTagAttribute("generation");
    if(lGenerationStr.empty() == false) mGeneration = str2uint(lGenerationStr);
    else mGeneration = 0;

    std::string lPopSizeStr = inNode->getTagAttribute("popsize");
    if(lPopSizeStr.empty() == false) mPopSize = str2uint(lPopSizeStr);
    else mPopSize = 0;
  
    size_type lSize=0;
    for(XMLNode::Handle lChild=inNode->getFirstChild();
        lChild != NULL; lChild=lChild->getNextSibling()) {
      if((lChild->getType() == XMLNode::eTag) && (lChild->getTagName() == "Measure")) ++lSize;
    }

    resize(lSize);
    mItemMap.clear();
      
    size_type lIndexMeasure=0;
    for(XMLNode::Handle lChild=inNode->getFirstChild();
        lChild != NULL; lChild=lChild->getNextSibling()) {
      if(lChild->getType() == XMLNode::eTag) {
        if(lChild->getTagName() == "Item") {
          std::string lKey = lChild->getTagAttribute("key");
          if(lKey.empty()) {
            std::ostringstream lOSS;
            lOSS << "expected a key attribute while reading a statistics item!";
            throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
          }
          XMLNode::Handle lChild2 = lChild->getFirstChild();
          if(lChild2==NULL) {
            std::ostringstream lOSS;
            lOSS << "expected an item value while reading a statistics item!";
            throw Beagle_IOExceptionNodeM(*lChild, lOSS.str());
          }
          if(lChild2->getType() != XMLNode::eString) {
            std::ostringstream lOSS;
            lOSS << "expected an item value while reading a statistics item!";
            throw Beagle_IOExceptionNodeM(*lChild2, lOSS.str());
          }
          mItemMap[lKey] = str2dbl(lChild2->getValue());
        }
        else if(lChild->getTagName() == "Measure") {
          (*this)[lIndexMeasure].mId  = lChild->getTagAttribute("id");
          (*this)[lIndexMeasure].mAvg = 0.0;
          (*this)[lIndexMeasure].mStd = 0.0;
          (*this)[lIndexMeasure].mMax = 0.0;
          (*this)[lIndexMeasure].mMin = 0.0;
          for(XMLNode::Handle lChild2=lChild->getFirstChild();
              lChild2!=NULL; lChild2=lChild2->getNextSibling()) {
            if(lChild2->getType() == XMLNode::eTag) {
              if(lChild2->getTagName() == "Avg") {
                XMLNode::Handle lChild3 = lChild2->getFirstChild();
                if(lChild3->getType() != XMLNode::eString) continue;
                else (*this)[lIndexMeasure].mAvg = str2dbl(lChild3->getValue());
              }
              else if(lChild2->getTagName() == "Std") {
                XMLNode::Handle lChild3 = lChild2->getFirstChild();
                if(lChild3->getType() != XMLNode::eString) continue;
                else (*this)[lIndexMeasure].mStd = str2dbl(lChild3->getValue());
              }
              else if(lChild2->getTagName() == "Max") {
                XMLNode::Handle lChild3 = lChild2->getFirstChild();
                if(lChild3->getType() != XMLNode::eString) continue;
                else (*this)[lIndexMeasure].mMax = str2dbl(lChild3->getValue());
              }
              else if(lChild2->getTagName() == "Min") {
                XMLNode::Handle lChild3 = lChild2->getFirstChild();
                if(lChild3->getType() != XMLNode::eString) continue;
                else (*this)[lIndexMeasure].mMin = str2dbl(lChild3->getValue());
              }
            }
          }
          ++lIndexMeasure;
        }
      }
    }
  }
  else if(lValid == "no") mValid = false;
  else throw Beagle_IOExceptionNodeM((*inNode), "bad stats validity value!");
}


/*!
 *  \brief Write a stats into a XML streamer.
 *  \param ioStreamer XML streamer used to write the stats into.
 */
void Stats::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Stats");
  if(mValid) {
    if(mId.empty() == false) ioStreamer.insertAttribute("id", mId);
    ioStreamer.insertAttribute("generation", uint2str(mGeneration));
    ioStreamer.insertAttribute("popsize", uint2str(mPopSize));
    for(std::map<std::string,double>::const_iterator lIter=mItemMap.begin();
        lIter!=mItemMap.end(); ++lIter) {
      ioStreamer.openTag("Item");
      ioStreamer.insertAttribute("key", lIter->first);
      ioStreamer.insertFloat(lIter->second);
      ioStreamer.closeTag(false);
    }
    for(size_type i=0; i<size(); ++i) {
      ioStreamer.openTag("Measure");
      ioStreamer.insertAttribute("id", (*this)[i].mId);
      ioStreamer.insertFloat((*this)[i].mAvg, "Avg");
      ioStreamer.insertFloat((*this)[i].mStd, "Std");
      ioStreamer.insertFloat((*this)[i].mMax, "Max");
      ioStreamer.insertFloat((*this)[i].mMin, "Min");
      ioStreamer.closeTag();
    }
  }
  else ioStreamer.insertAttribute("valid", "no");
  ioStreamer.closeTag();
}

