/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/System.cpp
 *  \brief  Source code of class System.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a new system with the default context allocator, randomizer, register,
 *    and logger.
 */
System::System() :
  mContextAllocator(new Context::Alloc),
  mRandomizer(new RandomizerMersenne),
  mRegister(new Register),
  mLogger(new LoggerXML)
{ }


/*!
 *  \brief Construct a new system with a given context allocator and default randomizer,
 *    register, and logger.
 *  \param inContextAllocator Handle to the context allocator used.
 */
System::System(Allocator::Handle inContextAllocator) :
  mContextAllocator(castHandleT<Context::Alloc>(inContextAllocator)),
  mRandomizer(new RandomizerMersenne),
  mRegister(new Register),
  mLogger(new LoggerXML)
{ }


/*!
 *  \brief Construct a new system with a given context allocator, randomizer and register.
 *  \param inContextAllocator Handle to the context allocator used.
 *  \param inRandomizer Handle to the randomizer used.
 *  \param inRegister Handle to the register used.
 *  \param inLogger Handle to the logger used.
 */
System::System(Allocator::Handle  inContextAllocator,
               Randomizer::Handle inRandomizer,
               Register::Handle   inRegister,
               Logger::Handle     inLogger) :
  mContextAllocator(castHandleT<Context::Alloc>(inContextAllocator)),
  mRandomizer(inRandomizer),
  mRegister(inRegister),
  mLogger(inLogger)
{ }


/*!
 *  \brief Initialize the system.
 *  \param ioArgc Number of elements on the command-line.
 *  \param ioArgv Element on the command-line.
 */
void System::initialize(int& ioArgc, char** ioArgv)
{
  // Initializing the system.
  Beagle_LogDetailedM(
    getLogger(),
    "system", "Beagle::System",
    "Initializing system"
  );
  mLogger->initialize(*this);
  mRandomizer->initialize(*this);
  mRegister->initialize(*this);

  // Parsing the command-line.
  Beagle_LogTraceM(
    (*mLogger),
    "system", "Beagle::System",
    "Parsing the command-line"
  );
  mRegister->parseCommandLine(*this, ioArgc, ioArgv);
}


/*!
 *  \brief Initialize the system.
 *  \param inConfigFileName Name of the filename containing the system configuration.
 */
void System::initialize(std::string inConfigFileName)
{
  // Initializing the system.
  Beagle_LogDetailedM(
    getLogger(),
    "system", "Beagle::System",
    "Initializing system"
  );
  mLogger->initialize(*this);
  mRandomizer->initialize(*this);
  mRegister->initialize(*this, inConfigFileName);
}


/*!
 *  \brief Post-initialization hook of the system.
 */
void System::postInit()
{
  Beagle_LogDetailedM(
    getLogger(),
    "system", "Beagle::System",
    "Post-initializing system"
  );
  mLogger->postInit(*this);
  mRandomizer->postInit(*this);
  mRegister->postInit(*this);
}

