/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Tokenizer.cpp
 *  \brief  Source code of class Tokenizer.
 *  \author Marc Parizeau
 *  \author Christian Gagne
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Constructs a default tokenizer with no input stream.
 */
Tokenizer::Tokenizer() :
  mLine(1),
  mStream(NULL),
  mWhiteSpace(" \t\n\r")
{ }


/*!
 *  \brief Constructs a tokenizer with default input stream \c inStream.
 *  \param inStream inStream
 */
Tokenizer::Tokenizer(std::istream& inStream) :
  mLine(1),
  mStream(&inStream),
  mWhiteSpace(" \t\n\r")
{ }


/*!
 *  \brief Parse next token of default input stream.
 *  \return Next token found.
 *
 *  This function finds the next token in the input stream, after removing any leading white space. An 
 *  empty token (string) means that end of stream was reached. Throws an std::runtime_error exception 
 *  if no stream is defined.
 */
std::string Tokenizer::getNextToken()
{
  Beagle_NonNullPointerAssertM(mStream);
  std::string lToken;
  getNextToken(lToken);
  return lToken;
}

/*!
 *  \brief Parse next token of default input stream.
 *  \return True if a valid token was found, false otherwise.
 *
 *  This function finds the next token in the default input stream, after removing any leading white 
 *  space. The token is returned through argument \c outToken. An empty token (string) 
 *  means that end of stream was reached. Throws an std::runtime_error exception if no default stream 
 *  is defined.
 */
bool Tokenizer::getNextToken(std::string& outToken)
{
   Beagle_NonNullPointerAssertM(mStream);
   char lChar;
   outToken.erase(outToken.begin(), outToken.end());
   // get rid of leading white space
   while(mStream->get(lChar) && mWhiteSpace.find(lChar) != std::string::npos) if(lChar == '\n') ++mLine;
   // get token; if eof, return empty string
   if(mStream->good()) {
     // not eof
     if(lChar == '\n') ++mLine;
     outToken += lChar;
     if(mSingleCharTokens.find(lChar) == std::string::npos) {
       // not a single char token
       while(mStream->get(lChar)) {
         // append until next white space, single char token, or string token
         if(mWhiteSpace.find(lChar) != std::string::npos ||
            mSingleCharTokens.find(lChar) != std::string::npos) {
           mStream->putback(lChar);
           break;
         }
         else if(lChar == '\n') ++mLine;
           outToken += lChar;
           if(mStringTokens.find(outToken) != mStringTokens.end()) break;
         }
      }
   }
   return !outToken.empty();
}
