/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Vector.cpp
 *  \brief  Source code of class Vector.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/10/28 16:50:43 $
 */

#include <cfloat>
#include "beagle/Beagle.hpp"

using namespace Beagle;


/*!
 *  \brief Construct a real-valued vector of size inSize, with default value inValue.
 *  \param inSize Size of vector.
 *  \param inValue Default value of data in vector.
 */
Vector::Vector(unsigned int inSize, double inValue) :
  Matrix(inSize,1,inValue)
{ }


/*!
 *  \brief Construct a Beagle real-valued vector from a STL vector of double.
 *  \param inDoubleVector STL vector of double used to construct the STL vector.
 */
Vector::Vector(const std::vector<double>& inDoubleVector) :
  Matrix(inDoubleVector)
{ }


/*!
 *  \brief Construct a vector from a column matrix.
 *  \param inMatrix Matrix used to construct the vector.
 *  \throw Beagle::RunTimeException If the matrix given has more than one column.
 */
Vector::Vector(const Matrix& inMatrix) :
  Matrix(inMatrix)
{
  if(getNumberColumns() > 1) {
    throw Beagle_RunTimeExceptionM("Matrix used to construct Vector has more than one column!");
  }
}


/*!
 *  \brief Assign a matrix to a vector
 *  \param inRightMatrix Matrix assigned to vector.
 *  \throw Beagle::RunTimeException If the matrix given has more than one column.
 */
Vector& Vector::operator=(const Matrix& inRightMatrix)
{
  if(this==&inRightMatrix) return *this;
  if(inRightMatrix.getNumberColumns() > 1) {
    throw Beagle_RunTimeExceptionM("Matrix assigned Vector has more than one column!");
  }
  Matrix::operator=(inRightMatrix);
  return *this;
}


/*!
 *  \brief Assign a STL double vector to a vector.
 *  \param inRightVector STL double vector assigned to vector.
 */
Vector& Vector::operator=(const std::vector<double>& inRightVector)
{
  if(this==&inRightVector) return *this;
  Matrix::operator=(inRightVector);
  return *this;
}

  
/*!
 *  \brief Invalid method for class Beagle::Vector.
 *  \throw Beagle::InternalException Unless the method is properly overdefined.
 */
double Vector::computeDeterminant() const
{
  throw Beagle_UndefinedMethodInternalExceptionM("computeDeterminant","Vector");
  return 0.0;
}


/*!
 *  \brief Invalid method for class Beagle::Vector.
 *  \throw Beagle::InternalException Unless the method is properly overdefined.
 */
void Vector::computeEigenSystem(Vector&,Matrix&) const
{
  throw Beagle_UndefinedMethodInternalExceptionM("computeEigenSystem","Vector");
}


/*!
 *  \brief Invalid method for class Beagle::Vector.
 *  \throw Beagle::InternalException Unless the method is properly overdefined.
 */
void Vector::inverse()
{
  throw Beagle_UndefinedMethodInternalExceptionM("inverse","Vector");
}


/*!
 *  \brief Compare rank of two Beagle Vectors.
 *  \param inRightObj Right vector to compare to the actual one.
 *  \return True if actual vector rank comes before right one, false if equal or more.
 */
bool Vector::isLess(const Object& inRightObj) const
{
  const Matrix& lRightMatrix = castObjectT<const Matrix&>(inRightObj);
  const std::vector<double>& lLeft = *this;
  const std::vector<double>& lRight = lRightMatrix;
  return (lLeft < lRight);
}


/*!
 *  \brief Read Beagle Vector from XML node.
 *  \param inNode XML node to read the Beagle Vector from.
 *  \throw Beagle::RunTimeException If read vector got more than one column.
 *
 *  Format of matrix written is simply a text string composed of the floating-point numbers
 *  each separated by one delimiter character. Delimiters characters can be anything except
 *  whitespaces, ';', '.', and number characters.
 *
 */
void Vector::read(XMLNode::Handle& inNode)
{
  Matrix::read(inNode);
  if(getNumberColumns()>1) throw Beagle_RunTimeExceptionM("Vector read has more than one column!");
}


/*!
 *  \brief Resize Beagle Vector.
 *  \param inSize New size of vector.
 *  \param inValue Default value to use to initialize new data.
 */
void Vector::resize(unsigned int inSize, double inValue)
{
  std::vector<double>::resize(inSize, inValue);
  mRows=inSize;
}


/*!
 *  \brief Invalid method for class Beagle::Vector.
 *  \throw Beagle::InternalException Unless the method is properly overdefined.
 */
void Vector::resize(unsigned int, unsigned int, double)
{
  throw Beagle_UndefinedMethodInternalExceptionM("resize","Vector");
}

