/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/Vivarium.cpp
 *  \brief  Source code of class Vivarium.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.9 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <algorithm>

using namespace Beagle;


/*!
 *  \brief Construct a vivarium from a deme allocator.
 *  \param inDemeAlloc Allocator of demes.
 */
Vivarium::Vivarium(Deme::Alloc::Handle inDemeAlloc) :
  Deme::Bag(inDemeAlloc),
  mHOFAlloc(new HallOfFame::Alloc),
  mHallOfFame(new HallOfFame(inDemeAlloc->getContainerTypeAlloc())),
  mStatsAlloc(new Stats::Alloc),
  mStats(new Stats)
{ }


/*!
 *  \brief Construct a vivarium from a deme allocator and a statistics allocator.
 *  \param inDemeAlloc Allocator of demes.
 *  \param inStatsAlloc Statistics allocator.
 *  \param inHOFAlloc Hall-of-fame allocator.
 *  \param inN Number of demes of the vivarium.
 */
Vivarium::Vivarium(Deme::Alloc::Handle inDemeAlloc,
                   Stats::Alloc::Handle inStatsAlloc,
                   HallOfFame::Alloc::Handle inHOFAlloc,
                   size_type inN) :
  Deme::Bag(inDemeAlloc, inN),
  mHOFAlloc(inHOFAlloc),
  mHallOfFame(castObjectT<HallOfFame*>(inHOFAlloc->allocate())),
  mStatsAlloc(inStatsAlloc),
  mStats(castObjectT<Stats*>(inStatsAlloc->allocate()))
{
  mHallOfFame->setIndivAlloc(inDemeAlloc->getContainerTypeAlloc());
}


/*!
 *  \brief Construct a vivarium from a genotype allocator.
 *  \param inGenotypeAlloc Genotype allocator.
 */
Vivarium::Vivarium(Genotype::Alloc::Handle inGenotypeAlloc) :
  Deme::Bag(new Deme::Alloc(new Individual::Alloc(inGenotypeAlloc,new FitnessSimple::Alloc),
                            new Stats::Alloc, new HallOfFame::Alloc)),
  mHOFAlloc(new HallOfFame::Alloc), 
  mHallOfFame(new HallOfFame(new Individual::Alloc(inGenotypeAlloc, new FitnessSimple::Alloc))),
  mStatsAlloc(new Stats::Alloc),
  mStats(new Stats)
{ }


/*!
 *  \brief Construct a vivarium with a genotype allocator and a fitness allocator.
 *  \param inGenotypeAlloc Genotype allocator.
 *  \param inFitnessAlloc Fitness allocator.
 */
Vivarium::Vivarium(Genotype::Alloc::Handle inGenotypeAlloc, Fitness::Alloc::Handle inFitnessAlloc) :
  Deme::Bag(new Deme::Alloc(new Individual::Alloc(inGenotypeAlloc,inFitnessAlloc),
                            new Stats::Alloc, new HallOfFame::Alloc)),
  mHOFAlloc(new HallOfFame::Alloc),
  mHallOfFame(new HallOfFame(new Individual::Alloc(inGenotypeAlloc, inFitnessAlloc))),
  mStatsAlloc(new Stats::Alloc),
  mStats(new Stats)
{ }


/*!
 *  \brief Copy-construct a vivarium.
 *  \param inOrigVivarium Vivarium to copy.
 */
Vivarium::Vivarium(const Vivarium& inOrigVivarium)
{
  (*this) = inOrigVivarium;
}


/*!
 *  \brief Copy a vivarium.
 *  \param inOrigVivarium Vivarium to copy.
 */
Vivarium& Vivarium::operator=(const Vivarium& inOrigVivarium)
{
  if(this == &inOrigVivarium) return *this;
  Deme::Bag::operator=(inOrigVivarium);
  mHOFAlloc = inOrigVivarium.mHOFAlloc;
  mHallOfFame = castObjectT<HallOfFame*>(mHOFAlloc->clone(*inOrigVivarium.mHallOfFame));
  mStatsAlloc = inOrigVivarium.mStatsAlloc;
  mStats = castObjectT<Stats*>(mStatsAlloc->clone(*inOrigVivarium.mStats));
  return *this;
}


/*!
 *  \brief Copy a vivarium and clone the demes that it haves.
 *  \param inOrigContainer Original vivarium to copy.
 *  \throw BadCastException If the vivarium received as argument have no type allocator.
 */
void Vivarium::copyData(const Container& inOrigContainer)
{
  const Vivarium& lRightVivarium = castObjectT<const Vivarium&>(inOrigContainer);
  if(lRightVivarium.getTypeAlloc() == NULL) {
    std::string lMessage = "The copyData() method must be call only with as argument a vivarium";
    lMessage += " that have a type allocator!";
    throw Beagle_InternalExceptionM(lMessage);
  }
  if(this == &lRightVivarium) return;
  (*this) = lRightVivarium;
  clear();
  for(size_type i=0; i<lRightVivarium.size(); ++i) {
    Beagle_NonNullPointerAssertM(lRightVivarium[i]);
    push_back(getTypeAlloc()->cloneData(*lRightVivarium[i]));
  }
}


/*!
 *  \brief Reading a vivarium with method read is undefined. Use readWithContext instead.
 */
void Vivarium::read(XMLNode::Handle&)
{
  throw Beagle_UndefinedMethodInternalExceptionM("read","Vivarium");
}


/*!
 *  \brief Read a vivarium from an XML node.
 *  \param inNode XML node to read the vivarium from.
 *  \param ioContext Evolutionary context to use to read the vivarium.
 *  \throw Beagle::IOException If the format is not respected.
 */
void Vivarium::readWithContext(XMLNode::Handle& inNode, Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Vivarium"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Vivarium> expected!");
  for(XMLNode::Handle lChild=inNode->getFirstChild();
      lChild != NULL; lChild=lChild->getNextSibling()) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Stats") mStats->read(lChild);
      else if(lChild->getTagName() == "HallOfFame") mHallOfFame->readWithContext(lChild, ioContext);
      else if(lChild->getTagName() == "Population") readPopulation(lChild, ioContext);
    }
  }
}


/*!
 *  \brief Read a vivarium population from an XML node.
 *  \param inNode XML node to read the vivarium population from.
 *  \param ioContext Evolutionary context to use to read the vivarium population.
 *  \throw Beagle::IOException If the format is not respected.
 */
void Vivarium::readPopulation(XMLNode::Handle& inNode, Context& ioContext)
{
  if((inNode->getType() != XMLNode::eTag) || (inNode->getTagName() != "Population"))
    throw Beagle_IOExceptionNodeM(*inNode, "tag <Population> expected!");
  XMLNode::Handle lChild = inNode->getFirstChild();
  size_type lSize;
  for(lSize = 0; lChild != NULL; ++lSize) lChild = lChild->getNextSibling();
  if((getTypeAlloc()==NULL) && (lSize > size())) {
    std::ostringstream lMessageOSS;
    lMessageOSS << "Vivarium population size (" << lSize << ") is bigger than the actual size (";
    lMessageOSS << size() << "), and there is no type allocator for resizing!";
    throw Beagle_IOExceptionNodeM(*inNode, lMessageOSS.str());
  }
  resize(lSize);
  lChild = inNode->getFirstChild();
  unsigned int lOldDemeIndex = ioContext.getDemeIndex();
  Deme::Handle lOldDemeHandle = ioContext.getDemeHandle();
  for(size_type j=0; lChild != NULL; ++j) {
    if(lChild->getType() == XMLNode::eTag) {
      if(lChild->getTagName() == "Deme") {
        Beagle_NonNullPointerAssertM((*this)[j]);
        ioContext.setDemeIndex(j);
        ioContext.setDemeHandle((*this)[j]);
        (*this)[j]->readWithContext(lChild, ioContext);
        lChild = lChild->getNextSibling();
      }
    }
  }
  ioContext.setDemeHandle(lOldDemeHandle);
  ioContext.setDemeIndex(lOldDemeIndex);
}


/*!
 *  \brief Write the actual vivarium in a XML streamer.
 *  \param ioStreamer XML streamer to write the vivarium into.
 */
void Vivarium::write(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Vivarium");
  mStats->write(ioStreamer);
  mHallOfFame->write(ioStreamer);
  writePopulation(ioStreamer);
  ioStreamer.closeTag();
}


/*!
 *  \brief Write the vivarium population in a XML streamer.
 *  \param ioStreamer XML streamer to write the vivarium population into.
 */
void Vivarium::writePopulation(XMLStreamer& ioStreamer) const
{
  ioStreamer.openTag("Population");
  ioStreamer.insertAttribute("size", uint2str(size()));
  for(size_type i=0; i<size(); ++i) {
    Beagle_NonNullPointerAssertM((*this)[i]);
    (*this)[i]->write(ioStreamer);
  }
  ioStreamer.closeTag();
}


