/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/XMLNode.cpp
 *  \brief  Source code of class XMLNode.
 *  \author Marc Parizeau
 *  \author Christian Gagne
 *  $Revision: 1.9 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <stdexcept>
#include <fstream>
#include <sstream>
#include <queue>

using namespace Beagle;


/*!
 *  \brief Initializes with first child and next sibling.
 *  \param inChild First child.
 *  \param inSibling Next sibling.
 */
XMLNode::XMLNode(XMLNode::Handle inChild, XMLNode::Handle inSibling) :
  mChild(inChild),
  mSibling(inSibling)
{ }


/*!
 *  \brief Copy construct a XML node.
 *  \param inNode XML node to copy.
 */
XMLNode::XMLNode(const XMLNode& inNode) :
  mType(inNode.mType),
  mValue(inNode.mValue),
  mAttributes(inNode.mAttributes),
  mChild((inNode.mChild != NULL) ? (new XMLNode(inNode.mChild)) : NULL),
  mSibling((inNode.mSibling != NULL) ? (new XMLNode(inNode.mSibling)) : NULL)
{ }


/*!
 *  \brief Destruct a XML node.
 */
XMLNode::~XMLNode()
{
  Handle lIter = mChild;
  mChild = NULL;
  while((lIter!=NULL) && (lIter->getRefCounter()==1)) {
    Handle lTmp = lIter;
    lIter = lIter->mChild;
    lTmp->mChild = NULL;
  }
  lIter = mSibling;
  mSibling = NULL;
  while((lIter!=NULL) && (lIter->getRefCounter()==1)) {
    Handle lTmp = lIter;
    lIter = lIter->mSibling;
    lTmp->mSibling = NULL;
  }
}


/*!
 *  \brief Node assignment by deep copy of a subtree.
 *  \param inNode Subtree to deep copy.
 */
XMLNode& XMLNode::operator=(const XMLNode& inNode)
{
  if(this == &inNode) return *this;
  mType = inNode.mType;
  mValue = inNode.mValue;
  mAttributes = inNode.mAttributes;
  mChild = ((inNode.mChild != NULL) ? (new XMLNode(inNode.mChild)) : NULL);
  mSibling = ((inNode.mSibling != NULL) ? (new XMLNode(inNode.mSibling)) : NULL);
  return *this;
}


/*!
 *  \brief Breadth first search a node with given type and value in actual node subtree.
 *  \author Matthew Walker <m.g.walker@massey.ac.nz>
 *  \author Christian Gagne <cgagne@gel.ulaval.ca>
 *  \author Marc Parizeau <parizeau@gel.ulaval.ca>
 *  \param inValue Value of node to find in subtree.
 *  \param inType Type of node to find in subtree.
 *  \return Handle to XML node find with given specification, NULL handle if no node found.
 */
PointerT<XMLNode,Object::Handle> XMLNode::find(std::string inValue, XMLNode::Type inType) const
{
  std::queue<XMLNode::Handle> lQueue;
  lQueue.push(XMLNode::Handle(this));
  return XMLNode::find(lQueue, inValue, inType);
}


/*!
 *  \brief Breadth first search a node with given type and value in actual node subtree.
 *  \author Matthew Walker <m.g.walker@massey.ac.nz>
 *  \author Christian Gagne <cgagne@gel.ulaval.ca>
 *  \author Marc Parizeau <parizeau@gel.ulaval.ca>
 *  \param ioQueue Queue keeping the state of search in tree. Useful to continue search.
 *  \param inValue Value of node to find in subtree.
 *  \param inType Type of node to find in subtree.
 *  \return Handle to XML node find with given specification, NULL handle if no node found.
 *
 *  This find function version allows repeated search in XML tree by keeping the
 *  state of the search into a queue. When you first call this function, you must provide
 *  a queue with the root node of the subtree to do the search in it (remember that
 *  the actual method is static). Then, the search can be continued by repeated find calls
 *  using the same queue instance, until a NULL handle is returned. Here is an example
 *  on how to use of the actual find method by searching for all the tags named
 *  "Something" in a given XML tree:
 *  \code
      std::queue<XMLNode::Handle> lQueue;
      lQueue.push(lRootNode);
      for(XMLNode::Handle lFound=XMLNode::find(lQueue, "Something", XMLNode::eTag);
          lFound!=NULL; lFound=XMLNode::find(lQueue, "Something", XMLNode::eTag)) {
        // Process lFound here
      }
    \endcode
 */
PointerT<XMLNode,Object::Handle>
XMLNode::find(std::queue< PointerT<XMLNode,Object::Handle> >& ioQueue,
              std::string inValue,
              XMLNode::Type inType)
{
  while(ioQueue.empty()==false) {
    XMLNode::Handle lNode = ioQueue.front();
    for(XMLNode::Handle lChild=lNode->mChild;
        lChild!=NULL; lChild=lChild->mSibling) ioQueue.push(lChild);
    ioQueue.pop();
    if((lNode->mType==inType) && (lNode->mValue==inValue)) return lNode;
  }
  return XMLNode::Handle(NULL);
}


/*!
 *  \brief Get the value of an attribute.
 *  \param inName Name of the attribute to get the value.
 *  \return Value of attribute inName, or empty string if attribute does not exist.
 *  \warning Valid only for tag nodes (\see XMLNode::getType).
 */
std::string XMLNode::getTagAttribute(std::string inName) const
{
  std::map<std::string,std::string>::const_iterator lAttr = mAttributes.find(inName);
  return (lAttr != mAttributes.end() ? (*lAttr).second : std::string(""));
}


/*!
 *  \brief Insert node inChild as first child of this node.
 *  \param inChild New first child to this node.
 */
void XMLNode::insertFirstChild(PointerT<XMLNode,Object::Handle> inChild)
{
   if(inChild == NULL) throw Beagle_RunTimeExceptionM("Cannot insert null node!");
   // search for last sibling of inChild
   XMLNode::Handle lNode;
   for(lNode=inChild; lNode->mSibling != NULL; lNode=lNode->mSibling);
   // attach new child
   lNode->mSibling = mChild;
   mChild = inChild;
}


/*!
 *  \brief Insert node inSibling as next sibling of this node.
 *  \param inSibling New first sibling to this node.
 */
void XMLNode::insertNextSibling(PointerT<XMLNode,Object::Handle> inSibling)
{
   if(inSibling == NULL) throw Beagle_RunTimeExceptionM("Cannot insert null node!");
   // search for last sibling of inSibling
   XMLNode::Handle lNode;
   for(lNode=inSibling; lNode->mSibling != NULL; lNode=lNode->mSibling);
   // attach new sibling
   lNode->mSibling = mSibling;
   mSibling = inSibling;
}


/*!
 *  \brief Remove and delete the first child of this node.
 */
void XMLNode::removeFirstChild()
{
   if(mChild == NULL) throw Beagle_RunTimeExceptionM("Cannot remove null node!");
   XMLNode::Handle lNode = mChild;
   mChild = lNode->mSibling;
   lNode->mSibling = NULL;
}


/*!
 *  \brief Remove and delete the next sibling of this node.
 */
void XMLNode::removeNextSibling()
{
   if(mSibling == NULL) throw Beagle_RunTimeExceptionM("Cannot remove null node!");
   XMLNode::Handle lNode = mSibling;
   mSibling = lNode->mSibling;
   lNode->mSibling = NULL;
}


/*!
 *  \brief Parses next tag.
 *  \param inTokenizer Tokenizer used to parse the stream.
 *  \return Root of corresponding sub-tree.
 *  \throw Beagle::IOException
 */
XMLNode::Handle XMLNode::parse(Tokenizer& inTokenizer)
{
  XMLNode::Handle lNode = NULL;
  // look for start tag
  std::string lToken;
  inTokenizer.setWhiteSpace("");
  inTokenizer.setSingleCharTokens("<");
  if(!inTokenizer.getNextToken(lToken)) return XMLNode::Handle(NULL);
  if(lToken.find_first_not_of(" \t\n\r") == std::string::npos) {
    // flush any leading white space
    if(!inTokenizer.getNextToken(lToken)) return XMLNode::Handle(NULL);
  }
  if(lToken == "<") {
    // check for end tag
    char lChar;
    inTokenizer.getStreamPtr()->get(lChar);
    if(lChar == '/') {
       // found end tag;
       return 0;
    }
    inTokenizer.getStreamPtr()->putback(lChar);
    // found start tag
    lNode = new XMLNode;
    lNode->parseStartTag(inTokenizer, lToken);
    if(lToken == "/") {
       // found empty tag; next token must be '>'
       inTokenizer.setWhiteSpace("");
       inTokenizer.setSingleCharTokens(">");
       if(!inTokenizer.getNextToken(lToken))
         throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Unexpected eof");
       if(lToken != ">")
         throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Invalid start tag");
    }
    else if(lNode->getType() == eTag) {
      XMLNode::Handle lChild;
      XMLNode::Handle lLastChild = NULL;
      while((lChild = parse(inTokenizer)) != NULL) {
        if(lLastChild == NULL) lLastChild = lNode->mChild = lChild;
        else {
          lLastChild->mSibling = lChild;
          lLastChild = lChild;
        }
      }
      // test for valid end tag
      inTokenizer.setWhiteSpace("");
      inTokenizer.setSingleCharTokens(" \t\n\r>");
      if(!inTokenizer.getNextToken(lToken))
        throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Unexpected eof");
      if(lToken != lNode->mValue)
        throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Invalid end tag");
      // next token must be '>'
      inTokenizer.setWhiteSpace(" \t\n\r");
      inTokenizer.setSingleCharTokens(">");
      if(!inTokenizer.getNextToken(lToken))
        throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Unexpected eof");
      if(lToken != ">")
        throw Beagle_IOExceptionTokenizerM(*lNode, inTokenizer, "Invalid end tag");
    }
  }
  else {
    // found a string node
    lNode = new XMLNode;
    lNode->mType = eString;
    lNode->mValue = lToken;
  }
  return lNode;
}


/*!
 *  \brief Parses the attribute list.
 *  \param inTokenizer Tokenizer to use to parse the argument list.
 *  \param outToken Ending token parsed.
 *  \throw Beagle::IOException
 */
void XMLNode::parseAttributeList(Tokenizer& inTokenizer, std::string& outToken)
{
  inTokenizer.setWhiteSpace(" \t\n\r");
  inTokenizer.setSingleCharTokens("=/>");
  // next token should be an attribute name
  if(!inTokenizer.getNextToken(outToken))
    throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
  // parse all attributes
  while(outToken != "/" && outToken != ">")
  {
    // ok, found an attribute name!
    std::string lName = outToken;
    // next token should be '='
    inTokenizer.setSingleCharTokens("=");
    if(!inTokenizer.getNextToken(outToken))
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    if(outToken != "=")
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Invalid attribute");
    inTokenizer.setSingleCharTokens("\"");
    // next token must be '"'
    if(!inTokenizer.getNextToken(outToken))
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    if(outToken != "\"")
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Invalid attribute value");
    // next token is either a value or '"'
    inTokenizer.setWhiteSpace("");
    inTokenizer.setSingleCharTokens("\"");
    if(!inTokenizer.getNextToken(outToken))
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    if(outToken == "\"") {
      // value is empty string
      mAttributes[lName] = "";
    }
    else {
      mAttributes[lName] = outToken;
      inTokenizer.getNextToken(outToken);
      if(outToken != "\"")
        throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    }
    inTokenizer.setWhiteSpace(" \t\n\r");
    inTokenizer.setSingleCharTokens("=/>");
    if(!inTokenizer.getNextToken(outToken))
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
  }
}


/*!
 *  \brief Parses start tag, assuming that token "<" has been read.
 *  \param inTokenizer Tokenizer to use to parse the start tag.
 *  \param outToken Ending token found.
 *  \throw Beagle::IOException
 */
void XMLNode::parseStartTag(Tokenizer& inTokenizer, std::string& outToken)
{
  // parse tag name
  inTokenizer.setWhiteSpace("");
  inTokenizer.setSingleCharTokens(" \t\n\r/>");
  if(!inTokenizer.getNextToken(outToken))
    throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
  if(outToken.find_first_of(" \t\n\r/>") != std::string::npos)
    throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Invalid start tag name");
  switch(outToken[0]) {
  case '?':
  case '!':
    if(outToken[0] == '?') mType = eDecl;
    else mType = eSpecial;
    // parse special tag
    mValue = outToken;
    inTokenizer.setSingleCharTokens(">");
    if(!inTokenizer.getNextToken(outToken))
      throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    if(outToken != ">") {
      mValue += outToken;
      if(!inTokenizer.getNextToken(outToken))
        throw Beagle_IOExceptionTokenizerM(*this, inTokenizer, "Unexpected eof");
    }
    break;
  default:
    // token is regular tag name
    mType = eTag;
    mValue = outToken;
    parseAttributeList(inTokenizer, outToken);
  }
}


/*!
 *  \brief Set attribute inName with value inValue.
 *  \param inName Name of attribute to set value.
 *  \param inValue Value of attribute.
 */
void XMLNode::setTagAttribute(std::string inName, std::string inValue)
{
   mAttributes[inName] = inValue;
}


/*!
 *  \brief Set the tag name; valid only for tag nodes (see XMLNode::setType).
 *  \param inName New name of tag.
 */
void XMLNode::setTagName(std::string inName)
{
   if(mType != eTag) throw Beagle_RunTimeExceptionM("Invalid node type!");
   mValue = inName;
}


/*!
 *  \brief Set the node type (see enum Type).
 *  \param inType New type of node.
 */
void XMLNode::setType(Type inType)
{
   mType = inType;
}


/*!
 *  \brief Set node value (same as XMLNode::setTagName for tag nodes).
 *  \param inValue Set new value of node.
 */
void XMLNode::setValue(std::string inValue)
{
   mValue = inValue;
}


/*!
 *  \brief Serializes the sub-tree rooted by this node.
 *  \param outStream Serialization results.
 *  \param inIndent Indentation active or not.
 */
void XMLNode::serializeXML(XMLStreamer& outStream, bool inIndent) const
{
  bool lLastChildIsTag = true;
  switch(mType)
  {
  case eDecl:
  case eSpecial:
    outStream.insertSpecialTag(getValue());
    break;
  case eTag:
    outStream.openTag(getTagName(), inIndent);
    // serialize attribute list
    for(std::map<std::string,std::string>::const_iterator i = mAttributes.begin();
        i != mAttributes.end(); ++i)
    {
      outStream.insertAttribute((*i).first, (*i).second);
    }
    // serialize child nodes
    for(XMLNode::Handle lChild = mChild; lChild != NULL; lChild = lChild->mSibling)
    {
       lChild->serializeXML(outStream, lLastChildIsTag);
       lLastChildIsTag = (lChild->getType() != eString);
    }
    // indent closing tag only if last child is a tag
    outStream.closeTag(lLastChildIsTag);
    break;
  case eString:
    outStream.insertString(getValue());
    break;
  default:
    throw Beagle_IOExceptionNodeM((*this), "Internal error: unknown node type");
  }
}



