/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/XMLParser.cpp
 *  \brief  Source code of class XMLParser.
 *  \author Marc Parizeau
 *  \author Christian Gagne
 *  $Revision: 1.6 $
 *  $Date: 2004/04/25 23:12:21 $
 */

#include "beagle/Beagle.hpp"

#include <stdexcept>
#include <fstream>
#include <sstream>

using namespace Beagle;


/*!
 *  \return Root tag node of XML document. Return NULL handle if there is no node at the first
 *    level of tag type.
 */
XMLNode::Handle XMLParser::getRootTagNode() const
{
  for(XMLNode::Handle lNode=mRoot; lNode!=NULL; lNode=lNode->mSibling) {
    if(lNode->mType==XMLNode::eTag) return lNode;
  }
  return XMLNode::Handle(NULL);
}


/*!
 *  \brief Parses an XML document contained in a file.
 *  \param inFileName Filename of the XML document to read.
 */
void XMLParser::parse(const std::string& inFileName)
{
  std::ifstream lStream(inFileName.c_str());
  if(!lStream.good()) {
    std::string lMessage = "unable to open file ";
    lMessage += inFileName;
    throw Beagle_IOExceptionMessageM(lMessage);
  }
  parse(lStream, inFileName);
}


/*!
 *  \brief Parses an XML document.
 *  \param inStream Input stream to read the XML document from.
 *  \param inName Stream name for error messages.
 */
void XMLParser::parse(std::istream& inStream, const std::string& inName)
{
  Tokenizer lTokenizer(inStream);
  lTokenizer.setStreamName(inName);
  XMLNode::Handle lNode;
  XMLNode::Handle lLastChild;
  while((lNode = XMLNode::parse(lTokenizer)) != NULL)
  {
    if(mRoot == NULL) lLastChild = mRoot = lNode;
    else {
      lLastChild->mSibling = lNode;
      lLastChild = lNode;
    }
  }
  if(!inStream.eof())
    throw Beagle_IOExceptionTokenizerM(*lNode, lTokenizer, "invalid xml tag");
}
