/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   beagle/src/XMLStreamer.cpp
 *  \brief  Source code of class XMLStreamer.
 *  \author Marc Parizeau
 *  \author Christian Gagne
 *  $Revision: 1.6 $
 *  $Date: 2004/07/04 21:45:31 $
 */

#include "beagle/Beagle.hpp"

#include <stdexcept>

using namespace Beagle;


/*!
 *  \brief Constructs object for streaming XML tags.
 *  \param inStream Output stream in which data is written.
 *  \param inWidth Output indentation width.
 */
XMLStreamer::XMLStreamer(std::ostream& inStream, unsigned int inWidth) :
  mStream(inStream),
  mWidth(inWidth),
  mClosed(true),
  mDirty(false)
{ }


/*!
 *  \brief Closes the last opened tag.
 *  \param inIndent If the parameter is true, end tag indented.
 */
void XMLStreamer::closeTag(bool inIndent)
{
  if(mTags.empty()) throw Beagle_IOExceptionMessageM("No tag to close!");
  if(!mClosed) mStream << "/>" << std::flush;
  else {
    // indentation depends on the number of tags currently on the stack.
    if(inIndent && (mWidth!=0)) {
      mStream << std::endl;
      unsigned int lIndent = mWidth*(mTags.size()-1);
      mStream << std::string(lIndent, ' ');
    }
    mStream << "</" << mTags.top() << ">" << std::flush;
  }
  mClosed = true;
  mTags.pop();
}


/*!
 *  \brief Inserts a string valued attribute into current start tag.
 *  \param inName Name of attribute.
 *  \param inValue Value of attribute.
 */
void XMLStreamer::insertAttribute(const std::string& inName, std::string inValue)
{
  mStream << " " << inName << "=\"" << inValue << "\"";
}


/*!
 *  \brief Inserts a tag with float content.
 *  \param inValue Value of tag content.
 *  \param inName Tag name.
 *  \param inIndent If true, float number is indented if it is tagged.
 *
 *  If no tag name is specified, the float value is inserted into the current tag.
 */
void XMLStreamer::insertFloat(double inValue, const std::string& inName, bool inIndent)
{
  if(inName != "") openTag(inName, inIndent);
  if(!mClosed) mStream << ">";
  mClosed = true;
  if(isNaN(inValue)) mStream << "nan";
  else if(isInfinity(inValue)) {
    if(inValue < 0.0) mStream << "-inf";
    else mStream << "inf";
  }
  else mStream << inValue;
  if(inName != "") closeTag(false);
}


/*!
 *  \brief Insert a XML header in file.
 *  \param inEncoding Encoding of document.
 */
void XMLStreamer::insertHeader(std::string inEncoding)
{
  mStream << "<?xml version=\"1.0\"";
  if(inEncoding.empty() == false) mStream << " encoding=\"" << inEncoding << "\"";
  mStream << "?>" << std::endl << std::flush;
}


/*!
 *  \brief Insert a tag with integer content.
 *  \param inValue Value of tag content.
 *  \param inName Tag name.
 *  \param inIndent If true, integer number is indented if it is tagged.
 *
 *  If no tag name is specified, the integer value is inserted into the current tag.
 */
void XMLStreamer::insertInteger(long int inValue, const std::string& inName, bool inIndent)
{
  if(inName != "") openTag(inName, inIndent);
  if(!mClosed) mStream << ">";
  mClosed = true;
  mStream << inValue;
  if(inName != "") closeTag(false);
}


/*!
 *  \brief Inserts a special tag.
 *  \param inTag Special tag inserted.
 *  \param inIndent If true, integer number is indented if it is tagged.
 */
void XMLStreamer::insertSpecialTag(const std::string& inTag, bool inIndent)
{
  if(!mClosed) {
    mStream << ">";
    mClosed = true;
  }
  // indentation depends on the number of tags currently on the stack.
  if(inIndent && (mWidth!=0)) {
    unsigned int lIndent = mWidth*mTags.size();
    mStream << std::endl << std::string(lIndent, ' ');
  };
  mStream << "<" << inTag << ">";
}


/*!
 *  \brief Inserts a tag with string content.
 *  \param inValue Value of tag content.
 *  \param inName Tag name.
 *  \param inIndent Indent output or not.
 *
 *  If no tag name is specified, the string value is inserted into the current tag.
 */
void XMLStreamer::insertString(const std::string& inValue, const std::string& inName, bool inIndent)
{
  if(inName != "") openTag(inName, inIndent);
  if(inValue.empty() == false) {
    if(!mClosed) mStream << ">";
    mClosed = true;
    mStream << inValue;
  }
  if(inName != "") closeTag(false);
}


/*!
 *  \brief Open new start tag using tag name.
 *  \param inName Tag name of new tag.
 *  \param inIndent Indent (true) os not (false) XML output.
 */
void XMLStreamer::openTag(const std::string& inName, bool inIndent)
{
  if(!mClosed) {
    mStream << ">";
    mClosed = true;
  }
  // indentation depends on the number of tags currently on the stack.
  if(inIndent && (mWidth!=0)) {
    if(mDirty) mStream << std::endl;
    unsigned int lIndent = mWidth*mTags.size();
    mStream << std::string(lIndent, ' ');
  }
  mStream << "<" << inName << std::flush;
  mTags.push(inName);
  mClosed = false;
  mDirty = true;
}


/*!
 *  \brief Insert integer into a Streamer.
 *  \param inValue Integer inserted.
 *  \return Actual XMLStreamer.
 */
XMLStreamer& XMLStreamer::operator<<(int inValue)
{
  insertInteger(inValue);
  return *this;
}


/*!
 *  \brief Insert float into a Streamer.
 *  \param inValue Double inserted.
 *  \return Actual XMLStreamer.
 */
XMLStreamer& XMLStreamer::operator<<(double inValue)
{
  insertFloat(inValue);
  return *this;
}


/*!
 *  \brief Insert string into a Streamer.
 *  \param inValue String inserted.
 *  \return Actual XMLStreamer.
 */
XMLStreamer& XMLStreamer::operator<<(const std::string& inValue)
{
  insertString(inValue);
  return *this;
}


/*!
 *  \brief Insert Beagle Object into a Streamer.
 *  \param inObject Object inserted.
 *  \return Actual XMLStreamer.
 */
XMLStreamer& XMLStreamer::operator<<(const Object& inObject)
{
  inObject.write(*this);
  return *this;
}


