/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   KnapsackEvalOp.cpp
 *  \brief  Implementation of the class KnapsackEvalOp.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.7 $
 *  $Date: 2004/04/25 23:12:22 $
 */

#include "beagle/GA.hpp"
#include "KnapsackEvalOp.hpp"
#include <cmath>

using namespace Beagle;

/*!
 *  \brief Construct the individual evaluation operator for the Knapsack problem.
 *  \param inName Name of the operator.
 */
KnapsackEvalOp::KnapsackEvalOp(std::string inName) :
  EvaluationOp(inName),
  mMaxKnapsackWeight(0.0)
{ }


/*!
 *  \brief Evaluate the fitness of the given individual.
 *  \param inIndividual Current individual to evaluate.
 *  \param ioContext Evolutionary context.
 *  \return Handle to the fitness value of the individual.
 */
Fitness::Handle KnapsackEvalOp::evaluate(Individual& inIndividual, Context& ioContext)
{
  Beagle_AssertM(inIndividual.size() == 1);
  GA::BitString::Handle lBitString = castHandleT<GA::BitString>(inIndividual[0]);
  float lValue  = 0.0;
  float lWeight = 0.0;
  for(unsigned int i=0; i<lBitString->size(); ++i) {
    if((*lBitString)[i]) {
      lValue  += mObjects[i].first;
      lWeight += mObjects[i].second;
    } 
  }
  FitnessMultiObj::Handle lFitness = new FitnessMultiObj(2);
  (*lFitness)[0] = lValue;
  (*lFitness)[1] = (mMaxKnapsackWeight - lWeight);
  return lFitness;
}


/*!
 *  \brief Post-initialize knapsack evaluation operator by generating object's value/weight.
 *  \param ioSystem Evolutionary system.
 */
void KnapsackEvalOp::postInit(System& ioSystem)
{
  mObjects.clear();
  mMaxKnapsackWeight = 0.0;
  float lMaxKnapsackValue  = 0.0;
  for(unsigned int i=0; i<24; ++i) {
    std::pair<float,float> lObject;
    lObject.first  = ioSystem.getRandomizer().rollUniform(1.0, 10.0);
    lObject.second = ioSystem.getRandomizer().rollUniform(1.0, 10.0);
    mObjects.push_back(lObject);
    lMaxKnapsackValue  += lObject.first;
    mMaxKnapsackWeight += lObject.second;
  }
  Beagle_LogInfoM(
    ioSystem.getLogger(),
    "evaluation", "KnapsackEvalOp",
    std::string("Maximum knapsack value is ")+dbl2str(lMaxKnapsackValue)+
    std::string(", while maximum weight is ")+dbl2str(mMaxKnapsackWeight)
  );
}

