/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   ParityMain.cpp
 *  \brief  Implementation of the main routine for the parity-6 problem.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.5 $
 *  $Date: 2004/04/25 23:12:23 $
 */

#include "beagle/GP.hpp"
#include "ParityEvalOp.hpp"

#include <cstdlib>
#include <cmath>
#include <iostream>
#include <stdexcept>
#include <vector>
#include <numeric>

using namespace std;
using namespace Beagle;

/*!
 *  \brief Main routine for the even-6 parity problem.
 *  \param argc Number of arguments on the command-line.
 *  \param argv Arguments on the command-line.
 *  \return Return value of the program.
 *  \ingroup Parity
 */
int main(int argc, char *argv[]) {
  try {
    // 1: Build primitives.
    GP::PrimitiveSuperSet::Handle lSuperSet = new GP::PrimitiveSuperSet;
    GP::PrimitiveSet::Handle lSet1 = new GP::PrimitiveSet;
    lSet1->insert(new GP::And);
    lSet1->insert(new GP::Or);
    lSet1->insert(new GP::Xor);
    lSet1->insert(new GP::Not);
    lSet1->insert(new GP::AdfT<Bool,Bool>(1, 2, "ADF1", "ARG"));
    lSet1->insert(new GP::AdfT<Bool,Bool>(2, 2, "ADF2", "ARG"));
    lSet1->insert(new GP::TokenT<Bool>("IN0"));
    lSet1->insert(new GP::TokenT<Bool>("IN1"));
    lSet1->insert(new GP::TokenT<Bool>("IN2"));
    lSet1->insert(new GP::TokenT<Bool>("IN3"));
    lSet1->insert(new GP::TokenT<Bool>("IN4"));
    lSet1->insert(new GP::TokenT<Bool>("IN5"));
    lSet1->insert(new GP::TokenT<Bool>("0", Bool(false)));
    lSet1->insert(new GP::TokenT<Bool>("1", Bool(true)));
    lSuperSet->insert(lSet1);
    GP::PrimitiveSet::Handle lSet2 = new GP::PrimitiveSet;
    lSet2->insert(new GP::And);
    lSet2->insert(new GP::Or);
    lSet2->insert(new GP::Xor);
    lSet2->insert(new GP::Not);
    lSet2->insert(new GP::AdfT<Bool,Bool>(2, 2, "ADF2", "ARG"));
    lSet2->insert(new GP::TokenT<Bool>("IN0"));
    lSet2->insert(new GP::TokenT<Bool>("IN1"));
    lSet2->insert(new GP::TokenT<Bool>("IN2"));
    lSet2->insert(new GP::TokenT<Bool>("IN3"));
    lSet2->insert(new GP::TokenT<Bool>("IN4"));
    lSet2->insert(new GP::TokenT<Bool>("IN5"));
    lSet2->insert(new GP::TokenT<Bool>("ARG0"));
    lSet2->insert(new GP::TokenT<Bool>("ARG1"));
    lSet2->insert(new GP::TokenT<Bool>("0", Bool(false)));
    lSet2->insert(new GP::TokenT<Bool>("1", Bool(true)));
    lSuperSet->insert(lSet2);
    GP::PrimitiveSet::Handle lSet3 = new GP::PrimitiveSet;
    lSet3->insert(new GP::And);
    lSet3->insert(new GP::Or);
    lSet3->insert(new GP::Xor);
    lSet3->insert(new GP::Not);
    lSet3->insert(new GP::TokenT<Bool>("IN0"));
    lSet3->insert(new GP::TokenT<Bool>("IN1"));
    lSet3->insert(new GP::TokenT<Bool>("IN2"));
    lSet3->insert(new GP::TokenT<Bool>("IN3"));
    lSet3->insert(new GP::TokenT<Bool>("IN4"));
    lSet3->insert(new GP::TokenT<Bool>("IN5"));
    lSet3->insert(new GP::TokenT<Bool>("ARG0"));
    lSet3->insert(new GP::TokenT<Bool>("ARG1"));
    lSet3->insert(new GP::TokenT<Bool>("0", Bool(false)));
    lSet3->insert(new GP::TokenT<Bool>("1", Bool(true)));
    lSuperSet->insert(lSet3);
    // 2: Build a system.
    GP::System::Handle lSystem = new GP::System(lSuperSet);
    // 3: Build operators.
    EvaluationOp::Handle lEvalOp = new ParityEvalOp;
    // 4: Build an evolver and a vivarium.
    Evolver::Handle lEvolver = new GP::Evolver(lEvalOp);
    GP::Vivarium::Handle lVivarium = new GP::Vivarium;
    // 5: Initialize and evolve the vivarium.
    lEvolver->initialize(lSystem, argc, argv);
    lEvolver->evolve(lVivarium);
  }
  catch(Exception& inException) {
    inException.terminate();
  }
  catch(exception& inException) {
    cerr << "Standard exception catched:" << endl;
    cerr << inException.what() << endl << flush;
    return 1;
  }
  catch(...) {
    cerr << "Unknown exception catched!" << endl << flush;
    return 1;
  }
  return 0;
}

