/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   IfThenElse.cpp
 *  \brief  Implementation of the class IfThenElse.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:23 $
 */

#include "beagle/GP.hpp"
#include "IfThenElse.hpp"

#include <cmath>
#include <fstream>

using namespace Beagle;


/*!
 *  \brief Construct a new IfThenElse primitive.
 *  \param inName Name of the conditinal primitive.
 */
IfThenElse::IfThenElse(std::string inName) :
  Beagle::GP::Primitive(3, inName)
{ }


/*!
 *  \brief Return the type tag needed as input for the primitive (0=Bool, {1,2}=Double).
 *  \param inN Index of the argument to get the tag.
 *  \return String tagging the data type.
 *  \throw AssertException If the index inN given is > to 2.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
std::string IfThenElse::getArgType(unsigned int inN) const
{
#ifdef BEAGLE_HAVE_RTTI
  Beagle_AssertM(inN < 3);
  if(inN == 0) return typeid(Bool).name();
  return typeid(Double).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getArgType","IfThenElse");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Return the string tagging the Double type returned but the primitive IfThenElse.
 *  \return String tagging the Double type returned.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
std::string IfThenElse::getReturnType() const
{
#ifdef BEAGLE_HAVE_RTTI
  return typeid(Double).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getReturnType","GP::IfThenElse");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Depending the first argument value, return the 2nd (true) or the 3rd (false).
 *  \param outDatum Result of the comparison.
 *  \param ioContext Evolutionary context.
 */
void IfThenElse::execute(GP::Datum& outDatum, GP::Context& ioContext)
{
  Bool lCondition;
  get1stArgument(lCondition, ioContext);
  if(lCondition == true) get2ndArgument(outDatum, ioContext);
  else get3rdArgument(outDatum, ioContext);
}




