/*
 *  Open BEAGLE
 *  Copyright (C) 2001-2004 by Christian Gagne and Marc Parizeau
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU Lesser General Public
 *  License as published by the Free Software Foundation; either
 *  version 2.1 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  Lesser General Public License for more details.
 *
 *  You should have received a copy of the GNU Lesser General Public
 *  License along with this library; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  Contact:
 *  Laboratoire de Vision et Systemes Numeriques
 *  Departement de genie electrique et de genie informatique
 *  Universite Laval, Quebec, Canada, G1K 7P4
 *  http://vision.gel.ulaval.ca
 *
 */

/*!
 *  \file   LessThan.cpp
 *  \brief  Implementation of the class LessThan.
 *  \author Christian Gagne
 *  \author Marc Parizeau
 *  $Revision: 1.4 $
 *  $Date: 2004/04/25 23:12:23 $
 */

#include "beagle/GP.hpp"
#include "LessThan.hpp"

#include <cmath>
#include <fstream>

using namespace Beagle;


/*!
 *  \brief Construct a new LessThan primitive.
 *  \param inName Name of the comparision primitive.
 */
LessThan::LessThan(std::string inName) :
  Beagle::GP::Primitive(2, inName)
{ }


/*!
 *  \brief Return the Double tag needed as input for the primitive LessThan.
 *  \param inN Index of the argument to get the tag.
 *  \return String tagging the data type Double.
 *  \throw AssertException If the index inN given is > to 1.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
std::string LessThan::getArgType(unsigned int inN) const
{
#ifdef BEAGLE_HAVE_RTTI
  Beagle_AssertM(inN < 2);
  return typeid(Double).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getArgType","LessThan");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Return the string tagging the Bool type returned but the primitive LessThan.
 *  \return String tagging the Bool type returned.
 *  \throw InternalException If RTTI is disabled and the method is not properly overdefined.
 */
std::string LessThan::getReturnType() const
{
#ifdef BEAGLE_HAVE_RTTI
  return typeid(Bool).name();
#else  // BEAGLE_HAVE_RTTI
  throw Beagle_UndefinedMethodInternalExceptionM("getReturnType","GP::LessThan");
  return "";
#endif // BEAGLE_HAVE_RTTI
}


/*!
 *  \brief Execute the comparison operation between two Double, returning a Bool.
 *  \param outDatum Result of the comparison.
 *  \param ioContext Evolutionary context.
 */
void LessThan::execute(GP::Datum& outDatum, GP::Context& ioContext)
{
  Bool& lResult = castObjectT<Bool&>(outDatum);
  Double lArgs[2];
  getArguments(lArgs, sizeof(Double), ioContext);
  lResult = lArgs[0].getWrappedValue() < lArgs[1].getWrappedValue();
}




